
/* -----	Mac Spectacle --------------------------------------------------
			
	This program emulates a ZX Spectrum computer.
	Many thanks to AMSTRAD, the owner of the copyright for the ZX Spectrum ROM,
	to grant permission for it's use in software emulators. The original and unmodified 
	ROM is included in the resource fork of this application.
	
	needs mc68020 CPU or higher or PPC, System 7, color Quickdraw and Quicktime
						
***	Copyright:	1994-1999 Gnter Woigk
  				mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
			
	01.Aug.94		started work								KIO !
	05.Jan.95		Modified multilingual menu handling			
				Added windows menu	   						KIO !
	22.Jan.95		save .Z80 files added							KIO !
	28.Jan.95		command and pc profiles added					KIO !
	18.Jun.95		revised message verbosity handling				KIO !
	26.Dec.99  	made compile again after i had messed up my libs	kio !
*/

// ***** The Main Application File ******************************

#include	<Events.h>
#include	<resources.h>
#include	<files.h>
#include	<quickdraw.h>
#include	<memory.h>
#include	<LowMem.h>
#include	<OSutils.h>
#include	<string.h>

#include	"kio.h"
#include	"options.h"
#include	"z80.options"
#include	"strings.h"
#include	"messages.h"
#include	"load+save.h"
#include	"preferences.h"
#include	"z80 main.h"
#include	"disassembler.h"
#include	"menue.h"
#include	"application.h"
#include	"terminal.h"
#include	"z80.h"
#include	"help+about.h"
#include	"display.h"
#include	"cassette.h"
#include	"movie.h"
#include	"debugger.h"
#include	"profiler.h"
#include	"sound.h"
#include	"keyboard.h"
#include	"windows.h"
#include	"buttons.h"
#include	"freeze.h"


/* **************************************************************** */
/* ********************    A  P  P  L  I  K  A  T  I  O  N    *********************** */
/* **************************************************************** */


short		language;


short 		oDocNumTypes	= 4;			// number of file types to use for OPEN
SFTypeList 	oDocTypeList 	= {'Snap',' Z80',' ROM',' SCR'};	// list of file types 


// -----	Procedure pointers for ExitToShell patch ---------------------------
static UniversalProcPtr OldExitToShellPtr;
static UniversalProcPtr MyExitToShellPtr;


// -----	Prototypes ------------------------------------------------------
extern	short DoBox 		( short id, Str255 s, Str255 t, Str255 u );
extern	HandleTEST		( );
extern	PrintCopyrightKIO	( );
static void MyExitToShell		( void );


// 	Special Test entry in debugger menu
// 	This procedure should do just nothing.
// 	Anything it does is solely for testing pupose
HandleTEST()
{	Char *c; string s; int n;

	PrintLine(CStr("A  SZ-H-VNC BC   DE   HL   ix   iy   PC   SP   im ei"));

	PrintHex2(RA);
	s=SpaceStr(8);
	n=RF; for (c=s.text+8-1; c>=s.text; --c )  {  *c = n&0x01 ? '' : ' '; n>>=1; }
	PrintStr(s);
	PrintHex4(BC);
	PrintHex4(DE);
	PrintHex4(HL);
	PrintHex4(IX);
	PrintHex4(IY);
	PrintHex4(PC);
	PrintHex4(SP);
	PrintHex2(IM);
	PrintChar(IFF1==enabled ? '' : ' ');
	PrintNL();
}


// -----	Copyright ------------------------------------------------
PrintCopyrightKIO()
{	
	PrintLine ( CStr(" Mac Spectacle Version 1.9.7") );
#if GENERATINGPOWERPC
	PrintLine ( CStr("      PowerPC Native Version") );
#else
	PrintLine ( CStr("      68020-68040 Version") );
#endif
	PrintLine ( CatStr (CStr("      Compiled: "), CatStr( CStr(__DATE__), CatStr(CStr(" at "), CStr(__TIME__)) )));
	PrintNL();

	PrintLine ( CStr(" Copyright (c) 1994-2000 Gnter Woigk (kio !)") );
	PrintLine( CStr("      Sankt Johann 6/246; 91056 Erlangen; Germany") );
	PrintLine( CStr("      mailto:kio@little-bat.de") );
	PrintNL();

	PrintLine ( CStr(" C-coded Z80 engine originally based on fMSX") );
	PrintLine ( CStr("      Copyright (C) Marat Fayzullin 1994,1995") );
	PrintNL();

	PrintStr ( CStr("This program is distributed in the hope that it will be useful,") );
	PrintStr ( CStr("but WITHOUT ANY WARRANTY; without even the implied warranty of") );
	PrintStr ( CStr("MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  ") );
	PrintNL();

	PrintStr ( CStr("Permission to use, copy, modify, distribute, and sell this software and") );
	PrintStr ( CStr("its documentation for any purpose is hereby granted without fee, provided") );
	PrintStr ( CStr("that the above copyright notice appear in all copies and that both that") );
	PrintStr ( CStr("copyright notice and this permission notice appear in supporting") );
	PrintStr ( CStr("documentation, and that the name of the copyright holder not be used") );
	PrintStr ( CStr("in advertising or publicity pertaining to distribution of the software") );
	PrintStr ( CStr("without specific, written prior permission.  The copyright holder makes no") );
	PrintStr ( CStr("representations about the suitability of this software for any purpose.") );
	PrintStr ( CStr("It is provided \"as is\" without express or implied warranty.") );
	PrintNL();

	PrintStr ( CStr("THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,") );
	PrintStr ( CStr("INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO") );
	PrintStr ( CStr("EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR") );
	PrintStr ( CStr("CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,") );
	PrintStr ( CStr("DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER") );
	PrintStr ( CStr("TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR") );
	PrintStr ( CStr("PERFORMANCE OF THIS SOFTWARE.") );
	PrintNL();
}


main ( )
{
	MoreMasters();
	MoreMasters();
	InitGraf(&qd.thePort);		
	InitFonts();			
	InitWindows();			
	InitMenus();			
	TEInit();				
	InitDialogs(nil);
	InitCursor();			

	InitStrings(0x8000);

	CheckGestalt();

	OldExitToShellPtr = GetToolTrapAddress(_ExitToShell);
	MyExitToShellPtr  = NewRoutineDescriptor ( (ProcPtr)MyExitToShell, kPascalStackBased, GetCurrentISA() );
	SetToolTrapAddress ( MyExitToShellPtr, _ExitToShell );

	InitAppleEvents();
	InitMenuBar();	terminalmenu=windowsmenu; terminalitem=iConsole;

	Read_CIcons();
	InitDisplay();
	InstallTicker();

	ClearInstrProfiles();
	ClearLocProfiles();

	ReadPreferences();

	PrintCopyrightKIO();
	HandleOptionsMenu(i48k);

	blacked_out = true; HandleOptionsMenu ( iMultitasking );	// switch on

	if (sizeof(Boolean)!=1) do_abort("\psizeof(Boolean)!=1","\p","\p");	// 68k assembler part relies on this ...

	while (!done) RunZ80();

	HandleEXIT();
}


// -----	SHUT DOWN --------------------------------------------------------
HandleEXIT()
{	RemoveTicker(); 
	LMSetSysEvtMask ( LMGetSysEvtMask() & ~keyUpMask );
	MBarControl(multi_window);
	SetToolTrapAddress ( OldExitToShellPtr, _ExitToShell );	// remove patch
}


// -----	ExitToShell trap patch --------------------------------------------
static void MyExitToShell (void)
{	SetCurrentA5();
	HandleEXIT();
	ExitToShell();		// now exit via old ExitToShell
}


static HandleAppleMenu(long itemNum)
{	Str255 daName;
	if (itemNum==iAbout) 	{ if (aboutscreen) SelectWindow(aboutscreen);	else open_about();	return; }
	if (itemNum==iHelp) 		{ if (helpscreen) SelectWindow(helpscreen);	else open_help();	return; }
	GetItem ( applemenu, itemNum, daName );
	OpenDeskAcc ( daName );
}


static HandleSoundMenu ( long itemNum )
{	
	if (itemNum>=iSoundOff && itemNum<=iMaxVolume)
	{	SetSoundVolume ( itemNum-iSoundOff );
		return;
	};

	switch (itemNum)
	{	
	case	i22:		SetSampleRate(sample_rate_22);	return;
	case i44:		SetSampleRate(sample_rate_44);	return;
	case iAY:		SetSoundChannels ( !GetSoundChannels() ); return;
	case iAY_A:	SetSoundChannels ( GetSoundChannels() ^ channel_A ); return;
	case iAY_B:	SetSoundChannels ( GetSoundChannels() ^ channel_B ); return;
	case iAY_C:	SetSoundChannels ( GetSoundChannels() ^ channel_C ); return;
	};
}


StartStop ( Boolean run )
{	
	z80_running = run;
	
	if (!zxwindow) NewSpecciWindow(scrnsize); 
#if include_movie
	StartStopMovie(run);
#endif
}




HandleSpeedMenu ( long itemNum )
{	
	switch(itemNum)
	{	case iUnlimitedSpeed:	
			run_unlimited=true; run_exact=run_highres=false;	
			DisableItem(soundmenu,0); DrawMenuBar(); 
			break;
		case iExactSpeed:	
			run_unlimited=run_highres=false; run_exact=true;	
			EnableItem(soundmenu,0); DrawMenuBar(); 
			break;
		case iHighresMode:	
			run_unlimited=run_exact=false; run_highres=true;	
			EnableItem(soundmenu,0); DrawMenuBar(); 
			break;
		case iBetter:		
			use_best_z80=false; use_better_z80=true;	 	
			DisableItem ( debugmenu, iWritableROM );
			break;
		case iBest:		
			use_better_z80=false; use_best_z80=true;	 	
			EnableItem ( debugmenu, iWritableROM );
			break;
		case iLessVideoAdjust:
			line_cycle_adjust = max(-4,line_cycle_adjust-2); 
			return;
		case iMoreVideoAdjust:
			line_cycle_adjust = min(80,line_cycle_adjust+2); 
			return;
	}

	UpdateSpecciWindow();
	
	CheckItem ( speedmenu, iUnlimitedSpeed,	run_unlimited );
	CheckItem ( speedmenu, iExactSpeed,	run_exact );
	CheckItem ( speedmenu, iHighresMode,	run_highres );
	CheckItem ( speedmenu, iBetter,		use_better_z80 );
	CheckItem ( speedmenu, iBest,			use_best_z80 );
}


HandleFileMenu(long itemNum)
{	StandardFileReply reply;
		
	switch (itemNum)
	{
	case iNew:			HandleNEW();				return; 
	case iSave:			HandleSAVE();		 		return;
	case iCloseTopWindow:	HandleCLOSE(FrontWindow());	return;
	case iOpen:
			if (event.modifiers&(optionKey|shiftKey))
				StandardGetFile ( nil, -1, oDocTypeList, &reply ); 
			else	StandardGetFile ( nil, oDocNumTypes, oDocTypeList, &reply ); 
			if (reply.sfGood) HandleOPEN ( reply.sfFile );
			return;
	case iSuspend:		StartStop(false); return;
	case iResume:		StartStop(true);  return;
	case iQuit:		done=true;		return; 
	case	iPreferences:	WritePreferences();	return;
	};	
}


HandleSAVE()
{	StandardFileReply reply;
	OSErr 	err;
	short	fileID;
	long		fileType;
	Char		*name;
	FInfo		fndrInfo;
	string	p;

	if (!zxwindow) return;
	StandardPutFile  (  localized ( "\pSave file as:","\pDatei speichern als:" ), specciWindowName, &reply );
	if (!reply.sfGood) return;
	
	name = reply.sfFile.name;
	fileType = '????';
	if (*name>=4)
	{	p = UpperStr ( QRightStr(PStr(name),4) );		// file name extension
		switch (*(long*)(p.text))
		{	case '.SNA':
			case 'SNAP':	fileType = 'Snap'; break;
			case '.Z80':	fileType = ' Z80'; break;
			case '.SCR':	fileType = ' SCR'; break;
			case '.ROM':	fileType = ' ROM'; break;
			case '.PIC':
			case 'PICT':	fileType = 'PICT'; break;
			case '.TAP':
			case 'TAPE':	fileType = 'Tape'; break;
#if include_movie
			case '.VHS':	fileType = 'ZMov'; break;
#endif
		}
	};

	if (reply.sfReplacing)
	{	if (fileType=='????') fileType=reply.sfType;				// if file exists, use it's file type as default
		if (fileType!=' Z80' && fileType!='Snap' && fileType!=' ROM' && fileType!=' SCR' && fileType!='PICT') fileType=' Z80';		// if unknown file type, use .Z80 as default
		if (fileType!=reply.sfType)
		{	err = FSpGetFInfo (&reply.sfFile, &fndrInfo);
			fndrInfo.fdType = fileType;
			if (!err) err = FSpSetFInfo (&reply.sfFile, &fndrInfo);
			if (err) return PrintAlert(CStr(localized("Error: failed to change file type","Fehler: Der Dateityp konnte nicht gendert werden")));
		};
	}
	else
	{	if (fileType=='????') fileType=' Z80';	
		if (fileType=='PICT')	err = FSpCreate ( &reply.sfFile, 'ttxt', fileType, 0 );
		else				err = FSpCreate ( &reply.sfFile, 'ZXSP', fileType, 0 );
		if (err)  return PrintNumberAlert(CStr(localized("Error while creating file:","Fehler beim Erstellen der Datei:")),err);
	};
	
	err = FSpOpenDF ( &reply.sfFile, fsWrPerm, &fileID );
	if (err)  return PrintNumberAlert(CStr(localized("Error while opening file:","Fehler beim ffnen der Datei:")),err);
	
	err = SetFPos ( fileID, fsFromStart, 0 );
	if (!err) err = SetEOF ( fileID,0 );
	if (err)  return PrintNumberAlert(CStr(localized("Error while rewinding file:","Fehler beim Zurckspulen der Datei:")),err); 

	if (fileType=='Snap') SaveSNA(fileID); else
	if (fileType==' Z80') SaveZ80(fileID); else
	if (fileType==' ROM') SaveROM(fileID); else
	if (fileType==' SCR') SaveSCR(fileID); else
	if (fileType=='PICT') SavePICT(fileID);
	if (fileType=='Tape') SaveTAPE(fileID);
#if include_movie
	if (fileType=='ZMov') SaveMOVIE(fileID);
#endif

	FSClose ( fileID ); fileID = 0;
	FlushVol (nil, reply.sfFile.vRefNum);
}
	

HandleLoggingMenu ( long itemNum )
{	register long i;

	switch ( itemNum )
	{
	case iLogIrpt:		CheckItem ( logmenu, itemNum, show_irpt ^= 1 ); return;
	case iLogNMI:		CheckItem ( logmenu, itemNum, show_nmi ^= 1 ); return;
	case iLogIllegals:	CheckItem ( logmenu, itemNum, show_illegals ^= 1 ); return;
	case iLogWeird:		CheckItem ( logmenu, itemNum, show_weird ^= 1 ); return;
	case iLogAllLegals:	
	case iLogNoLegals:	show_ei = show_di = show_im0 = show_im1 = show_im2 = 
					show_reti = show_retn = show_halt = show_daa = show_i = show_r =
					show_rst0 = show_rst1 = show_rst2 = show_rst3 = 
					show_rst4 = show_rst5 = show_rst6 = show_rst7 = itemNum==iLogNoLegals;
					for ( i=iLogEI; i<=iLogRST7; i++ ) HandleLoggingMenu(i);	break;
	case iLogEI:		CheckItem ( logmenu, itemNum, show_ei ^= 1 ); break;
	case iLogDI:		CheckItem ( logmenu, itemNum, show_di ^= 1 ); break;
	case iLogIM0:		CheckItem ( logmenu, itemNum, show_im0 ^= 1 ); break;
	case iLogIM1:		CheckItem ( logmenu, itemNum, show_im1 ^= 1 ); break;
	case iLogIM2:		CheckItem ( logmenu, itemNum, show_im2 ^= 1 ); break;
	case iLogRETI:		CheckItem ( logmenu, itemNum, show_reti ^= 1 ); break;
	case iLogRETN:		CheckItem ( logmenu, itemNum, show_retn ^= 1 ); break;
	case iLogHALT:		CheckItem ( logmenu, itemNum, show_halt ^= 1 ); break;
	case iLogDAA:		CheckItem ( logmenu, itemNum, show_daa ^= 1 ); break;
	case iLogR:		CheckItem ( logmenu, itemNum, show_r ^= 1 ); break;
	case iLogI:		CheckItem ( logmenu, itemNum, show_i ^= 1 ); break;
	case iLogRST0:		CheckItem ( logmenu, itemNum, show_rst0 ^= 1 ); break;
	case iLogRST1:		CheckItem ( logmenu, itemNum, show_rst1 ^= 1 ); break;
	case iLogRST2:		CheckItem ( logmenu, itemNum, show_rst2 ^= 1 ); break;
	case iLogRST3:		CheckItem ( logmenu, itemNum, show_rst3 ^= 1 ); break;
	case iLogRST4:		CheckItem ( logmenu, itemNum, show_rst4 ^= 1 ); break;
	case iLogRST5:		CheckItem ( logmenu, itemNum, show_rst5 ^= 1 ); break;
	case iLogRST6:		CheckItem ( logmenu, itemNum, show_rst6 ^= 1 ); break;
	case iLogRST7:		CheckItem ( logmenu, itemNum, show_rst7 ^= 1 ); break;
	}
	CheckItem ( logmenu, iLogAllLegals,	show_ei && show_di && show_im0 && show_im1 && show_im2 && 
								show_reti && show_retn && show_halt && show_daa && show_i && show_r &&
								show_rst0 && show_rst1 && show_rst2 && show_rst3 && 
								show_rst4 && show_rst5 && show_rst6 && show_rst7 ); 
	CheckItem ( logmenu, iLogNoLegals,	!(show_ei || show_di || show_im0 || show_im1 || show_im2 || 
								show_reti || show_retn || show_halt || show_daa || show_i || show_r ||
								show_rst0 || show_rst1 || show_rst2 || show_rst3 || 
								show_rst4 || show_rst5 || show_rst6 || show_rst7) ); 
}


HandleProfileMenu ( long itemNum )
{
	switch ( itemNum )
	{
#if CMD_PROFILE
	case iClearInstrProf:	ClearInstrProfiles();		return;
	case iCountInstr:		CheckItem ( profilemenu, itemNum, count_instr ^= 1 );		
						if (!count_instr && show_1st_instr) HandleProfileMenu(iLogFirstInstr);
						return;
	case iLogFirstInstr:		CheckItem ( profilemenu, itemNum, show_1st_instr ^= 1 );
						if (!count_instr && show_1st_instr) HandleProfileMenu(iCountInstr);
						if (show_1st_location && show_1st_instr) HandleProfileMenu(iLogFirstLoc);
						return;
	case iShowInstrXProf:	ShowCmdProfile();		return;
	case iShowInstrProf:	ShowUsedCmds();		return;
#endif
#if PC_PROFILE
	case iClearLocProf:		ClearLocProfiles();		return;
	case iCountLoc:			CheckItem ( profilemenu, itemNum, count_pc ^= 1 );
						if (!count_pc && show_1st_location) HandleProfileMenu(iLogFirstLoc);
						return;
	case iLogFirstLoc:		CheckItem ( profilemenu, itemNum, show_1st_location ^= 1 );
						if (!count_pc && show_1st_location) HandleProfileMenu(iCountLoc);
						if (show_1st_location && show_1st_instr) HandleProfileMenu(iLogFirstInstr);
						return;
	case iShowLocProf:		ShowPcProfile();		return;
	case iShowLocXProf:		ShowExecutedLocations();	return;
#endif
	default:				PrintAlert(CStr(localized("Unknown item in profile menu selected","Unbekannter Eintrag im Profile-Men ausgewhlt")));
	}
}


HandleDebugMenu ( long itemNum )
{	
	switch ( itemNum )
	{
	case iSuspOnIllegals:		CheckItem ( debugmenu, itemNum, stop_on_illegals ^= 1 );		return;
	case iSuspOnWeird:		CheckItem ( debugmenu, itemNum, stop_on_weird ^= 1 );		return;
	case iTriggerNMI:		WUFF |= is_nmi; 									return;
	case iTriggerIrpt:		WUFF += 1;  WUFF &= 0x0f;							return;
	case iTriggerReset:		HandleReset(false); 									return;
	case iWritableROM:		CheckItem ( debugmenu, itemNum, writable_rom ^= 1 );
						if (writable_rom) wpage[0]=rpage[0]; else wpage[0]=(Char*)norom;
																		return;
	case	iTest:			HandleTEST(); 										return;
	};
}


// -----	Handle options menu -------------------------------------
HandleOptionsMenu ( long itemNum )
{	
	switch (itemNum)
	{	
	case i48k:
		InitZ80(zx48);
		HandleReset(true);		// load ROM, name specci window, reset paging, reset video page, etc.
		CheckItem ( optionsmenu, i48k, true );
		CheckItem ( optionsmenu, i128, false );
		CheckItem ( optionsmenu, iplus2, false );
#if !GENERATINGPOWERPC
		EnableItem ( speedmenu, iBetter );
#endif
		return;

	case i128:
		if (!use_best_z80) HandleSpeedMenu(iBest);
		InitZ80(zx128);
		HandleReset(true);		// load ROMs, name specci window
		CheckItem ( optionsmenu, i48k, false );
		CheckItem ( optionsmenu, i128, true );
		CheckItem ( optionsmenu, iplus2, false );
#if !GENERATINGPOWERPC
		DisableItem ( speedmenu, iBetter );
#endif
		return;

	case iplus2:
		if (!use_best_z80) HandleSpeedMenu(iBest);
		InitZ80(zxplus2);
		HandleReset(true);		// load ROMs, name specci window
		CheckItem ( optionsmenu, i48k, false );
		CheckItem ( optionsmenu, i128, false );
		CheckItem ( optionsmenu, iplus2, true );
#if !GENERATINGPOWERPC
		DisableItem ( speedmenu, iBetter );
#endif
		return;

	case iMultitasking:
		MBarControl ( blacked_out?multi_window:hide_mbar );
		CheckItem ( optionsmenu, iMultitasking, !blacked_out );
		PollPendingEvents();
		return;
	
	case iForceOffscreen:
	case iAvoidOffscreen:
		force_offscreen = itemNum==iForceOffscreen;
		CheckItem ( optionsmenu, iForceOffscreen, force_offscreen );
		CheckItem ( optionsmenu, iAvoidOffscreen, !force_offscreen );
		return;
		
	case iNormalSize:
	case iDoubleSize:	
	case iTripleSize:
	case iQuadSize:
		if (zxwindow)  NewSpecciWindow(itemNum-iNormalSize+1); 
		scrnsize = itemNum-iNormalSize+1;
		CheckItem ( optionsmenu, iNormalSize, itemNum==iNormalSize );
		CheckItem ( optionsmenu, iDoubleSize, itemNum==iDoubleSize );
		CheckItem ( optionsmenu, iTripleSize, itemNum==iTripleSize );
		CheckItem ( optionsmenu, iQuadSize, itemNum==iQuadSize );
		return;

	case iBasicKeys:
	case iGameKeys:
		InitKeyboard(itemNum);
		CheckItem ( optionsmenu, iBasicKeys, itemNum == iBasicKeys );
		CheckItem ( optionsmenu, iGameKeys, itemNum == iGameKeys );
		return;

	case iInternational:
	case iGerman:
		MenuBarNewLanguage(itemNum);
		CheckItem ( optionsmenu, iInternational, itemNum==iInternational );
		CheckItem ( optionsmenu, iGerman, itemNum==iGerman );
		if ( tape_window)	{ SetWTitle (tape_window,	localized("\pTape recorder","\pKassettenrekorder")); RefreshTapeWindow(); }
#if include_movie
		if ( movie_window)	{ SetWTitle (movie_window,	localized("\pMovie recorder","\pFilmrekorder")); RefreshMovieWindow(); }
#endif
		if ( aboutscreen)	{ SetWTitle (aboutscreen,		localized("\pAbout Mac Spectacle","\pber Mac Spectacle")); refresh_about(); }
		if ( terminalwindow)	{ SetWTitle (terminalwindow,	localized("\pConsole","\pKonsole")); }
		return;

	case iLogOnlyCon:
		CloseLogfile(); return;
	case iLogToCon:
		log_con = !log_con; if (!log_file) log_con=true;
		UpdateLogMenu(); return;
	case iLogToFile:
		if (log_file) CloseLogfile(); else OpenLogfile();
		return;

	case iQuiet:		log_level = alert; goto ll;
	case iVerbose:		log_level = msg; goto ll;
	case iEverything:	log_level = debug;
ll:		CheckItem ( optionsmenu, iQuiet, log_level==alert );
		CheckItem ( optionsmenu, iVerbose, log_level==msg );
		CheckItem ( optionsmenu, iEverything, log_level==debug );
		return;

	case iOpenDebugger:
		{	static Boolean debug_open = false;
			CheckItem ( optionsmenu, iOpenDebugger, debug_open^=1 );
			if (debug_open) AddDebuggerMenus(); else RemoveDebuggerMenus();
			return;
		}
	};
}


UpdateLogMenu()
{	
	CheckItem ( optionsmenu, iLogToFile, log_file );
	CheckItem ( optionsmenu, iLogToCon, log_con );
}


// -----	Selection in windows menu ----------------------------------------
static HandleWindowsMenu ( long itemNum )
{	switch (itemNum)
	{
	case	iConsole:			if (terminalwindow) SelectWindow(terminalwindow);	else TermOpen();				return;
	case	iTape:			if (tape_window) SelectWindow(tape_window);		else OpenTapeWindow();			return;
#if include_movie
	case	iMovie:			if (movie_window) SelectWindow(movie_window);	else OpenMovieWindow();			return;
#endif
	case	iKeyboard:		if (helpscreen) SelectWindow(helpscreen);			else open_help();				return;
	case	iAboutWindow:	 	if (aboutscreen) SelectWindow(aboutscreen);		else open_about();				return;
	case iSpectrum:		if (zxwindow) SelectWindow(zxwindow);			else NewSpecciWindow(scrnsize);	return;
	};
}


// -----	Auswahl in einem Men -------------------------------------------
HandleMENU ( short menuNum, short itemNum )
{ 	switch ( menuNum)
	{	
	case mDebug:		HandleDebugMenu(itemNum);		break;
	case mLogging:		HandleLoggingMenu(itemNum);		break;
	case mProfile:		HandleProfileMenu(itemNum);		break;
	case mFile:		HandleFileMenu(itemNum); 		break;
	case mOptions:		HandleOptionsMenu(itemNum);		break;
	case mWindows:	HandleWindowsMenu(itemNum);		break;
	case mApple:		HandleAppleMenu(itemNum);		break;
	case mSpeed:		HandleSpeedMenu(itemNum);		break;
	case mSound:		HandleSoundMenu (itemNum);		break;
	default:			PrintAlert(CStr(localized("Unknown menu selected","Unbekanntes Men selektiert")));
	};
}


// -----	Window refresh -----------------------------------------------
HandleUPDATE(WindowPtr w)	
{	if ( w==terminalwindow)	TermRfsh();
	if ( w==tape_window)	RefreshTapeWindow(); 
#if include_movie
	if ( w==movie_window)	RefreshMovieWindow(); 
#endif
	if ( w==helpscreen)		refresh_help();
	if ( w==aboutscreen)		refresh_about();
	if ( w==zxwindow)		UpdateSpecciWindow();
}


// -----	Drag window ---------------------------------------------------------
HandleDRAG(WindowPtr w, Point p)
{	
	DragWindow ( w, p, &qd.screenBits.bounds );	
	if ( w==zxwindow)  UpdateSpecciWindow();
}


// -----	Close top window ----------------------------------------------
HandleCLOSE ( WindowPtr w )	
{ 	if (w==terminalwindow)	TermClose();
	if (w==zxwindow)		{ DisposeSpecciWindow(); StartStop(false); }
	if (w==tape_window)	CloseTapeWindow();
#if include_movie
	if (w==movie_window)	CloseMovieWindow();
#endif
	if (w==helpscreen)		close_help();
	if (w==aboutscreen)		close_about();
}


// -----	Click in FrontWindow --------------------------------------------------
HandleINCONTENT(WindowPtr w)
{	
	if (w==tape_window) 	{ ClickInTapeWindow(); return; }
#if include_movie
	if (w==movie_window) 	{ ClickInMovieWindow(); return; }
#endif
	if (w==helpscreen) 	{ ClickInKeyboardWindow(event.where); return; }
	LocalToGlobal(&event.where); 
	HandleDRAG ( w, event.where );
}


// -----	Load Snapshot  OR  insert Tape --------------------------------------
HandleOPEN(FSSpec file) 
{	OSErr	err;
	FInfo		info;
	short	fileID;
	long	typ;	
	int i;
	string ns;
	
	if (!zxwindow) HandleNEW();
	if (!zxwindow) return;

	err = FSpGetFInfo ( &file, &info );
	if (err) return PrintNumberAlert ( CStr(localized("Error while accessing file:","Fehler beim Dateizugriff:")),err );
	typ = info.fdType;
	
	if (*file.name>=4)
	{	switch ( *(long*)(LowerStr ( QRightStr(PStr(file.name),4) )).text )
		{
		case '.tap':	typ = 'Tape'; break;	
#if include_movie
		case '.vhs':	typ = 'ZMov'; break;		// <-- suggested for movies
		case '.mov':	typ = 'ZMov'; break;		// <-- probably already used
		case 'zmov':	typ = 'ZMov'; break;		// <-- not that nice
		case 'ovie':	typ = 'ZMov'; break;		// <-- .movie ?!?
#endif
		case '.rom':	typ = ' ROM'; break;
		case '.sna':	typ = 'Snap'; break;
		case '.z80':	typ = ' Z80'; break;
		case '.scr':	typ = ' SCR'; break;
		}
	}

	if (typ=='Tape') 	{ OpenTape(nil,file); return; };
#if include_movie
	if (typ=='ZMov') 	{ OpenMovie(file); return; };
#endif

	err = FSpOpenDF ( &file, fsRdPerm, &fileID );
	if ( err || !fileID )  return PrintNumberAlert(CStr(localized("Error while opening file:","Fehler beim ffnen der Datei:")),err);
	
	err = SetFPos ( fileID, fsFromStart, 0 );
	if (err)  { PrintNumberAlert(CStr(localized("Error while rewinding file:","Fehler beim Zurckspulen der Datei:")),err); goto xx; };

	switch (typ)
	{	case ' ROM':	err=LoadROM(fileID); break;
		case 'Snap':	err=LoadSNA(fileID); break;
		case ' Z80':	err=LoadZ80(fileID);	 break;
		case ' SCR':	err=LoadSCR(fileID);	 break;
		default:		err=error; PrintAlert ( CStr(localized("Unknown file type!","Unbekannter Dateityp!" )));
	};

	if (err)	HandleReset(true);		// reset and reload rom
	else	
	{	SetSpecciWindowName ( file.name );
		
		ns = LowerStr(PStr(file.name));
		i = FindChar(ns,'.',0);
		ns = QLeftStr(ns,i);
		if (SameStr(ns,CStr("academy"))) line_cycle_adjust = 0; else
		if (SameStr(ns,CStr("aquaplan"))) line_cycle_adjust = 74; else
		if (SameStr(ns,CStr("eel_demo"))) line_cycle_adjust = 34; else
		if (SameStr(ns,CStr("hercules"))) line_cycle_adjust = 20; else
		if (SameStr(ns,CStr("loco"))) line_cycle_adjust = -4; else
		if (SameStr(ns,CStr("weird_sc"))) line_cycle_adjust = -4; else line_cycle_adjust = 0;
	}

xx:	FSClose(fileID); fileID=0;
	FlushVol (nil, file.vRefNum);
}
	

// -----	Reset Z80 and open window -----------------------------------------------
HandleNEW()
{	HandleReset(true);
}


// -----	Suspend / Resume application ---------------------------------------------

Boolean	old_z80_running = true;
Boolean	old_blacked_out = false;

HandleSUSPEND()	// Application is juggled out
{	old_z80_running = z80_running;
	old_blacked_out = blacked_out;
	//PrintLine ( CStr(" SUSPEND") );
	if (old_z80_running) HandleFileMenu(iSuspend);
	if (blacked_out) HandleOptionsMenu(iMultitasking);
}

HandleRESUME()	// Application is juggled in
{	//PrintLine ( CStr(" RESUME") );
	if (old_z80_running) HandleFileMenu(iResume);
	if (old_blacked_out) HandleOptionsMenu(iMultitasking);
}


short null_events;

HandleNULL()		// no event available
{	null_events++;
}

PollPendingEvents()
{
	null_events=0;
	while(!null_events) PollEvents(everyEvent,0,nil);
}


HandleSLIDEBAR(WindowPtr, ControlHandle, short){}

HandleFASTUPDATE(WindowPtr w)
{	if (w=terminalwindow) TermFastRfsh();
}


HandleZOOM(WindowPtr w)
{	if ( w==terminalwindow) TermZoom();
	if ( w==tape_window ) TapeZoom();
}


// ***** not used: ***************************************

HandlePRINT(FSSpec)
{	PrintAlert(CStr(localized("Printing of tapes and snaps is not supported.","Kassetten und Snapshots knnen nicht gedruckt werden")));
}

HandleBUTTON		(WindowPtr,ControlHandle) {}
HandleCHECKBOX	(WindowPtr,ControlHandle) {}
HandleGROW(WindowPtr){ }	// in FrontWindow(); only required if scrollbars==false
HandleMOUSEUP()	 	{ }
HandleDISK()			{ }	// Disk inserted (( ejected? ))
HandleDRIVER() 		{ }	// used by system tasks only
HandleAPP3()			{ }	// application event
HandleREGION()			{ }	// mouse leaves 'eventRegion'
HandleACTIVATE()		{ }	// window 'thePort' activated
HandleDEACTIVATE()		{ }	// window 'thePort' deactivated

