
/*	Tape recorder window handler

***	Copyright:	1994-1996 Gnter Woigk
  				mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	01.Jun.94:	Started work on this file											KIO !
	07.Jan.95:	Made tape recorder window title bar 'standard', full multilingual support		KIO !
	09.Jan.95:	now using NewCWindow instead of GetNewCWindow						KIO !
	22.Jan.95:	replaced tape file format with .tap file format as used in Gerton Lunter's "Z80"	KIO !
	11.Jun.95:	watchCursor while loading											KIO !
	19.Nov.95:	using Infinity Windoid, some modifications								KIO !
	12.Oct.96:	dual tape recorder (copy deck)										KIO !
*/

#include	<quickdraw.h>
#include	<windows.h>
#include	<string.h>

#include	"InfinityWindoid.h"
#include	"kio.h"
#include	"strings.h"
#include	"preferences.h"
#include	"menue.h"
#include	"cassette.h"
#include	"terminal.h"
#include	"application.h"
#include	"messages.h"
#include	"display.h"
#include	"keyboard.h"
#include	"z80.h"
#include	"z80_opcodes.h"
#include	"buttons.h"

#define		tape_wdef 	kInfinityZoomProc
//#define		tape_wdef 	noGrowDocProc


#define		WIDTH		250				// tape recorder window width
#define		HEIGHT		102				// tape recorder window height (single recorder window)

#define		SPOOL_DELAY	15				// delay for NextBlock() and PrevBlock()
#define		ANIM_DELAY	4				// video frames per animation frame
#define		SYNC_DELAY	40				// duration of sync sequence in 1/60 sec
#define		BPS			4000			// save/load speed: bytes per second

RGBColor		bkgr_color	= { 34*257,44*257,30*257 };
RGBColor		hilite_color	= { 0*257,187*257,0*257 };


// -----	all required info for 2 x 7 tape recorder buttons + zoom in/out button
btn			button[15];


// -----	button positions
#define		y_btn		73
#define		y_btn2		(y_btn+HEIGHT)

#define		x_x			7
#define		x_o			x_x+36
#define		x_rewind		x_o+36
#define		x_prev		x_rewind+33
#define		x_start		x_prev+33
#define		x_next		x_start+33
#define		x_end		x_next+33

// -----	cicn IDs for normal icons on buttons
#define		i_eject		135				// cicn: insert / eject
#define		i_stop_go		155				// cicn: start tape / pause
#define		i_start		145
#define		i_pause		150
#define		i_rewind		160				// cicn: go to first block of tape
#define		i_end		165				// cicn: go to end of tape
#define		i_prev		170				// cicn: go to previous block of tape
#define		i_next		175				// cicn: go to next block of tape
#define		i_record		180				// cicn: recording on
#define		i_erase		190				// cicn: erase block
#define		i_copy		185				// cicn: copy block
#define		i_new_tape	195				// cicn: create new tape file
enum { grey,green,orange,red,yellow };			// offsets to get hilit versions


// -----	cicn IDs for released and pushed buttons
#define		i_button		130				// cicn: button, released
#define		i_pbutton	 	131				// cicn: button, pressed


// -----	icons & position for flipping rectange to open/close 2nd tape deck
#define		i_zoom_out	128				// cicn: open tape deck 2
#define		i_zoom_in		129				// cicn: close tape deck 2
Rect			r_zoom		= { HEIGHT-8,7,HEIGHT+8,7+12 };


// -----	cicn IDs & locations for 'cassette' logo & animation
#define		i_cassette		200				// cicn: CD player; following: CD window animation 1,2,3,4
Rect			r_cassette	= { 7,7, 7+32,7+32 };


// -----	cicn IDs & locations for LED tape counter
#define		i_led			400				// cicn: LED-"0", normal version; following: 1,2,3,4,5,6,7,8,9,off,-,1,1off,:,:off
#define		y_led		45				// vertical position
Rect			r_led[5]		=  {	y_led,8,  	y_led+21,8+12,  
							y_led,24,	y_led+21,24+12,  
							y_led,37,	y_led+21,37+8,
							y_led,45,	y_led+21,45+12,
							y_led,61,	y_led+21,61+12	};

// -----	pict & location of slide bar indicator
#define		i_slide		133	 							// pict: slide bar indicator
Rect			r_slide_o		= { 54,x_rewind, 54+12,WIDTH-7 };		// outer box (whole pict)
Rect			r_slide_i		= { 59,x_rewind+5,59+2,WIDTH-12 };	// inner box (slider opening)
short		w_slider		= 5;								// slider width
PicHandle		slide_pic		= nil;

// -----	file name, block info, popup box
#define		textsize		9								// text size of file name, block info and popup menu
short		x_filename	= 80;							// position of file name
short		y_filename	= 7+textsize;
Rect			r_filename	= { 7,x_rewind, 9+textsize,WIDTH-7 };
short		x_blockinfo	= x_rewind+3;						// position of block info
short		y_blockinfo	= 25+textsize;						// position of block info
Rect			r_blockinfo	= { 23, x_rewind, 29+textsize, WIDTH-8 };


WindowPtr	tape_window	= nil;
Rect			tapeRect		= { 0,0,HEIGHT,WIDTH};				// tape window rect in glob. coord (as used for prefs)

Rect			tape_rect		= { 0,0,HEIGHT,WIDTH };
Rect			r_tape_hi		= { 2,2,HEIGHT-3,WIDTH-3 };
Rect			r_tape_lo		= { 3,3,HEIGHT-2,WIDTH-2 };

tap			tape[2];



/***************************************************************
***************************************************************/


static OSErr err	= ok;


// -----	get actual block length -----------------------------------------------------
//		-> err
static Short block_len ( tap *tap )
{	Short len;

	err = SetFPos (tap->id, fsFromStart, tap->pos);
	if (!err) err = File2Buffer(tap->id,&len,2);
	return (len<<8)+(len>>8);
}


// -----	get actual block type -----------------------------------------------------
//		-> err
static Char block_type ( tap *tap )
{	Char type;

	err = SetFPos (tap->id, fsFromStart, tap->pos+2);
	if (!err) err = File2Buffer(tap->id,&type,1);
	return type;
}


// -----	copy screen contents -> highres screen buffer ----------------------------------------
static update_highres ( )
{	
	while (!EXIT) Do_Cycles();
	EXIT=false;
	while (!EXIT) Do_Cycles();
}


// -----	animate cassette icon ---------------------------------------------------------------
animate (tap *tap);
animate (tap *tap)
{	long n;

	n = TickCount()+ANIM_DELAY;

	SetPort(tape_window);
	SetOrigin ( 0, tap->dy );
	tap->frame = (tap->frame+1) % 5;
	PlotCIcon ( &r_cassette, Get_CIcon(i_cassette+tap->frame) );
	SetOrigin(0,0);

	RefreshSpecciWindow();
	while ( (long)TickCount()-n<0 ) PollEvents ( updateMask, n-TickCount(), nil );
}


// -----	draw dimmed or highlighted frame around controls -------------------------------------
draw_frame ( tap *tap );
draw_frame ( tap *tap )
{	RGBColor rgb;

	SetPort(tape_window);
	SetOrigin ( 0, tap->dy );

	ForeColor ( blackColor );
	FrameRect ( &r_tape_lo );

	if (tap->selected)
		RGBForeColor ( &hilite_color );					
	else
	{	rgb.red = bkgr_color.red *2;
		rgb.green = bkgr_color.green *2;
		rgb.blue = bkgr_color.blue *2;
		RGBForeColor ( &rgb );
	}
	FrameRect ( &r_tape_hi );
	SetOrigin(0,0);
	Draw_Button ( button+14,false );
}


// -----	draw tape counter and slide bar indicator ---------------------------------------
draw_counter ( tap *tap );
draw_counter ( tap *tap )
{	int i;
	Rect box;
	short dx;

	SetPort(tape_window);
	SetOrigin(0,tap->dy);

	i = tap->id ? tap->pos*8/1500 : 0;
	PlotCIcon ( r_led+0, Get_CIcon(i_led + i/600) );
	PlotCIcon ( r_led+1, Get_CIcon(i_led + (i%600)/60) );
	PlotCIcon ( r_led+2, Get_CIcon(415));
	PlotCIcon ( r_led+3, Get_CIcon(i_led + (i%60)/10) );
	PlotCIcon ( r_led+4, Get_CIcon(i_led + i%10) );

	if (tap->id)
	{	
		dx = tap->pos * ((r_slide_i.right-r_slide_i.left)-w_slider)/max(tap->len,1);

		ForeColor(blackColor);
		box = r_slide_i; box.right = box.left+dx;
		PaintRect(&box);
		box = r_slide_i; box.left += dx+w_slider;
		PaintRect(&box);

		ForeColor ( yellowColor );
		box = r_slide_i; box.left += dx; box.right = box.left+5;
		PaintRect ( &box );
	}
	else
	{	ForeColor(blackColor);
		PaintRect(&r_slide_i);
	}

	SetOrigin(0,0);
}


// -----	draw block info, tape counter and slide bar indicator --------------------------
draw_blockinfo ( tap *tap );
draw_blockinfo ( tap *tap )
{	Char ns[12];
	Short len; Char typ;
	struct { Char dummy, art, name[10]; Short clen,cadr; } head;
	

	SetPort(tape_window);
	SetOrigin(0,tap->dy);

	ForeColor ( whiteColor );
	EraseRect ( &r_blockinfo );
	FrameRect ( &r_blockinfo );	

	draw_counter(tap);
	SetOrigin(0,0);
	
	if (!tap->id) return;

	ForeColor ( whiteColor );
	MoveTo ( x_blockinfo,y_blockinfo-tap->dy );

	if ( tap->pos>=tap->len )
	{	if (tap->pos==0)	DrawString ( localized ("\pEmpty tape","\pLeere Kassette") );		else
						DrawString ( localized ("\pEnd of tape","\pKasetten-Ende") );
		return;
	}

	len = block_len(tap) - 2; if(err) goto xx;
	typ = block_type(tap); if (err) goto xx;

	if ( typ==255 )			// data
	{	DrawString(localized("\pBlock of data","\pDatenblock")); 
		goto pl;
	}

	if ( typ==0 )			// header
	{	if (len!=17)
		{ 	DrawString("\pLong Header");	goto pl;
		}

		err = File2Buffer(tap->id,&head.art,15); if (err) goto xx;
		head.clen = (head.clen>>8)+(head.clen<<8);
		head.cadr = (head.cadr>>8)+(head.cadr<<8);
		switch ( head.art )
		{	
		case 0:	DrawString("\pPROG"); break;
		case 1:
		case 2:	DrawString("\pDATA"); break;
		case 3:	DrawString("\pCODE");  break;
		default:	DrawString("\pLong Header"); goto pl;
		};
		
		if (*head.name!=0x00ff )
		{	DrawString ("\p: \"");DrawText ( head.name, 0, 10 );DrawChar ('"');
		}
		else
		{	DrawString (localized("\punnamed","\pohne Namen"));
		};
		
		DrawString(localized("\p; Length: ","\p; Lnge: ")); 
		NumToString(head.clen,ns);
		DrawString(ns);
		return;
	}
	
	DrawString(localized("\pUndefined block type","\pUnbekannter Blocktyp")); 
pl:	DrawString(localized("\p; Length: ","\p; Lnge: ")); 
	NumToString(len,ns);
	DrawString(ns);
	return;

xx:	DrawString ( localized("\p File error: ", "\p Dateifehler: ") );
	NumToString(err,ns);
	DrawString(ns);
}


// -----	draw file name, block info, tape counter and slide bar indicator ------------------
draw_filename ( tap *tap );
draw_filename ( tap *tap )
{	
	SetPort(tape_window);
	SetOrigin(0,tap->dy);

	EraseRect ( &r_filename );
	ForeColor ( whiteColor );
	MoveTo ( x_filename,y_filename );
	if (tap->id) DrawString ( tap->name ); else DrawString ( localized ("\p(No tape loaded)","\p(Keine Kassette eingelegt)") );

	draw_blockinfo(tap);
}


// -----	draw background, frame, file name, block info, tape counter and slide bar indicator -------
//		-> background
//		-> slide bar
//		-> icon
refresh_tape ( tap *tap );
refresh_tape ( tap *tap )
{
	if (!tap->visible) return;
	SetPort(tape_window);
	SetOrigin ( 0, tap->dy );

	EraseRect ( &tape_rect );
	PlotCIcon ( &r_cassette, Get_CIcon(i_cassette+tap->frame) );
	
	if ( !slide_pic )	slide_pic = GetPicture ( i_slide ); 
	if ( slide_pic && *slide_pic ) DrawPicture ( slide_pic, &r_slide_o );

	draw_filename(tap);
	draw_frame(tap);
}


// -----	Redraw window -------------------------------------------------------
RefreshTapeWindow ()
{	int i;

	if ( !tape_window ) return;
	
	refresh_tape ( tape );
	refresh_tape ( tape+1 );

	for (i=0;i<=14;i++) Draw_Button ( button+i, 0 );	// buttons
}


/*****************************************************************/
/*****************************************************************/
/*****************************************************************/



// -----	Eject selected tape ------------------------------------------------------------------
eject_tape ( tap *tap );
eject_tape ( tap *tap )
{	int i;

	if (!tap->id) return;		// no tape loaded

	FSClose(tap->id);
	tap->id = 0;
	tap->running=tap->recording=false;
	tap->pos=tap->len=0;

	if (!tape_window || !tap->visible) return;
	draw_filename(tap);
	
	if (tap==tape)	for (i=0;i<7;i++) Set_Button(button+i,0);
	else			for (i=7;i<14;i++) Set_Button(button+i,0);
}


// -----	Open new tape file -----------------------------------------------------
OpenTape ( tap *tap, FSSpec file )
{	OSErr err; int i;

	OpenTapeWindow();
	if (!tap) tap = tape+!tape->selected;
	eject_tape(tap);

	err = FSpOpenDF ( &file, fsCurPerm, &tap->id );
	if (err) { tap->id=0; PrintNumberAlert(CStr(localized("Error while opening tape:","Fehler beim Einlegen der Kassette:")),err); return; };

	memcpy ( tap->name, file.name, 64 );
	tap->running = true;
	tap->recording = false;
	tap->pos = 0;
	err = GetEOF (tap->id, &tap->len); if (err) { eject_tape(tap); return; }	// should never happen
	
	if (!tape_window || !tap->visible) return;
	draw_filename(tap);

	if (tap==tape)	for (i=0;i<7;i++) Set_Button(button+i,1);
	else			for (i=7;i<14;i++) Set_Button(button+i,1);
}


// -----	Handle "Insert Tape" --------------------------------------------------------------
//		Ask for a file and open it
static insert_tape ( tap *tap )
{	StandardFileReply reply;	
	SFTypeList	typeList 	= {'Tape',0,0,0};	// list of file types 

	StandardGetFile ( nil, 1, typeList, &reply ); 
	if (reply.sfGood) OpenTape(tap,reply.sfFile);
}


static set_btn ( short idx, short o, short l, short i0, short i1, short i2, short state)
{	btn *b; Rect r;

	b=button+idx;
	r=(**Get_CIcon(i_button)).iconPMap.bounds;
		
	Link ( button, b );

	SetRect ( &b->bbox, l,o,l+r.right-r.left,o+r.bottom-r.top );
	b->btn[0] = b->btn[1] = i_button;
	b->pbtn[0] = b->pbtn[1] = i_pbutton;
	b->icn[0] = i0;
	b->icn[1] = i1;
	b->icn[2] = i2;
	b->icn[3] = nil;
	b->nxt_sta[0] = 0;
	b->nxt_sta[1] = 2;
	b->nxt_sta[2] = 1;
	b->nxt_sta[3] = 0;
	b->state = state;
	b->id = idx;
}
	

// -----	Close the tape recorder window -----------------------------------------------------
//		updates tapeRect for preferences
CloseTapeWindow ()
{	
	CheckItem ( windowsmenu, iTape, false );
	tape[0].visible = tape[1].visible = false;
	if (!tape_window) return;
	SetPort(tape_window);
	tapeRect = tape_window->portRect;
	LocalToGlobal ( (Point*)&tapeRect ); 
	LocalToGlobal (((Point*)&tapeRect)+1 ); 
	DisposeWindow ( tape_window ); 
	tape_window=nil;
}


// -----	Open the tape recorder window ------------------------------------------------------
OpenTapeWindow ()
{	short i; 
	btn *b;
	
	if (tape_window) goto xx;							// already open -> only SetPort() etc.

	tape[0].visible=true;
	tape[1].visible=tape[1].selected||tape[1].id;

	if ( tape[0].selected || tape[1].selected ) goto yy;		// re-open: tape[] and button[] already initialized

	tape[0].selected=true;	tape[0].frame = 0;	tape[0].dy = 0;
	tape[1].selected=false;	tape[1].frame = 0;	tape[1].dy = -HEIGHT;

// init linked list of buttons
	button->next = button->prev = button;
// 7 buttons for tape recorder 1
	i=tape[0].id ? 1 : 0;
	set_btn ( 0,	y_btn,x_x,		i_new_tape, i_erase,i_erase+red,		i); button[0].nxt_sta[0]=3; button[0].nxt_sta[3]=0; button[0].icn[3]=i_new_tape+green;
	set_btn ( 1,	y_btn,x_o,		i_record, i_record, i_record+red,		i);
	set_btn ( 2,	y_btn,x_rewind,	i_rewind, i_rewind, i_rewind+green, 	i);
	set_btn ( 3,	y_btn,x_prev,		i_prev, i_prev, i_prev+green,			i);
	set_btn ( 4,	y_btn,x_start,		i_stop_go, i_start+green, i_pause+red,	i);
	set_btn ( 5,	y_btn,x_next,		i_next, i_next, i_next+green,			i);
	set_btn ( 6,	y_btn,x_end,		i_end, i_end, i_end+green,			i);
// 7 buttons for tape recorder 2
	i=tape[1].id ? 1 : 0;
	set_btn ( 7,	y_btn2,x_x,		i_new_tape, i_erase,i_erase+red,		i); button[7].nxt_sta[0]=3; button[7].nxt_sta[3]=0; button[7].icn[3]=i_new_tape+green;
	set_btn ( 8,	y_btn2,x_o,		i_copy, i_copy, i_copy+green,			i);
	set_btn ( 9,	y_btn2,x_rewind,	i_rewind, i_rewind, i_rewind+green, 	i);
	set_btn ( 10,	y_btn2,x_prev,	i_prev, i_prev, i_prev+green,	 		i);
	set_btn ( 11,	y_btn2,x_start,	i_stop_go, i_start+green, i_pause+red,	i);
	set_btn ( 12,	y_btn2,x_next,	i_next, i_next, i_next+green,			i);
	set_btn ( 13,	y_btn2,x_end,		i_end, i_end, i_end+green,			i);
// button for tape recorder 2 on/off
	b=button+14;
	Link(button,b);
	b->bbox = r_zoom;
	b->btn[0] = b->pbtn[0] = i_zoom_in;		// never used
	b->btn[1] = b->pbtn[1] = i_zoom_in;
	b->btn[2] = b->pbtn[2] = i_zoom_out;
	b->icn[0] = b->icn[1] =  b->icn[2] = nil;
	b->nxt_sta[0]=0;
	b->nxt_sta[1]=2;
	b->nxt_sta[2]=1;
	b->state=2;
	b->id=14;


// create and open window
yy:	tapeRect.right = tapeRect.left + (tape_rect.right-tape_rect.left);
	tapeRect.bottom = tapeRect.top + (tape_rect.bottom-tape_rect.top)*(1+tape[1].visible);
	
	tape_window = NewCWindow (nil, &tapeRect, localized("\pTape recorder","\pKassettenrekorder"), true, tape_wdef, (WindowPtr)-1, true, 0);
	if (!tape_window)  abort_oomem("\pOpenTapeWindow()"); 

xx:	SetWRefCon(tape_window,iTape);
	SetPort ( tape_window );
	GetFNum ( "\pGeneva", &i );
	TextFont ( i );
	TextSize ( textsize );
	TextMode ( srcOr );
	RGBBackColor (&bkgr_color );
	CheckItem ( windowsmenu, iTape, true );
}


// -----	wait delay -------------------------------------------------------------------------
//		-> for key clicks etc.
Wait ( long delay )
{	
	delay += TickCount();			// final tick number
	update_highres();
	
	while ( ((long)TickCount()-delay)<0 )
	{	RefreshSpecciWindow();
		if (mbar_present) PollEvents ( /*updateMask*/everyEvent, delay-TickCount(), nil );
	}
}


// -----	wait delay -------------------------------------------------------------------------
//		animation of cassette icon
static Boolean delay ( tap *tap, long delay )
{
	delay += TickCount();			// final tick number
	update_highres();
	
	while ( (long)TickCount()-delay<0 )
	{	animate(tap);
		UpdateKeyboard();
		if ( ~keyboard.byte[7] & 1 ) return error;
	}
	return ok;
}


// -----	wait delay for sync border ------------------------------------------------------
//		animation of cassette icon
//		test for break
static Boolean sync_delay ( tap *tap )
{
	bordercolor = syncborder;
	return delay ( tap, SYNC_DELAY );
}


// -----	wait delay for data border ------------------------------------------------------
//		animation of cassette icon
//		test for break
static Boolean data_delay ( tap *tap, long len )
{
	bordercolor = databorder;
	return delay ( tap, len*60/BPS );
}

		
// -----	Save (insert) block at actual tape position ---------------------------------
//		tape must be inserted, running and recording!
//		returns error code (ok, 'ubrk' etc.)
static long save_block ( tap *tap, Short addr, Short len, Char type )
{	Handle bu=nil; Char *p,*e; long bytes; long result;
	Char crc;

	err = ok;
	result='ubrk';
	WaitCursor();
	if ( delay ( tap, SYNC_DELAY/4 ) ) 		goto xx;
	if ( sync_delay ( tap ) )				goto xx;
	if ( data_delay ( tap, len ) )			goto xx;

// extend file length before anything else -> ensure, file is writable and will fit on disk
// all other file error checkings are for security only, all positioning and writing should succeed!
	result = 'wrer';
	err = SetEOF(tap->id,tap->len+len+4);				if (err) goto xx;

	if (tap->pos<tap->len)	// insert ?
	{	
		bytes = tap->len-tap->pos;
		bu = TempNewHandle(bytes,&err); 				if (err) goto xx;
		HLock(bu);
		err = SetFPos(tap->id,fsFromStart,tap->pos);		if (err) goto xx;
		err = File2Buffer(tap->id,*bu,bytes);			if (err) goto xx;
		err = SetFPos(tap->id,fsFromStart,tap->pos+len+4);if (err) goto xx;
		err = Buffer2File(*bu,tap->id,bytes);			if (err) goto xx;
		DisposeHandle(bu); bu=nil;
	}

	bytes = len+4;
	bu = TempNewHandle(bytes,&err);					if (err) goto xx;
	HLock(bu);
	
	p = (Char*)*bu; 
	*p++ = len+2;
	*p++ = (len+2)>>8;
	*p++ = type;	
	
	Core2Buffer(addr,p,len);		
	
	e = p+len;	crc = type; while (p<e) crc ^= *p++; *e=crc;
	
	err = SetFPos(tap->id,fsFromStart,tap->pos);			if (err) goto xx;
	err = Buffer2File(*bu,tap->id,bytes);				if (err) goto xx;
	tap->pos += bytes;
	tap->len += bytes;

	result = ok;

xx:	if (bu) DisposeHandle(bu); bu=nil;
	if (err)
	{	SetEOF (tap->id,tap->len);
		PrintNumberAlert(CStr(localized("Write Error: ","Schreibfehler: ")),err);
	}

	draw_blockinfo ( tap );
	ArrowCursor();
	return result;
}


// -----	Load block from current tape position --------------------------------------
//		tape must be inserted, running and not recording!
//		returns error code (ok, 'ubrk', 'wtyp' etc.)
static long load_block ( tap *tap, Short addr, Short len, Char type)
{	Handle bu=nil; long bytes; long result;
	Char *p,*e,crc;
	
	err = ok;
	result='ubrk';
	WaitCursor();
	if ( delay ( tap, SYNC_DELAY/4 ) ) 		goto xx;
	if ( sync_delay ( tap ) )				goto xx;

	result='rder';
	bytes = block_len(tap); 				if (err) goto xx;
	bu = TempNewHandle	( bytes, &err);		if (err) goto xx;
	HLock (bu);
	err = File2Buffer(tap->id,*bu,bytes);	if (err) goto xx;
	tap->pos += 2+bytes;

	p = (Char*)*bu;
	result='wtyp';	if ( *p++ != type )		goto xx;
	
	result='ubrk';
	if ( data_delay(tap,len) )				goto xx;

	Buffer2Core ( p, addr, min(len,bytes-1) );
	result = 'rder';	if (bytes<len+2)		goto xx;
	
	crc = type;
	e = p+len;
	while (p<=e) crc ^= *p++;
	result = 'wcrc'; if (crc) goto xx;
	
	result = ok;

xx:	if (bu) DisposeHandle(bu); bu=nil;
	if (err) PrintNumberAlert(CStr(localized("Read Error: ","Lesefehler: ")),err);
	draw_blockinfo ( tap );
	ArrowCursor();
	return result;
}


// -----	Verify block at actual tape position --------------------------------------
//		tape must be inserted, running and not recording!
//		returns error code (ok, 'ubrk', 'wtyp' etc.)
static long verify_block ( tap *tap, Short addr, Short len, Char type)
{	Handle bu=nil; long bytes; long result;
	Char *p,*e,crc;
	
	err = ok;
	result='ubrk';
	WaitCursor();
	if ( delay ( tap, SYNC_DELAY/4 ) ) 		goto xx;
	if ( sync_delay ( tap ) )				goto xx;

	result='rder';
	bytes = block_len(tap); 				if (err) goto xx;
	bu = TempNewHandle	( bytes, &err);		if (err) goto xx;
	HLock (bu);
	err = File2Buffer(tap->id,*bu,bytes);	if (err) goto xx;
	tap->pos += 2+bytes;

	p = (Char*)*bu;
	result='wtyp';	if ( *p++ != type )		goto xx;
	
	result='ubrk';
	if ( data_delay(tap,len) )				goto xx;

	result = 'rder';	if (bytes<len+2)		goto xx;
	
	crc = type;
	e = p+len;
	while (p<e) 
	{	if (Peek(addr++)!=*p) goto xx;
		crc ^= *p++;
	}
	result = 'wcrc'; if (crc!=*p) goto xx;
	
	result = ok;

xx:	if (bu) DisposeHandle(bu); bu=nil;
	if (err) PrintNumberAlert(CStr(localized("Read Error: ","Lesefehler: ")),err);
	draw_blockinfo ( tap );
	ArrowCursor();
	return result;
}


// -----	Copy block from tape A to tape B at actual positions -------------------------
//		tapes must be inserted, running and not recording!
//		returns error code (ok, 'ubrk' etc.)
copy_block ( );
copy_block ( )
{	Handle bu=nil; Handle b2=nil; long bytes,n;
	Boolean writing=false;

	if ( !tape[0].id || !tape[0].running || tape->recording || tape->pos>=tape->len)	return;
	if ( !tape[1].id || !tape[1].running )									return;


// read source
	bytes = 2 + block_len(tape); 							if (err) goto xx;
	err = SetFPos(tape->id,fsFromStart,tape->pos);			if (err) goto xx;

	bu = TempNewHandle	( bytes, &err);						if (err) goto xx;
	HLock (bu);

	err = File2Buffer(tape->id,*bu,bytes);					if (err) goto xx;
	tape->pos += bytes;

	draw_blockinfo(tape);
	delay(tape,SPOOL_DELAY);

	writing = true;
// extend destination file length before anything else -> ensure, file is writable and will fit on disk
// all other file error checkings are for security only, all positioning and writing should succeed!
	err = SetEOF(tape[1].id,tape[1].len+bytes);				if (err) goto xx;

	if (tape[1].pos<tape[1].len)	// insert ?
	{
		n = tape[1].len-tape[1].pos;
		b2 = TempNewHandle(n,&err); 						if (err) goto xx;
		HLock(b2);
		err = SetFPos(tape[1].id,fsFromStart,tape[1].pos);		if (err) goto xx;
		err = File2Buffer(tape[1].id,*b2,n);					if (err) goto xx;
		err = SetFPos(tape[1].id,fsFromStart,tape[1].pos+bytes);if (err) goto xx;
		err = Buffer2File(*b2,tape[1].id,n);					if (err) goto xx;
	}


// write to destination file
	err = SetFPos(tape[1].id,fsFromStart,tape[1].pos);			if (err) goto xx;
	err = Buffer2File(*bu,tape[1].id,bytes);					if (err) goto xx;

	draw_blockinfo(tape+1);
	delay(tape+1,SPOOL_DELAY);

	tape[1].pos += bytes;
	tape[1].len += bytes;

	draw_blockinfo(tape+1);

xx:	if (bu) DisposeHandle(bu); bu=nil;
	if (b2) DisposeHandle(b2); b2=nil;
	if (err)
	{	if (writing)
		{	SetEOF (tape[1].id,tape[1].len);
			PrintNumberAlert(CStr(localized("Write Error: ","Schreibfehler: ")),err);
			eject_tape(tape+1);
		}
		else
		{	PrintNumberAlert(CStr(localized("Read Error: ","Lesefehler: ")),err);
			eject_tape(tape);
		}
	}
}




// -----	Erase (delete) block at actual tape position --------------------------------
//		tape must be inserted and running!
erase_block ( tap *tap );
erase_block ( tap *tap )
{	long len; Handle bu=nil; long bytes;

	if ( !tap->id || !tap->running || tap->pos>=tap->len ) return;

	if (!(event.modifiers&(optionKey|shiftKey)))
		if (do_confirm ( localized("\pDo you really want to erase this block?","\pSoll dieser Block wirklich gelscht werden?"),
					 localized("\p(shift-clicking bypasses this requester)","\p(Shift-Klicken umgeht diese Abfrage)")) !=1 ) return;

	len = block_len(tap); if (err) return;

	bytes = tap->len - (tap->pos + 2 + len);
	bu = TempNewHandle ( bytes, &err ); 		// allocate buffer
	if (!bu) { PrintAlert(CStr(localized("Out of memory!","Speicherberlauf"))); return; };
	HLock(bu);
	
	err = SetFPos ( tap->id, fsFromStart, tap->pos+2+len );
	if (!err) err = File2Buffer ( tap->id, *bu, bytes );
	if (!err) err = SetFPos ( tap->id, fsFromStart, tap->pos );
	if (err) 
	{	PrintNumberAlert ( CStr(localized("Read error:","Lesefehler:")), err ); goto xx; 
	}
	
	err = Buffer2File ( *bu, tap->id, bytes );
	if (!err) { tap->len = tap->pos+bytes; err = SetEOF ( tap->id, tap->len ); }
	if (err)
	{	PrintNumberAlert ( CStr(localized("Write error:","Schreibfehler:")), err ); 
	}

xx:	if (err) eject_tape(tap); else draw_blockinfo(tap);
	if (bu) DisposeHandle(bu);
}


// ----- Step to next data block on tape --------------------------------------
next_block ( tap *tap );
next_block ( tap *tap )
{	Short len;
	
	if ( !tap->id ) return;							// no tape loaded

	WaitCursor();
	delay(tap,SPOOL_DELAY);
	ArrowCursor();
	
	if (tap->pos>=tap->len) return;					// already at end
	len = block_len(tap); if (err) { eject_tape(tap); return; }

	tap->pos += 2 + len;
	draw_blockinfo(tap);
}


// ----- Step to previous data block on tape ------------------------------------
static prev_block ( tap *tap )
{	long ap,lp;
	Short len;

	if (!tap->id ) return;

	WaitCursor();
	delay(tap,SPOOL_DELAY);
	ArrowCursor();
	
	ap = tap->pos;
	tap->pos = lp = 0;
	
	while ( tap->pos < ap )
	{	len = block_len(tap); if (err) { eject_tape(tap); return; }
		lp = tap->pos;
		tap->pos += 2 + len;
	};

	tap->pos = lp;
	draw_blockinfo(tap);
}


// -----	Rewind tape --------------------------------------------------------------
static first_block ( tap *tap )
{
	if (!tap->id ) return;

	WaitCursor();
	delay ( tap, tap->pos>0 ? SPOOL_DELAY*2 : SPOOL_DELAY );
	ArrowCursor();
	
	tap->pos = 0;
	draw_blockinfo(tap);
}


// -----	Wind tape to end ------------------------------------------------------
static last_block ( tap *tap )
{
	if (!tap->id ) return;

	WaitCursor();
	delay ( tap, tap->pos<tap->len ? SPOOL_DELAY*2 : SPOOL_DELAY );
	ArrowCursor();
	
	tap->pos = tap->len;
	draw_blockinfo(tap);
}


// -----	zoom in / zoom out ------------------------------------------------------------
TapeZoom()
{
	if (!tape_window) return;
	SetPort ( tape_window );

	if (tape[1].visible)
	{
		tape[1].visible=false;
		SizeWindow (tape_window, WIDTH, HEIGHT, false);
		Set_Button(button+14,2);
		if (tape[1].selected)
		{	tape[0].selected=true;
			tape[1].selected=false;
			draw_frame(tape);
		}
	}
	else
	{
		tape[1].visible=true;
		SizeWindow (tape_window, WIDTH, HEIGHT*2, true);
		Set_Button(button+14,2);
	}
}


// -----	create and insert an empty tape ----------------------------------------------
empty_tape (tap *tap);
empty_tape (tap *tap)
{	StandardFileReply reply;
	FInfo fndrInfo;
	Boolean fndrInfoValid = false;
		
	if (tap->id) eject_tape(tap);
	StandardPutFile  (  localized ( "\pName the empty tape:","\pName fr die leere Kassette:" ), "\pEmpty.tape", &reply );
	if (!reply.sfGood) return;
	
	if (reply.sfReplacing)
	{
		err = FSpGetFInfo (&reply.sfFile, &fndrInfo);
		fndrInfo.fdType='Tape';
		fndrInfo.fdCreator='ZXSP';
		if (!err) fndrInfoValid = true;
		if (!err) err = FSpDelete (&reply.sfFile);
	}

	err = FSpCreate ( &reply.sfFile, 'ZXSP', 'Tape', 0 );

	if (!err && fndrInfoValid) err = FSpSetFInfo (&reply.sfFile, &fndrInfo);

	if (err)	PrintNumberAlert(CStr(localized("Could not create the file: ","Die Datei konnte nicht angelegt werden: ")),err);
	else		OpenTape ( tap, reply.sfFile );
}


// -----	Mouse click in tape recorder window -------------------------------------------
ClickInTapeWindow ()
{	Point p; tap *tap; btn *b;

	if (!tape_window) return;
	SetPort ( tape_window );

	p = event.where;

	if ( PtInRect ( p,&tape_rect ) != tape->selected )
	{	tape[0].selected ^= 1;
		tape[1].selected ^= 1;
		if (tape->visible) draw_frame(tape);
		if (tape[1].visible) draw_frame(tape+1);
	}

	tap = tape + !tape->selected;

	if ( b = Find_Button(button,p) )
	{	if ( Handle_Button(b) )
		{
			switch (b->id)
			{
			case 0:									
			case 7:	if (b->state==2)	erase_block(tap);			// Tape A/B: erase block
					else				empty_tape(tap);	break;	// Tape A/B: insert empty tape
			case 1:	tap->recording = b->state==2;			return;	// Tape A: record on/off
			case 8:	copy_block();						break;	// Tape B: copy
			case 2:											// Tape A: rewind
			case 9:	first_block(tap);					break;	// Tape B: rewind	
			case 3:											// Tape A: prev block
			case 10:	prev_block(tap);					break;	// Tape B: prev block
			case 4:											// Tape A: start/stop
			case 11:	tap->running = b->state==1;			return;	// Tape B: start/stop
			case 5:											// Tape A: next block
			case 12:	next_block(tap);					break;	// Tape B: next block
			case 6:											// Tape A: end of tape
			case 13:	last_block(tap);					break;	// Tape B: end of tape
			case 14:	TapeZoom();						return;	// zoom in/out
			}
		
		// reset button icon for those, who breaked (and not returned)
			if (b->state>=2)
			{	SetPort(tape_window);
				Set_Button ( b, 3-b->state );
			}
		}
		return;
	}
	
	#define In_Box(b,p) PtInRect(p,&b)
	
	if ( In_Box ( r_cassette, p ) )	// insert/eject tape
	{	if (tap->id) eject_tape(tap); else insert_tape(tap);
		return;
	}

	if ( In_Box ( r_blockinfo, p ) )	// popup menu of blocks
	{	return;
	}

	p.v -= HEIGHT;	

	if ( In_Box ( r_cassette, p ) )	// insert/eject tape
	{	if (tap->id) eject_tape(tap); else insert_tape(tap);
		return;
	}

	if ( In_Box ( r_blockinfo, p ) )	// popup menu of blocks
	{	return;
	}
	
	p.v += HEIGHT;	

}


long LoadBlock ( tap *tap, Char typ, Short addr, Short len )
{
	OpenTapeWindow();
	if (!tap) tap=tape+!tape->selected;

	if ( !tap->id || !tap->running || tap->recording || tap->pos>=tap->len ) 
	{	SelectWindow(tape_window);
		UpdateKeyboard();
		return (keyboard.byte[7]&1) ? 'wait' : 'ubrk';
	}

	return load_block(tap,addr,len,typ);
}


long VerifyBlock ( tap *tap, Char typ, Short addr, Short len )
{
	OpenTapeWindow();
	if (!tap) tap=tape+!tape->selected;

	if ( !tap->id || !tap->running || tap->recording || tap->pos>=tap->len ) 
	{	SelectWindow(tape_window);
		UpdateKeyboard();
		return (keyboard.byte[7]&1) ? 'wait' : 'ubrk';
	}

	return verify_block(tap,addr,len,typ);
}


long SaveBlock ( tap *tap, Char typ, Short addr, Short len )
{
	OpenTapeWindow();
	if (!tap) tap=tape+!tape->selected;

	if ( !tap->id || !tap->running || !tap->recording ) 
	{	SelectWindow(tape_window);
		UpdateKeyboard();
		return (keyboard.byte[7]&1) ? 'wait' : 'ubrk';
	}

	return save_block(tap,addr,len,typ);
}


/* -----	Save Core as .tap file ------------------------------------------------------------
		note:	a basic loader (header+data) and a binary block (data only) are appended to the open file.
			this routine puts a little loader on the stack of the saved snapshot and this may result in a corrupt snapshot.
			the risk is especially high if the interrupts are detected to be disabled.
		note: assumes ZX 48k (to be improved!)
*/
OSErr SaveTAPE ( short fileID )	
{
	fileID;
/*	OSErr err;
	Short old_pc;
	short old_tape_ID;
	Char bu[120];
	Char *p;

	Char prg_head[17] = 
	{	0,  							// type of data (program/code/etc.)
		'T','H','E','_','L','O','A','D','E','R', 	// name
		0x6f,0,  						// size of data block
		10,0,  						// auto-start line number
		0x6f,0 						// size of program without variables
	};

	Char prg_data[111] = 
	{	0x00, 0x0A, 0x31, 0x00, 		// line number and length
		0xF5, 0xC0, 0x28, 0xBE, 		// PRINT USR '(' PEEK
		0x32, 0x33, 0x36, 0x33, 0x35, 	// 23635 (ASCII)
		0x0E, 0, 0, 0x53, 0x5C, 0, 		// 23635 (internal notation)
		0x2B, 						// '+'
		0x32, 0x35, 0x36, 				// 256 (ASCII)
		0x0E, 0, 0, 0x00, 0x01, 0, 		// 256 (internal notation)
		0x2A, 0xBE, 					// '*' PEEK
		0x32, 0x33, 0x36, 0x33, 0x36, 	// 23636 (ASCII)
		0x0E, 0, 0, 0x54, 0x5C, 0, 		// 23636 (internal notation)
		0x2B, 						// '+'
		0x36, 0x36, 					// 66 (ASCII exact value not yet calculated!)
		0x0E, 0, 0, 0x42, 0x00, 0, 		// 66 (internal notation)
		0x29, 0x0D, 					// ')' 'return'
		0x00, 0x14, 0x36, 0x00, 		// line number and length
		0xE4 ,  '"',					// 	DATA '"'
		// now 50 bytes mcode follow:
		LD_BC_NN, RC2, RB2,			//	LD	BC,REG_BC2
		LD_DE_NN, RE2, RD2,			//	LD	DE,REG_DE2
		LD_HL_NN, RL2, RH2,			//	LD	HL,REG_DE2
		PFX_IY, LD_HL_NN, YL, YH,		//	LD	IY,REG_IY
		EXX, 						// 	EXX
		LD_A_N, RI,					//	LD	A,REG_I
		PFX_ED, LD_I_A,				//	LD	I,A
		LD_A_N, RR,					//	LD	A,REG_R
		PFX_ED, LD_R_A,				//	LD	R,A
		LD_SP_NN, (SP-10), (SP-10)>>8,	//	LD	SP,REG_SP - 10
		LD_A_N, 0xff,					//	LD	A,$FF
		LD_DE_NN, 0x00, 0xc0,			//	LD	DE,0XC000
		PFX_IX, LD_HL_NN, 0x00, 0x40,	//	LD	IX,0X4000
		SCF,							//	SCF				; -> LOAD (NOT VERIFY)
		INC_D,						//	INC	D			; Zero loeschen
		EX_AF_AF,					//	EX	AF,AF'		; und Flags retten
		DEC_D,						//	DEC	D			; D regen.
		DI,							//	DI
		LD_A_N, 0x0f,					//	LD	A,0FH		; Border color WHITE
		OUTA, 0xfe,					//	OUT	(0FEH),A
		JP, 0x62, 0x05,				//	JP	$0562
		NOP,NOP,NOP,NOP,				// 	fill up to 50 bytes
		'"',	0x0d 					//	'"' 'return'
	};

	Char boot_data[28] = 
	{	LD_BC_NN,RC,RB,				//	LD	BC,REG_BC
		LD_DE_NN,RE,RD,				//	LD	DE,REG_DE
		LD_HL_NN,RL,RH,				//	LD	HL,REG_HL
		POP_AF,						//	POP	AF
		EX_AF_AF,					//	EX	AF,AF'
		POP_AF,						//	POP	AF
		PFX_IX,POP_HL,				//	POP	IX
		PFX_ED,IM_1,					//	IM0/IM1/IM2
		EI,							//	EI/NOP
		RET,							//	RET
		SP-10-18,(SP-10-18)>>8,		//	DW	REG_SP - 10 - SIZE_OF_BOOTSTRAP_CODE_SO_FAR
		RF,RA,						//	DW	REG_AF
		RF2,RA2,						//	DW	REG_AF2
		XL,XH,						//	DW	REG_IX
		PCL,PCH						//	DW	REG_PC
	};


	old_tape_ID=tape_ID; tape_ID=fileID;
	old_pc = PC;
	Core2Buffer ( 0x4000, bu, 120 );

// save program header
	Buffer2Core ( &prg_head, 0x4000,17 );
	err=SaveBlock(0,0x4000,17);

// save program body
	Buffer2Core ( &prg_data, 0x4000, 111 );
	err=SaveBlock(0xff,0x4000,111);

// create boot loader & save core
	Buffer2Core ( bu, 0x4000, 120 );
	SP-=28;
	Buffer2Core ( boot_data, SP, 28 );
	err=SaveBlock(0xff,	0x4000,0xc000);
*/
}


