
/*	ZX Spectrum window handler

***	Copyright:	1994-1996 Gnter Woigk
  				mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	01.Apr.94:	Started work on this file					KIO !
	07.Apr.95:	8 bit indexed color support added			KIO !
	24.Apr.95:	offscreen drawing added					KIO !
	25.Jun.95:	screensize 3 and 4; improved b&w			KIO !
*/


#include	<resources.h>
#include	<palettes.h>
#include	<files.h>
#include	<quickdraw.h>
#include	<QDoffscreen.h>
#include	<events.h>
#include	<Windows.h>
#include	<string.h>

#include	"kio.h"
#include	"terminal.h"
#include	"z80.h"
#include	"preferences.h"
#include	"display.h"
#include	"application.h"
#include	"messages.h"
#include	"z80 main.h"
#include	"menue.h"
#include	"strings.h"
#include	"movie.h"

#define MemCopy(a,b,c) memcpy(b,a,c)


typedef struct { Long a,b; }	Long2;
typedef struct { Long a,b,c; }	Long3;
typedef struct { Long a,b,c,d; }	Long4;

typedef struct { Short a,b; }	Short2;
typedef struct { Short a,b,c; }	Short3;
typedef struct { Short a,b,c,d;}	Short4;

typedef struct { Char a,b; }	Char2;
typedef struct { Char a,b,c; }	Char3;
typedef struct { Char a,b,c,d; }	Char4;

typedef struct { Char x[32*3]; }	LineB3;	// line length: triple sized b&w display
typedef struct { Char x[32*4]; }	LineB4;	// line length: quad sized b&w display
typedef struct { Char x[0x100]; }	Line1;
typedef struct { Char x[0x200]; }	Line2;
typedef struct { Char x[0x300]; }	Line3;
typedef struct { Char x[0x400]; }	Line4;
typedef struct { Char x[0x600]; }	Line6;
typedef struct { Char x[0x800]; }	Line8;
typedef struct { Char x[0x900]; }	Line9;
typedef struct { Char x[0xC00]; }	Line12;
typedef struct { Char x[0x1000]; }	Line16;


// -----	Handling Black Out Screen ---------------------------------------------------
extern	Boolean	blacked_out	= false;
extern	Boolean	mbar_present	= true;

RgnHandle mbarRgn				= nil;	// menu bar (in global coord.)
RgnHandle grayRgn				= nil;	// desktop (in global coord.)

short	mbarHeight			= 0;


// -----	The Spectrum window ------------------------------------------------------
WindowPtr	zxwindow			= nil;
GWorldPtr 	offscreenGWorldPtr	= nil;

#define	maxmon			4
GDHandle	monitor[maxmon+1];		// list of GDevices the screenRect intersects

short	bordercolor		= 0;		// which color should the border be?
short	oldbordercolor		= 99;	// which color is the border actually?
long*	brdrlog			= nil;	// border log buffer in highres mode
long*	brdrlogptr			= nil;	// write pointer used by DoOutput()
long*	brdrlogend		= nil;	// buffer end
#define	brdrlogsize		1000
#define	brdrlines			1000	// 312
char		oldbrdr[brdrlines];

short	scrnsize			= 1;		// 1 ... 4
short	blkborder			= 0;		// additional black border around specci border
short	hborder				= 16;		// 16*n; border left & right of writable interior screen
short	vborder				= 16;		// 16*n; border above & below of writable interior screen
short	force_offscreen 	= false;	// menu item: force usage of offscreen gworld ?
short	require_offscreen	= false;	// Specci screen across multiple monitors or not supported color depth?
short	last_update_method	= 0;
Char*	video_page			= nil;	// address of speccy's video RAM

RgnHandle	tempRgn			= nil;	// scratch
RgnHandle	borderRgn			= nil;	// border (local coord.)
RgnHandle	screenRgn			= nil;	// writable interior screen (local coord.)
RgnHandle	blackRgn			= nil;	// remaining window around border (if rest of screen blacked out)
Rect		screenRect;			
Rect		specciRect[5];				// default rects for screensize 1 ... 4   ((0 is a lost entry))


// -----	attribute -> color conversion -----------------------------------
RGBColor	rgb[16];				// rgb colors for the 16 ZX Spectrum colors

Char		c8[16];				// 8 bit indexed colors or grey scale screen bytes
Short	c16[16];				// 16 bit rgb color screen words
Long		c32[16];				// 32 bit rgb color screen longwords

Char2	d8[16];				// 8 bit ic or grey	(scrnsize=2)
Short2	d16[16];				// 16 bit rgb color	(scrnsize == 2)
Long2	d32[16];				// 32 bit rgb color	(scrnsize == 2)

Char3	t8[16];				// 8 bit ic or grey	(scrnsize=3)
Short3	t16[16];				// 16 bit rgb color	(scrnsize == 3)
Long3	t32[16];				// 32 bit rgb color	(scrnsize == 3)

Char4	q8[16];				// 8 bit ic or grey	(scrnsize=4)
Short4	q16[16];				// 16 bit rgb color	(scrnsize == 4)
Long4	q32[16];				// 32 bit rgb color	(scrnsize == 4)

long		pap[256];				// convert attr -> paper color index
long		ink[256];				// convert attr -> ink color index
Char		pink[256];			// swap paper <-> ink color if flashing

Char		bwn[256];			// b&w: attr_byte -> negative pixel_byte ?
Char		bwm[256];			// b&w: attr_byte -> ignore pixel_byte if ink==paper


// -----	screen refresh buffers ------------------------------------------------

long		pixbits[24*8];			// bitflags flagging differences
long		n2n[24*8];			// convert: display line number <-> pixmap line index

Char		*old_screen;
Char		*old_pixels;			// what IS displayed
Char		*old_attr;
Char		*old_end;

Char		*new_screen;
Char		*new_pixels;			// what SHOULD be displayed
Char		*new_attr;
Char		*new_end;

long		rowbytes;				//  of specci window (immediate) or offscreen bitmap
Char*	screenbase;			//  of specci window (immediate) or offscreen bitmap


// **************************************************************** 


// -----	find differences between old_pixels and new_pixels and flag them in pixbits[]
static update_pixflags()
{	Char *p1,*p2;
	long i,y,n=0;

	p1 = old_pixels;
	p2 = new_pixels;

	for (y=0;y<24*8;y++)				// scan pixels
	{	for (i=32;i--;) n += n+(*p1++!=*p2++);
		pixbits[n2n[y]] = n;
	}

	if (run_highres)
	{	for (y=0;y<24*8;y++)			// scan attributes
		{	for (i=32;i--;) n += n+(*p1++!=*p2++);
			pixbits[y] |= n;
		}
	}
	else
	{	for (y=0;y<24*8;y+=8)			// scan attributes
		{	for (i=32;i--;) n += n+(*p1++!=*p2++);
			for (i=y+7;i>=y;i--) pixbits[i] |= n;
		}
	}
}


InitDisplay()
{	Short c,i;

	for (i=0;i<24*8;i++)  n2n[i] = ((i&0x07)<<3) + ((i&0x38)>>3) + (i&0xC0);

// allocate handles etc.
	tempRgn		= NewRgn();
	borderRgn		= NewRgn();
	screenRgn		= NewRgn();
	blackRgn		= NewRgn();
	mbarRgn		= NewRgn();
	old_screen 	= (Char*)NewPtr(24*16*32);
	new_screen 	= (Char*)NewPtr(24*16*32);
	if ( !(old_screen&&new_screen&&tempRgn&&borderRgn&&screenRgn&&blackRgn&&mbarRgn) ) abort_oomem("\pInitDisplay()");

// -----	initialize color conversion tables ------------------------------------
//		window dependant indexed color tables c8[] and d8[] are recalculated in RefreshSpecciWindow()
	for ( i=0; i<16; i++ ) 
	{	c=0xdddd; if (i>=8) c=0xffff;			// 16 * rgb colors
		rgb[i].blue	  = (i&1) ? c : 0;
		rgb[i].red 	  = (i&2) ? c : 0;
		rgb[i].green = (i&4) ? c : 0;

		q8[i].a = q8[i].b = q8[i].c = q8[i].d =			// 16 * 8-bit indexed color quad_chars		(quad size window)
		t8[i].a = t8[i].b = t8[i].c = 				// 16 * 8-bit indexed color triple_chars		(quad size window)
		d8[i].a = d8[i].b = 						// 16 * 8-bit indexed color double_chars		(quad size window)
		c8[i] = Color2Index(rgb+i);				// 16 * 8-bit indexed color bytes			(single size window)
		
		q16[i].a = q16[i].b = q16[i].c = q16[i].d =		// 16 * 16-bit indexed color quad_word		(quad size window)
		t16[i].a = t16[i].b = t16[i].c = 				// 16 * 16-bit indexed color triple_word		(quad size window)
		d16[i].a = d16[i].b = 					// 16 * 16-bit indexed color double_word	(quad size window)
		c16[i] = ((rgb[i].red&0xf800)>>1) + ((rgb[i].green&0xf800)>>6) + (rgb[i].blue>>11);
		
		q32[i].a = q32[i].b = q32[i].c = q32[i].d =		// 16 * 32-bit indexed color quad_long		(quad size window)
		t32[i].a = t32[i].b = t32[i].c = 				// 16 * 32-bit indexed color triple_long		(quad size window)
		d32[i].a = d32[i].b = 					// 16 * 32-bit indexed color double_long		(quad size window)
		c32[i] = ((rgb[i].red&0xff00)<<8) + (rgb[i].green&0xff00) + (rgb[i].blue>>8);
	};
	
	for ( i=0; i<256; i++ )
	{	ink[i]	=  (i&7)+((i>>3)&8);			// convert attribute byte -> color index of ink
		pap[i]	=  (i>>3)&0xf;					// convert attribute byte -> color index of paper
		pink[i]	=  (i&0xc0)+((i>>3)&7)+((i&7)<<3);// swap paper<->ink in attribute byte

		bwm[i]	=  (i&7) == ((i>>3)&7)  ?  0 : 0xff;	// mask to ignore pixel bits if paper == ink
		bwn[i] 	=  bwm[i]  ?  (  (i&7) < ((i>>3)&7)  ?  0 : 0xff  )  :  (  (i&7) > 3 ? 0 : 0xff  );	// mask to invert pixel bits
	};
}


// -----	Accelerated refresh routine for 1-bit screens and scrnsize=1 ------------------------
//		the bitmap must start on byte boundary without a bit-offset!
static RefreshSingle1 ()
{	long		x,y; 
	Char		*attr, *pix;
	Char		*pixel;

	for ( y=0; y<24*8; y++ )
	{	if ( true || pixbits[y] )
		{	attr	= new_attr+32* ( run_highres ? y : (y>>3) );
			pix	= new_pixels+32*n2n[y];
			pixel	= screenbase + y*rowbytes;
			for ( x=pixbits[y]; x; x<<=1 )
			{	*pixel++ = ( *pix++ & bwm[*attr] ) ^ bwn[*attr]; attr++;
			};
		};
	};
}


// -----	Accelerated refresh routine for 1-bit screens and scrnsize=2 ------------------------
//		the bitmap must start on byte boundary without a bit-offset!
static RefreshDouble1 ()
{	long			x,y; 
	Char			*attr, *pix, c;
	Short		*pixel,*paxel;
	static Short	Lnibble[]={0,0x0003,0x000c,0x000f,0x0030,0x0033,0x003c,0x003f,0x00c0,0x00c3,0x00cc,0x00cf,0x00f0,0x00f3,0x00fc,0x00ff};
	static Short	Hnibble[]={0,0x0300,0x0c00,0x0f00,0x3000,0x3300,0x3c00,0x3f00,0xc000,0xc300,0xcc00,0xcf00,0xf000,0xf300,0xfc00,0xff00};

	for ( y=0; y<24*8; y++ )
	{	if ( true || pixbits[y] )
		{	attr	= new_attr+32* ( run_highres ? y : (y>>3) );
			pix	= new_pixels+32*n2n[y];
			pixel	= (Short*)(screenbase + y*2*rowbytes);
			paxel= (Short*)((long)pixel+rowbytes);
			for ( x=pixbits[y]; x; x<<=1 )
			{	if ( x>=0 )  {  pixel++; paxel++; attr++; pix++; continue;  };
				c = ( *pix++ & bwm[*attr] ) ^ bwn[*attr]; attr++;	
				*pixel++ = *paxel++ = Hnibble[c>>4] + Lnibble[c&0x0f];
			};
		};
	};
}


// -----	Accelerated refresh routine for 1-bit screens and scrnsize=3 ------------------------
//		the bitmap must start on byte boundary without a bit-offset!
static RefreshTriple1 ()
{	long			x,y; 
	Char			*attr, *pix, c;
	Char			*pixel,*p0;
	static Char	tri1[] = { 0x00, 0x07, 0x38, 0x3F, 0xC0, 0xC7, 0xF8, 0xFF };
	static Char	tri2[] = { 0x00, 0x01, 0x0E, 0x0F, 0x70, 0x71, 0x7E, 0x7F, 0x80, 0x81, 0x8E, 0x8F, 0xF0, 0xF1, 0xFE, 0xFF };
	static Char	tri3[] = { 0x00, 0x03, 0x1C, 0x1F, 0xE0, 0xE3, 0xFC, 0xFF };

	for ( y=0; y<24*8; y++ )
	{	if ( true || pixbits[y] )
		{	attr	= new_attr+32* ( run_highres ? y : (y>>3) );
			pix	= new_pixels+32*n2n[y];
			p0 = pixel	= screenbase + y*3*rowbytes;
			for ( x=pixbits[y]; x; x<<=1 )
			{	if ( x>=0 )  {  pixel+=3; attr++; pix++; continue;  };
				c = ( *pix++ & bwm[*attr] ) ^ bwn[*attr]; attr++;	
				*pixel++ = tri3[c>>5];
				*pixel++ = tri2[(c>>2)&0x0F];
				*pixel++ = tri1[c&0x07];
			};
			*(LineB3*)(p0+rowbytes) = *(LineB3*)(p0+rowbytes*2) = *(LineB3*)p0;
		};
	};
}


static RefreshQuad1()
{	long		x,y; 
	Char		c, *attr, *pix;
	long		*pixel, *p0;
	static long Lnibble[]={0,0x000F,0x00F0,0x00FF,0x0F00,0x0F0F,0x0FF0,0x0FFF,
					0x0F000,0x0F00F,0x0F0F0,0x0F0FF,0x0FF00,0x0FF0F,0x0FFF0,0x0FFFF};
	static long	Hnibble[]={0,0x000F0000,0x00F00000,0x00FF0000,0x0F000000,0x0F0F0000,0x0FF00000,0x0FFF0000,
					0xF0000000,0xF00F0000,0xF0F00000,0xF0FF0000,0xFF000000,0xFF0F0000,0xFFF00000,0xFFFF0000};

	for ( y=0; y<24*8; y++ )
	{	if ( pixbits[y] )
		{	attr	= new_attr+32* ( run_highres ? y : (y>>3) );
			pix	= new_pixels+32*n2n[y];
			p0 = pixel	= (long*)(screenbase + y*4*rowbytes);
			for ( x=pixbits[y]; x; x<<=1 )
			{	if ( x>=0 )  {  pixel++; attr++; pix++; continue;  };
				c = ( *pix++ & bwm[*attr] ) ^ bwn[*attr]; attr++;	
				*pixel++ = Hnibble[c>>4] + Lnibble[c&0x0f];
			};
			*(LineB4*)((long)p0+rowbytes) = *(LineB4*)((long)p0+rowbytes*2) = *(LineB4*)((long)p0+rowbytes*3) = *(LineB4*)p0;
		};
	};
}


// -----	Accelerated refresh routine for 8-bit screens and scrnsize=1 ------------------------
static RefreshSingle8 ()
{	long		x,y,c,n; 
	Char		*attr, *pix;
	Char		p, i, *pixel;

	for ( y=0; y<24*8; y++ )
	{	if ( pixbits[y] )
		{	attr	= new_attr+32* ( run_highres ? y : (y>>3) );
			pix	= new_pixels+32*n2n[y];
			pixel	= screenbase + y*rowbytes;
			for ( x=pixbits[y]; x; x<<=1 )
			{	if ( x>=0 ) { pixel += 8; attr++; pix++; continue; }
				p = c8[pap[*attr]];    i = c8[ink[*attr++]];    c = *pix++;
				for ( n=0x80; n; n>>=1 )  *pixel++ = (c&n) ? i : p;
			};
		};
	};
}


// -----	Beschleunigte Refreshroutine fr 8-Bit-Bildschirm und scrnsize=2 ------------------------
static RefreshDouble8 ()
{	long		x,y,c,n;
	Char 	*attr,*pix;
	Short	p, i, *pixel, *paxel;
	
	for ( y=0; y<24*8; y++ )
	{	if ( pixbits[y] )
		{	attr	= new_attr+32* ( run_highres ? y : (y>>3) );
			pix	= new_pixels+32*n2n[y];
			pixel	= (Short*)(screenbase + y*2*rowbytes);
			paxel= (Short*)((long)pixel + rowbytes);
			for ( x=pixbits[y]; x; x<<=1 )
			{	if ( x>=0 ) { pixel += 8;  paxel += 8; attr++; pix++; continue; }
				p = *(Short*)(d8+pap[*attr]);    i = *(Short*)(d8+ink[*attr++]);   c = *pix++;
				for ( n=0x80; n; n>>=1 )  *pixel++ = *paxel++ = (c&n) ? i : p;
			};
		};
	}
}


// -----	Beschleunigte Refreshroutine fr 8-Bit-Bildschirm und scrnsize=3 ------------------------
static RefreshTriple8 ()
{	long		x,y,c,n;
	Char 	*attr,*pix;
	Char		p, i, *pixel, *p0;	
	
	for ( y=0; y<24*8; y++ )
	{	if ( pixbits[y] )
		{	attr	= new_attr+32* ( run_highres ? y : (y>>3) );
			pix	= new_pixels+32*n2n[y];
			p0 = pixel	= (Char*)(screenbase + y*3*rowbytes);
			for ( x=pixbits[y]; x; x<<=1 )
			{	if ( x>=0 ) { pixel += 24; attr++; pix++; continue; };
				{	p = c8[pap[*attr]];    i = c8[ink[*attr++]];    c = *pix++;
					for ( n=0x80; n; n>>=1 )  *pixel++ = *pixel++ = *pixel++ = (c&n) ? i : p;
				}
			};
			*(Line3*)(p0+rowbytes) = *(Line3*)(p0+2*rowbytes) = *(Line3*)p0;
		};
	}
}


static RefreshQuad8()
{	long		x,y,c,n;
	Char 	*attr,*pix;
	Long		p, i, *pixel, *p0;
	
	for ( y=0; y<24*8; y++ )
	{	if ( pixbits[y] )
		{	attr	= new_attr+32* ( run_highres ? y : (y>>3) );
			pix	= new_pixels+32*n2n[y];
			p0 = pixel	= (Long*)(screenbase + y*4*rowbytes);
			for ( x=pixbits[y]; x; x<<=1 )
			{	if ( x>=0 ) { pixel+=8; attr++; pix++; continue; };
				p = *(Long*)(q8+pap[*attr]);    i = *(Long*)(q8+ink[*attr++]);    c = *pix++;
				for ( n=0x80; n; n>>=1 )  *pixel++ = (c&n) ? i : p;
			};
			*(Line4*)((long)p0+rowbytes) = *(Line4*)((long)p0+2*rowbytes) = *(Line4*)((long)p0+3*rowbytes) = *(Line4*)p0;			
		};
	}
}


// -----	Beschleunigte Refresh-Routine fr 16-Bit-Bildschirm und scrnsize=1 ------------------------
static RefreshSingle16 ()
{	long		x,y,c,n; 
	Char		*attr, *pix;
	Short	p, i, *pixel;

	for ( y=0; y<24*8; y++ )
	{	if ( pixbits[y] )
		{	attr	= new_attr+32* ( run_highres ? y : (y>>3) );
			pix	= new_pixels+32*n2n[y];
			pixel	= (Short*)(screenbase + y*rowbytes);
			for ( x=pixbits[y]; x; x<<=1 )
			{	if ( x>=0 ) { pixel += 8; attr++; pix++; continue; };
				p = c16[pap[*attr]];    i = c16[ink[*attr++]];    c = *pix++;
				for ( n=0x80; n; n>>=1 )  *pixel++ = (c&n) ? i : p;
			};
		};
	};
}


// -----	Beschleunigte Refreshroutine fr 16-Bit-Bildschirm und scrnsize=2 ------------------------
static RefreshDouble16 ()
{	long		x,y,c,n;
	Char 	*attr,*pix;
	Long		p, i, *pixel, *paxel;
	
	for ( y=0; y<24*8; y++ )
	{	if ( pixbits[y] )
		{	attr	= new_attr+32* ( run_highres ? y : (y>>3) );
			pix	= new_pixels+32*n2n[y];
			pixel	= (Long*)(screenbase + y*2*rowbytes);
			paxel= (Long*)((long)pixel + rowbytes);
			for ( x=pixbits[y]; x; x<<=1 )
			{	if ( x>=0 ) { pixel += 8;  paxel += 8; attr++; pix++; continue; };
				p = *(Long*)(d16+pap[*attr]);    i = *(Long*)(d16+ink[*attr++]);    c = *pix++;
				for ( n=0x80; n; n>>=1 )  *pixel++ = *paxel++ = (c&n) ? i : p;
			};
		};
	}
}


// -----	Beschleunigte Refreshroutine fr 16-Bit-Bildschirm und scrnsize=3 ------------------------
static RefreshTriple16 ()
{	long		x,y,c,n;
	Char 	*attr,*pix;
	Short3	p, i, *pixel, *p0;
	
	for ( y=0; y<24*8; y++ )
	{	if ( pixbits[y] )
		{	attr	= new_attr+32* ( run_highres ? y : (y>>3) );
			pix	= new_pixels+32*n2n[y];
			p0 = pixel	= (Short3*)(screenbase + y*3*rowbytes);
			for ( x=pixbits[y]; x; x<<=1 )
			{	if ( x>=0 ) { pixel+=8; attr++; pix++; continue; }
				p = t16[pap[*attr]];    i = t16[ink[*attr++]];    c = *pix++;
				for ( n=0x80; n; n>>=1 )  *pixel++ = (c&n) ? i : p;
			};
			*(Line6*)((long)p0+rowbytes) = *(Line6*)((long)p0+rowbytes*2) = *(Line6*)p0;
		};
	}
}


// -----	Beschleunigte Refreshroutine fr 16-Bit-Bildschirm und scrnsize=4 ------------------------
static RefreshQuad16 ()
{	long		x,y,c,n;
	Char 	*attr,*pix;
	Short4	p, i, *pixel, *p0;
	
	for ( y=0; y<24*8; y++ )
	{	if ( pixbits[y] )
		{	attr	= new_attr+32* ( run_highres ? y : (y>>3) );
			pix	= new_pixels+32*n2n[y];
			p0 = pixel	= (Short4*)(screenbase + y*4*rowbytes);
			for ( x=pixbits[y]; x; x<<=1 )
			{	if ( x>=0 ) { pixel+=8; attr++; pix++; continue; }
				p = q16[pap[*attr]];    i = q16[ink[*attr++]];    c = *pix++;
				for ( n=0x80; n; n>>=1 )  *pixel++ = (c&n) ? i : p;
			};
			*(Line8*)((long)p0+rowbytes) = *(Line8*)((long)p0+rowbytes*2) = *(Line8*)((long)p0+rowbytes*3) = *(Line8*)p0;
		};
	}
}


// -----	Beschleunigte Refresh-Routine fr 32-Bit-Bildschirm und scrnsize=1 ------------------------
static RefreshSingle32 ()
{	long		x,y,c,n; 
	Char		*attr, *pix;
	Long		p, i, *pixel;

	for ( y=0; y<24*8; y++ )
	{	if ( pixbits[y] )
		{	attr	= new_attr+32* ( run_highres ? y : (y>>3) );
			pix	= new_pixels+32*n2n[y];
			pixel	= (Long*)(screenbase + y*rowbytes);
			for ( x=pixbits[y]; x; x<<=1 )
			{	if ( x>=0 ) { pixel += 8; attr++; pix++; continue; }
				p = c32[pap[*attr]];    i = c32[ink[*attr++]];    c = *pix++;
				for ( n=0x80; n; n>>=1 )  *pixel++ = (c&n) ? i : p;
			};
		};
	};
}


// -----	Beschleunigte Refreshroutine fr 32-Bit-Bildschirm und scrnsize=2 ------------------------
static RefreshDouble32 ()
{	long		x,y,c,n;
	Char 	*attr,*pix;
	Long2	p, i, *pixel, *paxel;
	
	for ( y=0; y<24*8; y++ )
	{	if ( pixbits[y] )
		{	attr	= new_attr+32* ( run_highres ? y : (y>>3) );
			pix	= new_pixels+32*n2n[y];
			pixel	= (Long2*)(screenbase + y*2*rowbytes);
			paxel= (Long2*)((long)pixel + rowbytes);
			for ( x=pixbits[y]; x; x<<=1 )
			{	if ( x>=0 ) { pixel += 8;  paxel += 8; attr++; pix++; continue; }
				p = d32[pap[*attr]];    i = d32[ink[*attr++]];    c = *pix++;
				for ( n=0x80; n; n>>=1 )  *pixel++ = *paxel++ = (c&n) ? i : p;
			};
		};
	}
}


// -----	Beschleunigte Refreshroutine fr 32-Bit-Bildschirm und scrnsize=3 ------------------------
static RefreshTriple32 ()
{	long		x,y,c,n;
	Char 	*attr,*pix;
	Long3	p, i, *pixel, *p0;
	
	for ( y=0; y<24*8; y++ )
	{	if ( pixbits[y] )
		{	attr	= new_attr+32* ( run_highres ? y : (y>>3) );
			pix	= new_pixels+32*n2n[y];
			p0 = pixel	= (Long3*)(screenbase + y*3*rowbytes);
			for ( x=pixbits[y]; x; x<<=1 )
			{	if ( x>=0 ) { pixel+=8; attr++; pix++; continue; }
				p = t32[pap[*attr]];    i = t32[ink[*attr++]];    c = *pix++;
				for ( n=0x80; n; n>>=1 )  *pixel++ = (c&n) ? i : p;
			};
			*(Line12*)((long)p0+rowbytes) = *(Line12*)((long)p0+rowbytes*2) = *(Line12*)p0;
		};
	}
}


// -----	Beschleunigte Refreshroutine fr 32-Bit-Bildschirm und scrnsize=4 ------------------------
static RefreshQuad32 ()
{	long		x,y,c,n;
	Char 	*attr,*pix;
	Long4	p, i, *pixel, *p0;
	
	for ( y=0; y<24*8; y++ )
	{	if ( pixbits[y] )
		{	attr	= new_attr+32* ( run_highres ? y : (y>>3) );
			pix	= new_pixels+32*n2n[y];
			p0 = pixel	= (Long4*)(screenbase + y*4*rowbytes);
			for ( x=pixbits[y]; x; x<<=1 )
			{	if ( x>=0 ) { pixel+=8; attr++; pix++; continue; }
				p = q32[pap[*attr]];    i = q32[ink[*attr++]];    c = *pix++;
				for ( n=0x80; n; n>>=1 )  *pixel++ = (c&n) ? i : p;
			};
			*(Line16*)((long)p0+rowbytes) = *(Line16*)((long)p0+rowbytes*2) = *(Line16*)((long)p0+rowbytes*3) = *(Line16*)p0;
		};
	}
}


// -----	update specci window -----------------------------------------
//		-> flash colors
//		-> update border color
//		-> update screen contents

RefreshSpecciWindow()
{	Rect		box,border;
	PixMap	*winpixmap, 		// specci window pixmap
			*devpixmap;		// graphic device pixmap
	PixMap	**offpixmap;		// offscreen pixmap
	short	pixelsize;			// of specci window (immediate) or offscreen bitmap
	GDHandle	devhandle;		// graphic device of specci screen

	if ( !zxwindow ) return;
	SetPort ( zxwindow );
	
// check flashing colors:
	{	static Long 	oldTick = 0; 
		static int		flashphase = 0;
		Long 			tick;
		register Char	*attr, *anf;
		tick=TickCount();
		if ( (tick-oldTick) >= 30 )
		{	oldTick = tick;
			flashphase = !flashphase;
			anf	= old_attr;
			attr	= old_end;
			while ( attr>anf )  if ( *--attr & 0x80 )  *attr = pink[*attr];
			if (flashphase)
			{	MemCopy(ink,ink+128,sizeof(*ink)*128);
				MemCopy(pap,pap+128,sizeof(*pap)*128);
			}
			else
			{	MemCopy(pap,ink+128,sizeof(*ink)*128);
				MemCopy(ink,pap+128,sizeof(*pap)*128);
			};
		};
	};


// check border color:
	if (brdrlogptr>brdrlog+1)			// more than one border change in highres mode
	{	long t0,t1,t,y,*blp; char *obp,c ;

		GetClip(tempRgn);
		SetClip(borderRgn);
		PenSize(1,scrnsize);
		SetOrigin(0,0);
		
		if (oldbordercolor<0x80) memset ( oldbrdr, oldbordercolor, brdrlines );

		t0 = TOTAL+180-70000+line_cycles*(above_scrn-vborder/scrnsize);	// first time of interest
		
		t1 = t0+line_cycles*(192+2*vborder/scrnsize);				// last time of interest

		*brdrlogptr = t1;		// final value

		c = oldbordercolor&0x07;
		blp = brdrlog;
		obp = oldbrdr;
		y = blkborder;

		while ((t0-t1)<0)
		{	t = *blp;  if (t>t1) t=t1;
			while ( (t0-t)<0 )  
			{	if (*obp!=c)  {  RGBForeColor(rgb+c);  MoveTo(blkborder,y); Line(2000,0);  }
				y+=scrnsize; 
				*obp++ = c; t0+=line_cycles;
			}
			c = *blp++ & 0x07;
		}

		bordercolor = brdrlogptr[-1] & 0x07;
		oldbordercolor = 0x80 + bordercolor;
		brdrlogptr = brdrlog;
		SetClip ( tempRgn );
		PenSize(1,1);
	}
	else
	if (bordercolor!=oldbordercolor)	// border change normal mode or last frame in highres mode !!!
	{	short	i,x,y;
		
		border = box= zxwindow->portRect;

		GetClip(tempRgn);
		SetClip(borderRgn);
		
		switch (bordercolor)
		{
		case syncborder:
			RGBForeColor ( rgb+2 );		// red
			PaintRect ( &border ); 
			RGBForeColor ( rgb+8+5 );	// cyan
			x=9; y=4;
			goto db;
			
		case databorder:
			RGBForeColor (rgb+6 );		// yellow
			PaintRect ( &border ); 
			RGBForeColor ( rgb+1 );		// blue
			x=4; y=2;
db:			x*=scrnsize; y*=scrnsize;
			for ( i=1; i<border.bottom; i+=x )
			{	box.top=i; box.bottom=i+y;
				PaintRect ( &box );
			};
			break;

		default:
			RGBForeColor ( rgb + bordercolor );
			PaintRect ( &border ); 
		};

		oldbordercolor=bordercolor;
		SetClip ( tempRgn );
	};


// check writable interior screen:

	if (last_update_method)
	{	if ( !memcmp ( old_pixels, new_pixels, new_end-new_pixels ) ) return true;		// nothing changed
		update_pixflags();
	}
	MemCopy ( new_pixels, old_pixels, new_end-new_pixels );

	if (require_offscreen||force_offscreen) goto offscr; 			// multiple monitors covered or unsupported screen depth

	DiffRgn ( screenRgn, zxwindow->visRgn, tempRgn );
	if (!EmptyRgn(tempRgn)) goto offscr;
	
//	box = (**(zxwindow->visRgn)).rgnBBox;		// something eclipses parts of the specci window
//	if  (	box.left>screenRect.left || box.right<screenRect.right || box.bottom<screenRect.bottom || box.top>screenRect.top || 
//		(**(zxwindow->visRgn)).rgnSize != 10  )  goto offscr;

// *** draw immediately to the screen ***

	if (last_update_method!=immediate)  memset ( pixbits, 0xFF, 24*8*sizeof(*pixbits) );
	last_update_method = immediate;

	winpixmap	= *((CGrafPtr)zxwindow)->portPixMap;

	devhandle		= monitor[0];
	devpixmap	= *(**devhandle).gdPMap;
	pixelsize		= devpixmap->pixelSize;	
	rowbytes		= devpixmap->rowBytes & 0x3FFF;

	screenbase	= ((Char*)devpixmap->baseAddr)
				+ ( (hborder+blkborder) - winpixmap->bounds.left - devpixmap->bounds.left ) *  pixelsize / 8
				+ ( (vborder+blkborder) - winpixmap->bounds.top - devpixmap->bounds.top ) *  rowbytes;	

	ShieldCursor ( &screenRect, *(Point*)&(winpixmap->bounds) );

	switch ( scrnsize )
	{	case 1:
			switch ( pixelsize )
			{	case 1:	RefreshSingle1(); goto x;
				case 8:	RefreshSingle8(); goto x;
				case 16:	RefreshSingle16(); goto x;
				default:	RefreshSingle32(); goto x;
			}
		case 2:
			switch ( pixelsize )
			{	case 1:	RefreshDouble1(); goto x;
				case 8:	RefreshDouble8(); goto x;
				case 16:	RefreshDouble16(); goto x;
				default:	RefreshDouble32(); goto x;
			}
		case 3:
			switch ( pixelsize )
			{	case 1:	RefreshTriple1(); goto x;
				case 8:	RefreshTriple8(); goto x;
				case 16:	RefreshTriple16(); goto x;
				default:	RefreshTriple32(); goto x;
			}
		default:	// 4
			switch ( pixelsize )
			{	case 1:	RefreshQuad1(); goto x;
				case 8:	RefreshQuad8(); goto x;
				case 16:	RefreshQuad16(); goto x;
				default:	RefreshQuad32(); goto x;
			}
	}	
x:	ShowCursor(); 
	return;		
			
// *** draw offscreen
offscr:
	if (last_update_method!=offscreen)  memset ( pixbits, 0xFF, 24*8*sizeof(*pixbits) );
	last_update_method = offscreen;

	offpixmap		= offscreenGWorldPtr->portPixMap;
	pixelsize		= (**offpixmap).pixelSize;
	rowbytes		= (**offpixmap).rowBytes & 0x3FFF;
	screenbase 	= (Char*)GetPixBaseAddr(offpixmap); 
	if (pixelsize>=8) screenbase -= (**offpixmap).bounds.left*pixelsize/8;

	switch ( scrnsize )
	{	case 1:
			switch ( pixelsize )
			{	case 1:	RefreshSingle1(); goto xx;
				case 8:	RefreshSingle8(); goto xx;
				case 16:	RefreshSingle16(); goto xx;
				default:	RefreshSingle32(); goto xx;
			}
		case 2:
			switch ( pixelsize )
			{	case 1:	RefreshDouble1(); goto xx;
				case 8:	RefreshDouble8(); goto xx;
				case 16:	RefreshDouble16(); goto xx;
				default:	RefreshDouble32(); goto xx;
			}
		case 3:
			switch ( pixelsize )
			{	case 1:	RefreshTriple1(); goto xx;
				case 8:	RefreshTriple8(); goto xx;
				case 16:	RefreshTriple16(); goto xx;
				default:	RefreshTriple32(); goto xx;
			}
		default:	// 4
			switch ( pixelsize )
			{	case 1:	RefreshQuad1(); goto xx;
				case 8:	RefreshQuad8(); goto xx;
				case 16:	RefreshQuad16(); goto xx;
				default:	RefreshQuad32(); goto xx;
			}
	}	

xx:	ForeColor(blackColor);
	BackColor(whiteColor);

	{	GDHandle ogd,*gd;
		ogd = GetGDevice();
		for ( gd=monitor; *gd; gd++ )
		{	SetGDevice(*gd);
			SetPort(zxwindow);
			CopyBits (	&((GrafPtr)offscreenGWorldPtr)->portBits,	&zxwindow->portBits, 
					&offscreenGWorldPtr->portRect,			&screenRect,   srcCopy, nil );
		}
		SetGDevice(ogd);
	}
	
//	MoveTo(screenRect.left,screenRect.top+12);
//	TextMode(srcCopy);
//	TextSize(12);
//	DrawChar('');
}


/* ----	Update entire ZX Spectrum window -----------------------------------------------
		call this routine for updating and after moving the window
		calculate indexed colors index table, allocate offscreen bitmap and check bitmap alignment
		update preferences (windows location)
		on error, the window is closed and the z80 cpu is halted!
    note:	The Specci window is not updated immediately, all contents is only marked as invalid,
		because UpdateSpecciWindow() may be called to accomplish a MacOS window refresh event
		and at this moment additional clipping is applied, while RefreshSpecciWindow() would 
		still think that it refreshed the complete window and mark it entirely as up to date
*/
UpdateSpecciWindow()
{	PixMapPtr  pm; 
	int err; int i;
	Rect  box;
	GDHandle ogd;
	short depth;			// max. color depth of monitors overlapped by Specci screen
	
	 if ( !zxwindow ) return error;		// not open
	SetPort(zxwindow);
	ogd = GetGDevice();
	grayRgn = LMGetGrayRgn();

// calculate blackRgn, borderRgn, screenRgn and screenRect:
// screenRgn and screenRect describe the ZX Spectrum's writable screen
// borderRgn describes the border around the screen which receives border color and border effects
// blackRgn describes the region around the border, which is blacked out or outside the Specci window
	box = zxwindow->portRect;
	RectRgn ( blackRgn, &box );
	InsetRect ( &box, blkborder, blkborder );
	RectRgn ( borderRgn, &box );
	InsetRect ( &box, hborder, vborder );	screenRect = box;
	RectRgn ( screenRgn, &box );
	XorRgn ( blackRgn, borderRgn, blackRgn );
	XorRgn ( borderRgn, screenRgn, borderRgn );

// update preferences
	box = zxwindow->portRect;
	InsetRect ( &box, blkborder, blkborder );
	LocalToGlobal ( (Point*)&box ); 
	LocalToGlobal (((Point*)&box)+1 ); 
	specciRect[scrnsize] = box;

// black out rest of screen (e.g. menu bar)
	if (blacked_out)
	{	Point p={0,0};
		GlobalToLocal(&p);
		OffsetRgn (mbarRgn, p.h,p.v);
		if (mbar_present)	DiffRgn(zxwindow->visRgn,mbarRgn,zxwindow->visRgn);
		else				UnionRgn(zxwindow->visRgn,mbarRgn,zxwindow->visRgn);
		OffsetRgn (mbarRgn, -p.h,-p.v);
		GetClip(tempRgn);
		SetClip(blackRgn);
		ForeColor(blackColor);
		PaintRect(&zxwindow->portRect);
		SetClip(tempRgn);
		if (mbar_present) DrawMenuBar();
	}

// screen contents and offscreen bitmap are invalid
	if (run_highres)
	{	Char *q,*z; int i,j;
		
		old_pixels		= old_screen;
		old_attr		= old_pixels+32*24*8;
		old_end		= old_attr+32*24*8;
	
		new_pixels	= new_screen;
		new_attr		= new_pixels+32*24*8;
		new_end		= new_attr+32*24*8;
		
		MemCopy(video_page,new_screen,32*24*8);
		q = video_page+32*24*8; 
		z = new_attr;
		for ( i=24; i--; q+=32 )  for ( j=8; j--; z+=32 )  MemCopy(q,z,32);
	}
	else		
	{	old_pixels		= old_screen;
		old_attr		= old_pixels+32*24*8;
		old_end		= old_attr+32*24;
	
		new_pixels	= video_page;
		new_attr		= new_pixels+32*24*8;
		new_end		= new_attr+32*24;
	}
	last_update_method = 0;			// screen contents invalid

// force border update
	oldbordercolor=99;				// force redraw of border
	if (run_highres)
	{	if (!brdrlog) brdrlog = (long*)NewPtr((brdrlogsize+1)*sizeof(*brdrlog));
		if (brdrlog) { brdrlogptr=brdrlog; brdrlogend = brdrlog+brdrlogsize; };
		memset ( oldbrdr, 99, brdrlines );
	}
	else
	{	if (brdrlog) DisposPtr((Ptr)brdrlog);
		brdrlog = brdrlogptr = brdrlogend = nil;
	};
	
// find all monitors which intersect with screen rectangle
	{	Rect		sbox, gbox;
		GDHandle	gd;
		int i;
		depth = 1;				// max. color depth of monitors overlapped by Specci screen
		sbox = screenRect;
		LocalToGlobal((Point*)&sbox);
		LocalToGlobal(((Point*)&sbox)+1);
		for ( i=0, gd=GetDeviceList(); i<maxmon && gd; gd=GetNextDevice(gd) )
		{	gbox = (**gd).gdRect;
			if  (	sbox.left < gbox.right	&& 
				sbox.right > gbox.left	&&
				sbox.top < gbox.bottom	&&
				sbox.bottom > gbox.top
			   )	
			{	pm = *(**gd).gdPMap;
				if (depth<8)	{ if (pm->pixelSize>=depth)	{ depth=pm->pixelSize; SetGDevice(gd); }	}
				else			{ if (pm->pixelSize<depth)		{ depth=pm->pixelSize; SetGDevice(gd); }	}
				monitor[i++] = gd;
			}
		}
		monitor[i] = nil;		// end of list
		require_offscreen = (i!=1) || depth==2 || depth==4;	// multiple screens covered or not supported color depth
	}

// initialize indexed color conversion tables from actual GDevice (used only if really drawing on indexed color GDevice)
	for ( i=0; i<16; i++ ) 
	{	q8[i].a = q8[i].b = q8[i].c = q8[i].d =			// 16 * 8-bit indexed color quad_chars		(quad size window)
		t8[i].a = t8[i].b = t8[i].c = 				// 16 * 8-bit indexed color triple_chars		(quad size window)
		d8[i].a = d8[i].b = 						// 16 * 8-bit indexed color double_chars		(quad size window)
		c8[i] = Color2Index(rgb+i);				// 16 * 8-bit indexed color bytes			(single size window)
	};

// align window to byte/word/long word  boundary
	// force byte boundary if 1, 2 or 4 bit per pixel
	pm = *((CGrafPtr)zxwindow)->portPixMap;
	if ( depth<8 && pm->bounds.left&7 ) 	MoveWindow ( zxwindow, ((-pm->bounds.left+3)&0xfff8), -pm->bounds.top, false );
	// force corresponding pixels in same word or long word if scrnsize==2
	pm = *((CGrafPtr)zxwindow)->portPixMap;	// may have moved!
	if (scrnsize==2 && pm->bounds.left&1 )	MoveWindow ( zxwindow, ((0-pm->bounds.left)&0xfffe), -pm->bounds.top, false );
	if (scrnsize==4 && pm->bounds.left&3 )	MoveWindow ( zxwindow, ((2-pm->bounds.left)&0xfffc), -pm->bounds.top, false );

// allocate offscreenbitmap: use current GDevice's clut if indexed color
	if (offscreenGWorldPtr)  DisposeGWorld(offscreenGWorldPtr);  offscreenGWorldPtr=nil;
	box = screenRect;
	LocalToGlobal((Point*)&box);
	LocalToGlobal(((Point*)&box)+1);
	if (depth<8) depth = 1; else depth = 0;	//force b&w for 2 and 4 bit color else use depth of current GDevice using it's clut if ic
	err = NewGWorld ( &offscreenGWorldPtr, depth, &box, nil, nil, useTempMem );

	if (err)
	{	PrintAlert(CStr(localized("No memory for offscreen bitmap!","Kein Speicher mehr frei fr die Offscreen-Bitmap")));
		offscreenGWorldPtr = nil;
		DisposeSpecciWindow();
		StartStop(false);
	}

	SetGDevice(ogd);
	return err;
}


// -----	Close and discard ZX Spectrum window -------------------------------------------
DisposeSpecciWindow()
{	CheckItem ( windowsmenu, iSpectrum, false );
	if (zxwindow) DisposeWindow ( zxwindow ); 				zxwindow = nil;
	if (offscreenGWorldPtr) DisposeGWorld(offscreenGWorldPtr);	offscreenGWorldPtr = nil;
	if (brdrlog) DisposPtr((Ptr)brdrlog);						brdrlog = brdrlogptr = brdrlogend = nil;
}


SetSpecciWindowName ( Char* name )
{
	if (*name>31) *name=31;		// security only
	MemCopy ( name, specciWindowName, *name+1 );
	if (!zxwindow) return;	
	if (z80_running)	SetWTitle ( zxwindow, specciWindowName );
	else				SetWTitle ( zxwindow, PascalStr(CatStr(PStr(specciWindowName),PStr("\p (halted)"))) );
}


// -----	Create and open new ZX Spectrum window ----------------------------------------
OSErr NewSpecciWindow (short size)
{	Rect box;

	if (zxwindow)
	{	if (size==scrnsize) return ok;
		DisposeSpecciWindow();
	};
	scrnsize = size;

// allocate window record
	box = specciRect[scrnsize];
	hborder = (((box.right-box.left)-scrnsize*256)/16)*8;	hborder = minmax ( 8, hborder, 32*scrnsize );
	vborder = (((box.bottom-box.top)-scrnsize*192)/16)*8;	vborder = minmax ( 8, vborder, 16*scrnsize );
	SetRect ( &box, box.left, box.top, box.left+2*hborder+scrnsize*256, box.top+2*vborder+scrnsize*192 );
	InsetRect ( &box, -blkborder, -blkborder );
	zxwindow = NewCWindow ( nil, &box, "\p", false, noGrowDocProc, (WindowPtr)-1, true, 0 );
	if ( !zxwindow ) goto x;
	
// open window	
	SetSpecciWindowName ( specciWindowName );
	DisplayWindow ( zxwindow );
	SetWRefCon ( zxwindow, iSpectrum );

// allocate offscreen & redraw window
	UpdateSpecciWindow();	
	CheckItem ( windowsmenu, iSpectrum, true );
	if (zxwindow)  return ok;

x:	PrintAlert(CStr(localized("Out of memory: NewSpecciWindow()","Speicherberlauf: NewSpecciWindow()")));
	DisposeSpecciWindow();
	StartStop(false);
	return oomem;
}


/* -----	Black out and Menu bar control ----------------------------------------------------------
	defined constants for flag:
		0 = multi_window
		1 = mbar_on  /  black_out
		2 = mbar_off
*/
MBarControl ( short flag )
{	Rect mbarRect;
	GDHandle gdh;
	
	if (mbar_present) mbarHeight = LMGetMBarHeight();
	gdh = GetMainDevice ();
	mbarRect = (**gdh).gdRect;
	mbarRect.bottom = mbarRect.top+mbarHeight;
	RectRgn (mbarRgn, &mbarRect);				// region covers menu bar

	switch (flag)
	{
	case multi_window:
		if (!blacked_out) return;
		if (!mbar_present) MBarControl(show_mbar);
		DisposeSpecciWindow();
		blkborder = 0;
		NewSpecciWindow ( scrnsize );
		blacked_out = false;
		break;
	
	case black_out:		// or show_mbar
		if (!blacked_out)
		{	DisposeSpecciWindow();
			blkborder = 6000;
			NewSpecciWindow ( scrnsize );
			blacked_out = true;
		}
		else if (!mbar_present)
		{	DiffRgn(grayRgn,mbarRgn,grayRgn);
			LMSetMBarHeight(mbarHeight);
			DrawMenuBar();
			ShowCursor();
			mbar_present = true;
		}
		else return;
		break;
		
	case hide_mbar:
		if (!mbar_present) return;
		if (!blacked_out) MBarControl(black_out);
		UnionRgn(grayRgn,mbarRgn,grayRgn);
		LMSetMBarHeight(0);
		HideCursor();
		mbar_present = false;
	}
	
	UpdateSpecciWindow();
}







