/*
***	Copyright:	1997-1999 Gnter Woigk
  				mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/


#include 	<Gestalt.h> 
#include	<power.h>

#include	"kio.h"


// ==========================================================================
// ============================= global variables ===========================
// ==========================================================================


OSErr		err		= noerror;		// central error dump
OSErr		_err_;				// scratch error cell for TRAP()

Boolean		done		= false;		// quit application?
Boolean		abo		= false;		// abort current operation ?

Boolean		powerMgr	= false;		// Power manager present ?
Boolean		initappl	= false;		// InitAppl() already called ?

EventRecord	event;				// event returned by WaitNextEvent()



// ==========================================================================
// ========================== INITIALIZATION of APPLICATION =========================
// ==========================================================================



void InitAppl ( )
{
	if (initappl) return;		// <- don't rely on this; just security...

// init Apple stuff...
	MoreMasters();
	MoreMasters();

	InitGraf(&qd.thePort);		
	InitFonts();			
	InitWindows();			
	InitMenus();			
	TEInit();				
	InitDialogs(nil);
	InitCursor();

// now we can display alerts...
	initappl = true;

// init Application stuff...
	CheckGestalt();
	InitAppleEvents();

	GetDateTime((Long*) &qd.randSeed);
}



// ==========================================================================
// =========================== Check system environment ===========================
// ==========================================================================

#define DoAbort(a) do_abort(a,0,0)



/* the following options may be defined in "options.h":
	min_cpu		e.g.	gestalt68020
	req_fpu		e.g.	true
	min_QD		e.g.	gestalt32BitQD
	System7 stuff is always tested
	potentially, this file is obsolete for PowerPC
	though it also tests for power management stuff
*/


/* -----	performs additional checks depending on #defines in options.h -----------------------
*/
void CheckGestalt ( )
{	long answer;
	OSErr err;

// ----	check that Gestalt() is available ----------------------------------
	err = Gestalt (gestaltVersion, &answer);						
	if (err) DoAbort ( "\pThis program requires System 7.0 or higher" );

// ----	check cpu ---------------------------------------- (( if required ))
#if !__POWERPC__ && min_cpu
	err = Gestalt (gestaltProcessorType, &answer);				
	err = err || ( answer<min_cpu );
	if (err)
		if ( min_cpu==gestalt68020 )	DoAbort ( "\pThis program requires a 68020 processor or better" );
		else						DoAbort (	"\pThis program requires a newer processor" );
#endif

// ----	check for 68881 coprocessor ---------------------- (( if required ))
#if !__POWERPC__ && req_fpu
	err = Gestalt (gestaltFPUType, &answer);					
	err = err || answer<gestalt68881;
	if (err) DoAbort ( "\pThis program requires a floating point coprocessor" );
#endif

// ----	check for system 7.0 or higher -------------------------------------
#if !__POWERPC__
	err = Gestalt ( gestaltSystemVersion, &answer );
	err = err || answer<0x0700;					// e.g. 0x0604 for system 6.0.4
	if (err) DoAbort ( "\pThis program requires System 7.0 or higher" );
#endif
	
// ----	check for HighLevelEvents ------------------------------------------
//		always required for 'oapp', 'odoc', 'pdoc', and 'quit' 
// 		maybe obsolete: system 7.0 should imply that apple events are present
#if !__POWERPC__
	err = Gestalt(gestaltAppleEventsAttr, &answer);			
	err = err || !( answer & (1L<<gestaltAppleEventsPresent) );	
	if (err) DoAbort ( "\pApple Events not present" );
#endif

// ----	check for FindFolder() ---------------------------------------------
// 		mainly needed to find the Preferences folder
// 		maybe obsolete: system 7.0 should imply that FindFolder() is present
#if !__POWERPC__
	err = Gestalt ( gestaltFindFolderAttr, &answer );
	err = err || !( answer & (1L<<gestaltFindFolderPresent) );		
	if (err) DoAbort ( "\pFindFolder() not present" );	
#endif

// ----	check for temporary memory support ---------------------------------
//		i prefer allocation larger blocks of memory outside the applications heap
// 		maybe obsolete: system 7.0 should imply that temp mem is available
#if !__POWERPC__
	err = Gestalt ( gestaltOSAttr, &answer );	
	err = err || !( answer & (1L<<gestaltTempMemSupport) );		
	if (err) DoAbort ( "\pTemporary memory manager not present" );
#endif

// -----	check for power manager present ------------------------------------
	err = Gestalt ( gestaltPowerMgrAttr, &answer );
	powerMgr = err ? false : answer & (1L<<gestaltPMgrExists);

// -----	check for Color Quickdraw ------------------------------------------
#if !__POWERPC__
	err = Gestalt ( gestaltQuickdrawVersion, &answer );			
	err = err || answer<gestalt32BitQD;
	if (err) DoAbort ( "\pThis program requires 32 bit QuickDraw" );
#if min_QD
#if min_QD != gestalt32BitQD
	err = err || answer<min_QD;
	if (err) DoAbort ( "\pThis program requires a newer version of QuickDraw" );
#endif
#endif
#endif
}



// ==========================================================================
// =========================== Handle Apple Events ================================
// ==========================================================================


static OSErr _got_required_params ( AppleEvent *theAppleEvent )
{	DescType	returnedType;
	Size		actualSize;
	OSErr 	err;

	err = AEGetAttributePtr ( theAppleEvent, keyMissedKeywordAttr, typeWildCard, &returnedType, nil, 0, &actualSize);
	if (err==errAEDescNotFound)	return noErr;		// got all the required parameters
	if (err==noErr) return errAEParamMissed;			// missed a required parameter	
	return err;								// call to AEGetAttributePtr failed
}


static pascal OSErr _handle_NEW ( AppleEvent *eventIn, AppleEvent */*eventOut*/, long /*refCon*/ )
{	OSErr err;

	err = _got_required_params ( eventIn ); 
	if (!err) HandleNEW();
	return err; 
}


static pascal OSErr _handle_QUIT ( AppleEvent *eventIn, AppleEvent */*eventOut*/, long /*refCon*/ )
{	OSErr err;

	err = _got_required_params ( eventIn ); 
	done = true;
	return err;	
}


static pascal OSErr _handle_OPEN ( AppleEvent *eventIn, AppleEvent */*eventOut*/, long /*refCon*/ )
{	FSSpec	file;
	AEDescList	docList;
	long		index, itemsInList;
	Size		actualSize;
	AEKeyword	keywd;
	DescType	returnedType;
	OSErr 	err;

	err = AEGetParamDesc ( eventIn, keyDirectObject, typeAEList, &docList );
	if (err) return err; 
	err = _got_required_params ( eventIn );		
	if (!err) err = AECountItems ( &docList, &itemsInList );
	for (index=1; index<=itemsInList && !err; index++) 
	{	err = AEGetNthPtr ( &docList, index, typeFSS, &keywd, &returnedType, (Ptr)&file, sizeof(file), &actualSize );
		#if OPEN_FOLDERS
			if (!err) HandleFOLDER ( file );
		#else
			if (!err) HandleOPEN ( file );
		#endif
	}
	AEDisposeDesc(&docList);
	return err;
}


static pascal OSErr _handle_PRINT(AppleEvent *eventIn, AppleEvent */*eventOut*/, long /*refCon*/)
{	FSSpec	file;
	AEDescList	docList;
	long		index, itemsInList;
	Size		actualSize;
	AEKeyword	keywd;
	DescType	returnedType;
	OSErr 	err;
	
	err = AEGetParamDesc ( eventIn, keyDirectObject, typeAEList, &docList );
	if (err) return err; 
	err = _got_required_params ( eventIn );		
	if (!err) err = AECountItems ( &docList, &itemsInList );
	for (index=1; index<=itemsInList && !err; index++) 
	{	err = AEGetNthPtr ( &docList, index, typeFSS, &keywd, &returnedType, (Ptr)&file, sizeof(file), &actualSize );
		if (!err) HandlePRINT ( file );
	}
	AEDisposeDesc(&docList);
	return err;
}


void InitAppleEvents()
{	
	if ( AEInstallEventHandler ( kCoreEventClass,'oapp',NewAEEventHandlerProc(_handle_NEW),0L,false ) == noErr )
	if ( AEInstallEventHandler ( kCoreEventClass,'odoc',NewAEEventHandlerProc(_handle_OPEN),0L,false ) == noErr )
	if ( AEInstallEventHandler ( kCoreEventClass,'pdoc',NewAEEventHandlerProc(_handle_PRINT),0L,false ) == noErr )
	if ( AEInstallEventHandler ( kCoreEventClass,'quit',NewAEEventHandlerProc(_handle_QUIT),0L,false ) == noErr )
		return;
	DoAbort ( "\pApple event handlers could not be installed" );
}




