
/*	CONSOLE Window Handler

***	Copyright:	1993-1996 Gnter Woigk
  				mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	22.Nov.93:	Start of work on "console.c"								KIO !
	05.Feb.95:	now allocating text buffers at run time						KIO !
	18.Nov.95:	ConPrint() now does word wrap								KIO !
	21.apr.96:	auto crsr off on print										KIO !
	22.apr.96:	started work on "terminal.c" 								KIO !

TO DO:
	automatic tagging of 'console window' item in 'windows' menu
	querry cursor position
	scroll window to cursor position
	rows above row 0 = history
	character attributes
	print page
	scroll_up / scroll_down: korrekte berechnung der vacated area
		-> z.zt. wirk korrekter rfsh verhindert -> evtl. kein scrollbar rfsh und weie linien
		-> scrollen z.zt. zu langsam
*/

#include	<string.h>

#include	"kio.h"
#include	"options.h"
#include	"strings.h"
#include	"terminal.h"

#define		FASTPRINT	true

short		LINEHEIGHT;		// y-offset per line
short		CHARWIDTH;		// width of characters
short		ASCENT;
short		DESCENT;

#define		LOCATEPEN		MoveTo ( BORDER+CHARWIDTH*col, BORDER+LINEHEIGHT*row+ASCENT )


//	Mit dem ersten Aufruf einer Terminal-Funktion wird das Terminal-Fenster
//	automatisch geffnet. Fr den Window-Refresh wird line[] benutzt.

MenuHandle	terminalmenu	= nil;	// if supplied by main app, the console window item will be tagged automatically
short		terminalitem	= 0;

Boolean		terminalopen	= 0;	// use old line buffer contents or clear them?

WindowPtr	terminalwindow	= nil;
Rect		terminalrect	= { 50,50,360+50,440+50 };

string		line[MAXLINES];		// used for refresh

short		col,row;			// cursor location (character coordinates)
short		cols,rows;			// text columns & rows fitting in the window
short		oldwidth, oldheight;// for zoom box
short		oldhpos, oldvpos;	//  "   "    "
Boolean		reverse_video;
Boolean		crsr;				// is the cursor blob displayed ?

Str15		escseq;				// pending escape sequence

RgnHandle	termTempRgn = nil;


// ----	UP: show crsr blob ----------------------------------------
crsr_on ();
crsr_on ()
{
	LOCATEPEN;

	TextMode(notSrcCopy);
	if (col>=line[row].len)	DrawChar ( ' ' );
	else					DrawChar ( line[row].text[col] );
	TextMode(srcCopy);

	crsr=on;
}


// ----	UP: remove crsr blob ----------------------------------------
crsr_off ();
crsr_off ()
{
	LOCATEPEN;
	if (col>=line[row].len)	DrawChar ( ' ' );
	else					DrawChar ( line[row].text[col] );
	crsr=off;
}


// ---- show/hide cursor blob -------------------------------
void TermCrsr(Boolean f)
{
	if (!terminalwindow) TermOpen();
	SetPort(terminalwindow);

	if ( crsr==f ) return;
	if (f) crsr_on(); else crsr_off();
}


// ----- Open Console Window ---------------------------
void TermOpen()
{	FontInfo tfi;
	short    i;
	
	if (terminalwindow) return;

	if (!termTempRgn) termTempRgn=NewRgn();
	if (!termTempRgn) abort_oomem("\pTermOpen()");
	
	terminalwindow = NewWindow ( nil, &terminalrect, TERMNAME, false, zoomDocProc, (WindowPtr)-1, true, 0 );
	if (!terminalwindow) abort_oomem("\pTermOpen()");
	if (terminalmenu) CheckItem ( terminalmenu, terminalitem, true );

	SetPort ( terminalwindow );

	GetFNum(FONTNAME,&i);
	TextFont(i);
	TextMode(srcCopy);
	TextSize(CHARHEIGHT);
	TextFace(0);
	
	GetFontInfo (&tfi);
	ASCENT     = tfi.ascent;
	DESCENT	   = tfi.descent;
	LINEHEIGHT = tfi.ascent+tfi.descent+tfi.leading;
//	CHARWIDTH  = tfi.widMax;	!!! echt tolle funktion ... 
	CHARWIDTH  = ( TextWidth("\pTextbeispiel",0,12)+12-1 ) / 12;

	cols = MAXCOLUMNS;
	rows = MAXLINES;

	if (!terminalopen)	// clear screen on first call
	{	for (i=MAXLINES;i--;) line[i] = nullString;
		col = row = 0;
		*escseq = 0;
		oldwidth  = cols*CHARWIDTH+2*BORDER;
		oldheight = rows*LINEHEIGHT+2*BORDER;
		oldhpos = oldvpos = 0;
		reverse_video = false;
		terminalopen = true;
	}

	if (reverse_video) { ForeColor(whiteColor); BackColor(blackColor); }

	SetOrigin(0,0);
	SizeWindow   ( terminalwindow, cols*CHARWIDTH+2*BORDER, rows*LINEHEIGHT+2*BORDER, false );
	ResizeWindow ( terminalwindow, oldhpos, oldvpos, oldwidth, oldheight );
	DisplayWindow( terminalwindow );
	ResizeWindow ( terminalwindow, oldhpos, oldvpos, oldwidth, oldheight );
	SelectWindow ( terminalwindow );
	//InvalRect	 (&terminalwindow->portRect );
	//DrawSlidebars( terminalwindow );
	crsr=off;
}


// ----- Close Console Window --------------------------------
void TermClose ()
{	Rect box;

	if (terminalmenu) CheckItem ( terminalmenu, terminalitem, false );

	if (terminalwindow) 
	{
		box=terminalwindow->portRect;
		oldwidth  = box.right-box.left;
		oldheight = box.bottom-box.top;
		oldhpos   = box.left;
		oldvpos   = box.top;

		terminalrect=terminalwindow->portRect;
		SetPort(terminalwindow);
		LocalToGlobal ( (Point*)&terminalrect ); 
		LocalToGlobal (((Point*)&terminalrect)+1 ); 
		DisposeWindow ( terminalwindow );
		terminalwindow=nil;		
	};
}


// ----	clip trailing spaces of actual row --------------------------
//		called every time before the cursor leaves actual row
void clip_tail ();
void clip_tail ()
{	long i; Char *c;

#if FASTPRINT
	TermFastRfsh();
#endif	

	i=line[row].len;
	c=line[row].text+i-1;
	
	if (i && *c<=32)
	{	while (i && *c<=32) { i--; c--; };
		MoveMaster ( &line[row], LeftStr(line[row],i) );
	}
}


// ----	prepare any text output and control codes -------------
void set_port();
void set_port()
{
	if (!terminalwindow) TermOpen();
	SetPort ( terminalwindow );
	if (crsr) crsr_off();
}


// ----	clear range of lines ----------------------------------
//		range: [a ... [e
void clear_lines ( long a, long e );
void clear_lines ( long a, long e )
{	Rect box;

	SetRect (&box, 0, BORDER+a*LINEHEIGHT, 0x7fff, BORDER+e*LINEHEIGHT);
	while (a<e) line[a++]=nullString;
	EraseRect(&box);
	ValidRect(&box);
}


// ----	scroll range of screen up one line -------------------------
//		scrolls range [a ... [e; inserts blank line at bottom
void scroll_up ( long a, long e );
void scroll_up ( long a, long e )
{	Rect box;

	clip_tail();

	SetRect ( &box, 0, BORDER+a*LINEHEIGHT, 0x7fff, BORDER+e*LINEHEIGHT );
	ScrollWindowRect (terminalwindow, &box, 0, -LINEHEIGHT);

	while (++a<e) MoveMaster ( &line[a-1], line[a] );
	line[e-1] = nullString;
}


// ---- scroll range of screen down one line --------------------------
//		scrolls range [a ... [e; inserts blank line at top
void scroll_down ( long a, long e );
void scroll_down ( long a, long e )
{	Rect box;

	clip_tail();

	SetRect ( &box, 0, BORDER+a*LINEHEIGHT, 0x7fff, BORDER+e*LINEHEIGHT );
	ScrollWindowRect (terminalwindow, &box, 0, LINEHEIGHT);

	while (--e>a) MoveMaster ( &line[e], line[e-1] );
	line[a] = nullString;
}


// ----	clear to end of line -----------------------------------------------
void clear_eol();
void clear_eol()
{	Rect box;

	if (col<line[row].len)
	{	SetRect (&box, BORDER+col*CHARWIDTH, BORDER+row*LINEHEIGHT, 0x7fff, BORDER+(row+1)*LINEHEIGHT);
		EraseRect(&box);
		ValidRect(&box);
		memset ( &line[row].text[col], ' ', line[row].len-col );
	}
}


// ----	insert space at cursor position -------------------
void insert_char();
void insert_char()
{	string *s;

	s = &line[row];
	if (col<s->len)
	{	MoveMaster ( s, CatStr(QLeftStr(*s,col),CatStr(CharStr(' '),QSubStr(*s,col,s->len))) );
		LOCATEPEN;
		DrawText (s->text, col, s->len-col);
	}
}


// ----	delete character at cursor position -------------------
void delete_char();
void delete_char()
{	string *s;

	s = &line[row];
	if (col<s->len)
	{	MoveMaster ( s, CatStr(QLeftStr(*s,col),QSubStr(*s,col+1,s->len)) );
		DrawText (s->text, col, s->len-col);
		DrawChar (' ');
	}
}


// ----	locate cursor HORIZONTALLY -----------------------------------------
//		cursor may be positioned *one* character outside the screen!
void TermHLocate(short h)
{
	set_port();
	col = minmax(0,h,cols);
}


// ----	locate cursor VERTICALLY -----------------------------------------
void TermVLocate(short v)
{
	set_port();
	clip_tail();
	row = minmax(0,v,rows-1);
}


// ----	locate cursor ------------------------------------------------------
void TermLocate (short z, short s)
{	
	TermVLocate(z);
	col = minmax(0,s,cols);
}


// ----	Clear Console window ----------------------------------
void TermCLS()
{
	set_port();
	row = col = 0;
	clear_lines(0,rows);
}

	
// ----	move cursor UP ---------------------------------------------------------
//		may scroll
void TermUp()
{	
	set_port();
	if (row>0) { clip_tail(); row--; }
	else scroll_down(0,rows);
}


// ----	move cursor DOWN -------------------------------------------------------
//		may scroll
void TermDown()
{	
	set_port();
	if (row+1<rows) { clip_tail(); row++; }
	else scroll_up(0,rows);
}


// ----- move cursor to begin of next line -------------------------------------
//		may scroll
void TermNL ()
{
	TermDown();
	col = 0;
}


// ----	move cursor LEFT -------------------------------------------------------
//		wraps to next line; may scroll
void TermLeft()
{	
	if (col>0) { set_port(); col--; }
	else { TermUp(); col=cols-1; }
}


// ----	move cursor RIGHT ----- wrap around right border -----------------------
//		wraps to next line; may scroll
//		special: cursor may stay *one* character outside the screen!
void TermRight()
{	
	if (col<cols) { set_port(); col++; }
	else { TermDown(); col=1; };
}


// ----	move cursor to next TAB position -----------------------------------
//		wraps to next line; may scroll
//		special: cursor may stay *one* character outside the screen!
void TermTab()
{	
	set_port();
	if (col>=cols) TermNL();
	col = (col+TABS)/TABS*TABS;
	if (col>cols) col=cols;
}


// ====	print a character or control code ============================
//		TV 950 emulation (not complete)
void TermChar ( Char c );
void TermChar ( Char c )
{
	set_port();
	if (*escseq) goto es;	// escape sequence pending
	if (c<32) goto cc;		// control code


// ----	print the character -------------------------------------------
pc:	if ( col>=cols ) TermNL();
	
	if (col>=line[row].len) AddString ( &line[row], SpaceStr(cols-line[row].len) );
	line[row].text[col] = c;

#if FASTPRINT
	{	Rect box;
		SetRect ( &box,0,0,CHARWIDTH,LINEHEIGHT );
		OffsetRect(&box,BORDER+CHARWIDTH*col, BORDER+LINEHEIGHT*row);
		InvalRect(&box);
	}
#else
	LOCATEPEN;
	DrawChar(c); 
#endif
	col++;
	return;
	

// ----	handle control code -------------------------------------------
cc:	switch (c)
	{
	case 7: // ^G Bell
		SysBeep(20);
		return;
	case 8:	// ^H Backspace (cursor left, wrap to previous line but don't scroll)
		if (col>0) col--; 
		else if (row>0) { clip_tail(); row--; col=cols-1; }
		return;
	case 9:	// ^I Tab; tab positions every 8 chars, wraps to next line, may scroll
		if (col>=cols) { col=TABS; goto lf; }
		col = (col+TABS)/TABS*TABS;
		if (col>cols) col=cols;
		return;
	case 10:	// ^J Linefeed (cursor down, may scroll)
lf:		if (row+1>=rows) scroll_up(0,rows); else { clip_tail(); row++; }
		return;
	case 11:	// ^K Cursor up, don't scroll
		if (row>0) { clip_tail(); row--; }
		return;
	case 12:	// ^L Cursor right, wraps, may scroll
		if (col++>cols) { col=1; goto lf; }
		return;
	case 13:	// ^M Carriage return: back to column 0
		col=0;
		return;
	case 0x16:	// ^V Cursor down (without scroll)
		if (row<rows) { clip_tail(); row++; }
		return;
	case 0x1a:	// ^Z Home cursor and clear page
		clear_lines(0,rows);
		col = row = 0;
		return;
	case 0x1b:	// ^[ ESCAPE									// ] ((for MW editor))
		escseq[0] = 1;
		escseq[1] = c;
		return;
	case 0x1e:	// ^^ Home cursor
		clip_tail();
		col = row = 0;
		return;
	default:
		return;
	}


// ----	handle pending escape sequence ------------------------------
es:	if (*escseq==1) 
	{
		*escseq = 0;	// assuming no parameters
		
		switch(c)
		{
		case '(':	// normal brightness	=>  needs attribute handling
			return;
		case ')':	// half bright			=>  needs attribute handling
			return;
		case '*':	// clear screen
		case '+':	// clear screen
		case ',':	// clear screen
		case ':':	// clear screen
		case ';':	// clear screen
			TermCLS();
			return;
		case '?':	// ask cursor position	=>  needs to be implemented 
			return;	
		case 'E':	// insert line
			scroll_down(row,rows);
			return;
		case 'I':	// back tab (wraps to previous line, doesn't scroll)
			if (col<=0) 
			{	if (row<=0) return;
				row--;
				col=cols;
			}
			col=(col-1)/TABS*TABS;
			return;
		case 'Q':	// insert char (space)
			insert_char();
			return;
		case 'R':	// delete line
			scroll_up(row,rows);
			return;
		case 'T':	// clear to end of line
		case 't':	// clear to end of line
			clear_eol();
			return;
		case 'W':	// delete char (rest of line moves left)
			delete_char();
			return;
		case 'Y':	// clear to end of screen
		case 'y':	// clear to end of screen
			clear_eol();
			clear_lines(row+1,rows);
			return;
		case 'b':	// reverse video
			if (!reverse_video)
			{	ForeColor(whiteColor);
				BackColor(blackColor);
				InvalRect(&terminalwindow->portRect);
				reverse_video = true;
			};	return;
		case 'd':	// normal video
			if (reverse_video)
			{	ForeColor(blackColor);
				BackColor(whiteColor);
				InvalRect(&terminalwindow->portRect);
				reverse_video = false;
			};	return;
		case 'j':	// reverse linefeed (may scroll)
			if (row>0)	{ clip_tail(); row--; }
			else		scroll_down(0,rows);
			return;
		case 'P':	// print page  ( not yet implemented )
			return;
	
		case '=':	// locate curser to x,y
		case 'G':	// set video attribute to n
		case '.':	// set cursor attribute
		case 'U':	// control mode on
			escseq[0] = 2;
			escseq[2] = c;	// store cmd byte
			return;
	
		default:	// faulty esc
			return;
		}
	}

	switch(escseq[2])
	{
	case 'U':	// control mode active! (switch off via setup menu / reset button!)
		goto pc;	// print character and keep esc-sequence active
	case '=':	// locate curser to x,y
		if (escseq[0]==2) { escseq[++*escseq] = c; return; }
		TermLocate(escseq[3]+0x20,c+0x20);
		break;
	case 'G':	// set video attribute to n  =>  needs attribute handling
		break;
	case '.':	// set cursor attribute: not supported
		break;
	}
	*escseq = 0;
}


// ----	print a string --------------------------------------------------
//		no control codes
//		may wrap and may scroll
void term_type ( string s );
void term_type ( string s )
{	long i;

	while (col+s.len>cols) 
	{	i = cols-col;
		term_type(QLeftStr(s,i));
		TermNL();
		s = QSubStr(s,i,s.len);
	}
	
	if (s.len>0)
	{	if (col+s.len>line[row].len) AddString ( &line[row], SpaceStr(cols-line[row].len) );
		memcpy ( &line[row].text[col],s.text,s.len );
		LOCATEPEN;
		DrawText (s.text, 0, s.len);
		col+=s.len;
	}
}


void term_wrap ( string s );
void term_wrap ( string s )
{	long i;

a:	if (col+s.len>cols)
	{	for ( i=cols-col-1; i>=0; i-- )
		{	if (s.text[i]==' ')
			{	term_type(QLeftStr(s,i));
				TermNL();
				s = QSubStr(s,i+1,s.len);
				goto a;
			}
		};
		if (cols>0)
		{	TermNL();
			goto a;
		}
		term_type(QLeftStr(s,cols));
		TermNL();
		s = QSubStr(s,i,s.len);
		goto a;
	}
	term_type(s);
}


// ----	Print a space ------------------------------------------------------
//		intended to print a word-separating space
//		therefore the space is ignored if cursor is at right border
void TermSpace ( )
{
	set_port();
	term_wrap(CharStr(' '));
//	if (col<cols) TermChar(' ');
}


// ----- print a string ------------------------------------------------
void TermPrint ( string s )
{	long i;

	set_port();

tp:	if (*escseq)					// complete pending esc-sequence
	{	for (i=0; *escseq&&i<s.len; i++) TermChar(s.text[i]);
		s = QSubStr(s,i,s.len);
	}

	for (i=0; i<s.len; i++) 		// handle control codes
	{	if (s.text[i]>=32) continue;
		term_type(QLeftStr(s,i));
		TermChar(s.text[i]);
		s = QSubStr(s,i+1,s.len);
		goto tp;
	}

	term_type(s);
}


// ----- Print text with word wrap ------------------------------------------
void TermWrap ( string s )
{	long i;

	set_port();

tp:	if (*escseq)					// complete pending esc-sequence
	{	for (i=0; *escseq&&i<s.len; i++) TermChar(s.text[i]);
		s = QSubStr(s,i,s.len);
	}

	for (i=0; i<s.len; i++) 		// handle control codes
	{	if (s.text[i]>=32) continue;
		term_wrap(QLeftStr(s,i));
		TermChar(s.text[i]);
		s = QSubStr(s,i+1,s.len);
		goto tp;
	}

	term_wrap(s);
}


// ----- refresh screen -----------------------------------------------------
void term_rfsh ();
void term_rfsh ()
{	long i,y;
	Rect box;

	box = terminalwindow->portRect; box.right=BORDER; EraseRect(&box);
	box = terminalwindow->portRect; box.bottom=BORDER; EraseRect(&box);
	box = terminalwindow->portRect; box.top=BORDER+rows*LINEHEIGHT; EraseRect(&box);

	for ( i=0,y=BORDER; i<rows; i++,y+=LINEHEIGHT )
	{	SetRect ( &box, BORDER+line[i].len*CHARWIDTH, y, 0x7fff, y+ASCENT+DESCENT );
		EraseRect(&box);
		SetRect ( &box, 0, y+ASCENT+DESCENT, 0x7fff, y+LINEHEIGHT );
		EraseRect(&box);
		MoveTo ( BORDER,y+ASCENT );
		DrawText ( line[i].text,0,line[i].len );
	}
	if(crsr) crsr_on();
}


void TermRfsh ()
{
	if (!terminalwindow) return;			// this should never happen
	SetPort ( terminalwindow );				
	//EraseRect(&terminalwindow->portRect);	// in case background is black !!!
	term_rfsh();
}


void TermFastRfsh ()
{	
	if (!terminalwindow) return;			// this should never happen
	SetPort ( terminalwindow );				

	BeginUpdate		(terminalwindow);
	term_rfsh		();
//	UpdtControl		(terminalwindow,terminalwindow->visRgn);	there are no controls in terminal window
//	DrawSlidebars	(terminalwindow);							fast refresh needs not take care for hsb/vsb
	EndUpdate   	(terminalwindow); 
}


void TermZoom()
{	Rect box;
	ControlHandle hsb,vsb;

	hsb=FindRefControl(terminalwindow,hsb_magic);
	vsb=FindRefControl(terminalwindow,vsb_magic);
	
	box=terminalwindow->portRect;
	if ( !hsb && !vsb )					// was full size
	{	ResizeWindow ( terminalwindow,oldhpos, oldvpos, oldwidth,oldheight );
		DisplayWindow( terminalwindow );
	}
	else								// was shrinked
	{	oldwidth  = box.right-box.left;
		oldheight = box.bottom-box.top;
		oldhpos   = box.left;
		oldvpos   = box.top;
		ResizeWindow ( terminalwindow, 0, 0, cols*CHARWIDTH+2*BORDER, rows*LINEHEIGHT+2*BORDER );
	}
}




