/*	Z80 Emulator: header file

***	Copyright:	1995-1999 Gnter Woigk
  				mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#include	"z80.options"

#if CMD_PROFILE
extern	Boolean	count_instr;
extern	long		*cnt_xx;
extern	long		*cnt_cb;
extern	long		*cnt_ed;
extern	long		*cnt_xy;
extern	long		*cnt_xycb;
#endif

#if PC_PROFILE
extern	Boolean	count_pc;
extern	long		*cnt_pc;
#endif


// -----	Accessing the Core ---------------------------------------------------------

extern	Char*	RdPtr	( Short addr );			// get pointer for reading
extern	Char*	WrPtr	( Short addr );			// get pointer for writing

extern	Char		Peek	( Short addr );				// read byte
extern			Poke	( Short addr, Char byte );		// write byte


// -----	Block copying inside/from/to core -------------------------------------------

extern	long		tfu;			// bytes actually transferred  if  OSErr!=ok
extern	OSErr	Buffer2File	( void* bu,	short fileID,	long len );
extern	OSErr	File2Buffer	( short fileID,	void* bu,		long len );
extern	OSErr	Core2File		( Short addr,	short fileID,	long len );
extern	OSErr	File2Core		( short fileID,	Short addr,	long len );
extern	void		Core2Buffer	( Short addr,	void* bu,		long len );
extern	void		Buffer2Core	( void* bu,	Short addr,	long len );
extern	void		Core2Core	( Short src,	Short dest,	Short len );
extern	void		Do_LDIR		( Short src,	Short dest,	Short len );
extern	void		Do_LDDR		( Short src,	Short dest,	Short len );
extern	void		Buffer2Buffer	( void* src,	void* dest,	long len );


// -----	Flag tables ----------------------------------------------------------------

extern	Char		zlog_flags[256];	// convert:   A register  ->  z80 flags with V=parity and C=0
#if !GENERATINGPOWERPC
extern	Char		mlog_flags[256];	// convert:   A register  ->  m68 flags with V=parity and C=0
extern	Char		z80flags[256];		// convert:   m68 flag byte  ->  Z80 flag byte
extern	Char		m68flags[256];	// convert:   Z80 flag byte  ->  m68 flag byte
#endif


// -----	The Z80 core --------------------------------------------------------------
extern	Char**	CORE_M;			// master handle for all allocated RAM and ROM
extern	Char*	CORE;			// do not use any longer!!!
extern	Char*	rpage[];			// mapped in pages for reading
extern	Char*	wpage[];			// mapped in pages for writing
extern	Char*	ram[];			// real RAM pages
extern	Char*	rom[];			// real ROM pages
extern	Char*	norom;			// dummy page for writing to ROM


// -----	Z80 registers on entry and return of Z80() -----------------------------------

typedef union pair pair;
union pair
{	Short	rr;
	struct { Char hi,lo; } r;
};

typedef struct z80 z80;
struct z80
{	pair		bc,de,hl,ix,iy;			// registers	(z80() -> local struct)
	pair		pc,sp;				// registers	(z80()/z80_ppc() -> registers)
	pair		bc2,de2,hl2;			// registers
	pair		af,af2;				// registers	(a,f -> register;  z80(): a2,f2 -> register)
	pair		iic;					// register i + hibyte from bus
	pair		iff;					// interrupt enable flip flops
	Char		exit,wuff;				// watchdog flag & nmi/irpt flags
	Char		r, im;				// refresh counter & interrupt mode: 0 ... 2
	long		cycles;				// processor T states (count down for interrupt)
	long		total;				// T states since start of Z80 (overflows approx. every 20')
};

extern	z80		zreg;			// all z80 registers, flags and bits are stored in this struct


// -----	definitions to ease the use of the z80 registers & core:

#define	CYCLES	zreg.cycles		// processor T states (count down for interrupt)
#define	TOTAL	zreg.total			// T states since start of Z80 (overflows approx. every 20')
#define	WUFF	zreg.wuff			// watchdog flag
#define	EXIT		zreg.exit			// nmi & irpt flags
#define	IFF		zreg.iff.rr
#define	IFF1		zreg.iff.r.hi		// actually used irpt flip flop
#define	IFF2		zreg.iff.r.lo		// copy of iff1 during nmi processing
#define	IM		zreg.im			// interrupt mode: 0 ... 2
#define	RR		zreg.r			// 7 bit DRAM refresh counter
#define	RI		zreg.iic.r.hi		// hi byte of interrupt vector: i register
#define	IRPTCMD	zreg.iic.r.lo		// lo byte of interrupt vector: read from bus
#define	IRPTVEK	zreg.iic.rr			// interrupt vector in interrupt mode 2

#define	RA		zreg.af.r.hi		// Z80() uses register variable 'a'
#define	RF		zreg.af.r.lo		// Z80() uses register variable 'f'
#define	RA2		zreg.af2.r.hi
#define	RF2		zreg.af2.r.lo

#define	BC2		zreg.bc2.rr
#define	RB2		zreg.bc2.r.hi
#define	RC2		zreg.bc2.r.lo

#define	DE2		zreg.de2.rr
#define	RD2		zreg.de2.r.hi
#define	RE2		zreg.de2.r.lo

#define	HL2		zreg.hl2.rr
#define	RH2		zreg.hl2.r.hi
#define	RL2		zreg.hl2.r.lo

#define	BC		zreg.bc.rr
#define	RB		zreg.bc.r.hi
#define	RC		zreg.bc.r.lo

#define	DE		zreg.de.rr
#define	RD		zreg.de.r.hi
#define	RE		zreg.de.r.lo

#define	HL		zreg.hl.rr
#define	RH		zreg.hl.r.hi
#define	RL		zreg.hl.r.lo

#define	IX		zreg.ix.rr
#define	XH		zreg.ix.r.hi
#define	XL		zreg.ix.r.lo

#define	IY		zreg.iy.rr
#define	YH		zreg.iy.r.hi
#define	YL		zreg.iy.r.lo

#define	PC		zreg.pc.rr
#define	PCH		zreg.pc.r.hi
#define	PCL		zreg.pc.r.lo

#define	SP		zreg.sp.rr
#define	SPH		zreg.sp.r.hi
#define	SPL		zreg.sp.r.lo

#define	disabled		0x00		// irpt flags
#define	enabled		0xFF			// irpt flags

// bits in WUFF:
#define	is_nmi		0x80		// non maskable interrupt: handled inside Z80()
#define	is_irpt		0x7F			// normal interrupt counter: handled inside Z80()

// -----	Return values of Z80() -----------------------------------------------------

#define	watchdog_irpt	0			// watchdog exception
#define	nimp_instr	1			// not implemented instruction at pc -> stop engine
#define	halt_instr		2			// halt instruction encountered at pc-1 -> execute it!
#define	rst0_instr	3			// rst 0 instruction encountered at pc-1 -> execute it!
#define	irpt_error	4			// not supported interrupt mode/instruction -> stop engine
#define	ill_instr		5			// Z80_Info_Illegal() requested Z80() to exit
#define	weird_instr	6			// Z80_Info_Weird() requested Z80() to exit

// -----	Procedures -------------------------------------------------------------------

extern	Init_Z80 ( long roms, long rams );			// allocate core, init z80
extern	short	Z80 ( );					// same as Z80() but with T cycle counting
extern	short	Z80_PPC ( );				// same as Z80_P() but written in C


// -----	The following procedures must be supplied by the application program ---------------

extern	Do_Output		( Short addr, Char n );	// Output byte to port
Char		Do_Input 		( Short addr );			// Input byte from port
extern	void Do_Cycles		( void );				// Count down for T cycles reached 0

extern	Z80_Info			( long cc, Short ip );	// info call
Boolean	Z80_Info_Illegal	( long cc, Short ip );	// info call; may request Z80() to exit by returning 'true'
Boolean	Z80_Info_Weird	( long cc, Short ip );	// info call; may request Z80() to exit by returning 'true'
extern	Z80_Info_Irpt		( long cc );		// info call to debugger if INFO_IRPT==on
extern	Z80_Info_NMI		( long cc );		// info call to debugger if INFO_NMI==on
extern	Z80_1st_Loc		( long cc, Short ip );	// info call to debugger if PC_PROFILE==on
extern	Z80_1st_Instr		( long cc, Short ip );	// info call to debugger if CMD_PROFILE==on

