/*	KIO's standard header file.

***	Copyright:	1994-1999 Günter Woigk
  				mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

    20.Oct.01   revised for CW7Pro and ANSI C compliance									hdusel@bnv-gz.de
*/

#ifndef kio_h
#define kio_h


// ---- Data type definitions ---------------------------------------------
typedef unsigned char	Char;
typedef	unsigned short	Short;
typedef	unsigned long	Long;
typedef unsigned char	uchar;
typedef	unsigned short	ushort;
typedef	unsigned long	ulong;
typedef struct string string;
struct string { long len; Char* text; };


// ----	Constants ---------------------------------------------------------
#define	true	1
#define	false 	0
#define	yes		1
#define	no		0
#define	on		1
#define	off		0


// ----	Error numbers -----------------------------------------------------
#define	ok		0				// no error
#define	noerror	0				// no error
#define error	-1				// general error
#define	oomem	'kx01'			// memory overflow
#define	oorange	'kx02'			// argument out of range (general)
#define	badref	'kx03'			// Handle, pointer, or reference number is illegal
#define badsize	'kx04'			// argument out of range (size)
#define	badidx	'kx05'			// Array index out of range
#define	empty	'kx06'			// tried to retrieve element from empty list/queue/stack
#define nan		'kx07'			// not a number
#define toobig	'kx08'			// number too big


#include	"options.h"


// ----	Nice defines -------------------------------------------------------
#define	MIN(a,b)	((a)<(b)?(a):(b))
#define	MAX(a,b)	((a)>(b)?(a):(b))
#define	SWAP(a,b)	((a)^=(b),(b)^=(a),(a)^=(b))		// © (c) KIO !
#define	SIGN(a)		(((a)>0)-((a)<0))					// © (c) KIO !
#define	ABS(a)		((a)<0?-(a):(a))
#define	SORT(a,b)	if ((a)<=(b)) {} else SWAP(a,b)
#define	NELEM(feld)	(sizeof(feld)/sizeof(feld[0]))		// UNSIGNED !!!
#define	BIT(B,N)	(((N)>>(B))&1)
#define	RMASK(B)	(((ulong)(-1))>>(32-(B)))
#define	LIMIT(A,N,E)if (N<(A)) N=(A); else if (N<=(E)) {} else N=(E)


// ----	Calculate base 2 logarithm -----------------------------------------
#define LOG2BYTE(N)	((N)>127+(N)>63+(N)>31+(N)>15+(N)>7+(N)>3+(N)>1)
#define LOG2SHORT(N)(unsigned(N)>255 ? LOG2BYTE(unsigned(N)>>8)+8 : LOG2BYTE(N))
#define LOG2(N)		(unsigned(N)>0x7fff ? LOG2SHORT(unsigned(N)>>16)+16 : LOG2SHORT(N))

inline long	Min		( long a, long b )			{ return a<b ? a : b; }
inline long	Max		( long a, long b )			{ return a>b ? a : b; }	
inline long	MinMax	( long a, long n, long e )	{ return n<=e ? (n>=a?n:a) : e; }

inline int  Sign	( long a )					{ return SIGN(a); }
inline long	Abs		( long a )					{ return ABS(a); }

inline ulong Scale	( ulong oldval, ulong oldmax, ulong newmax )	// size conversion ((caveat: oldmax*newmax <= 0xFFFFFFFF))
					{ return ( oldval*newmax+oldmax/2 ) / oldmax; }

// ----	Nice defines -------------------------------------------------------
#define min(a,b)		( (a)<(b) ? (a) : (b) )
#define max(a,b)		( (a)>(b) ? (a) : (b) )
#define	minmax(a,n,e)	( (n)<=(e) ? ((n)>=(a)?(n):(a)) : (e) )

#define do_alert(s,t,u) { ParamText(s,t,u,0); Alert(666,nil); }
#define	do_abort(s,t,u)	{ do_alert(s,t,u); ExitToShell(); }
#define	abort_oomem(s)	{ do_abort("\pOut of memory:",s,0); }

inline int do_confirm(Str255 s, Str255 t)
{ 
    ParamText(s,t,0,0); 
    return Alert(669,nil); 
}


// ====	Standard functions & associated data ===============================


// ----	checkgestalt.c -----------------------------------------------------
extern	Boolean 		powerMgr;			// Power manager present?
void InitAppl ( );
void CheckGestalt ( );
extern	void InitAppleEvents	( );
void WaitCursor ( );
void ArrowCursor ( );

// ----	pollevents.c -------------------------------------------------------
extern	Boolean PollEvents ( short evMask, long evDelay, RgnHandle evRegion );
extern	EventRecord		event;
extern	Boolean 		done;				// Quit application?
extern	ControlHandle 	FindRefControl ( WindowPtr window, long refValue );
extern	void KIOResizeWindow ( WindowPtr window, short x, short y, short w, short h );
extern	void KIOScrollWindowRect ( WindowPtr window, Rect *box, short dx, short dy );
extern	void HandleFASTUPDATE(WindowPtr);
extern	void HandleBUTTON(WindowPtr,ControlHandle);
extern	void HandleCHECKBOX(WindowPtr,ControlHandle);
extern	void HandleSLIDEBAR(WindowPtr,ControlHandle,short);
#define	hsb_magic	938573995
#define	vsb_magic	937673997


// these procedures must be provided somewhere in your application:
extern	void HandleOPEN (FSSpec);
extern	void HandlePRINT (FSSpec);
extern	void HandleNEW(void);
extern	void HandleDISK(void);
extern	void HandleDRIVER(void);
extern	void HandleAPP3(void);
extern	void HandleKEYDOWN (Char);
extern	void HandleAUTOKEY (Char);
extern	void HandleKEYUP (Char);
extern	void HandleMENU (short, short);
extern	void HandleCLOSE (WindowPtr);
extern	void HandleGROW (WindowPtr);
extern	void HandleZOOM (WindowPtr);
extern	void HandleDRAG (WindowPtr, Point p);
extern	void HandleINCONTENT (WindowPtr);
extern	void HandleMOUSEUP(void);
extern	void HandleACTIVATE (WindowPtr);
extern	void HandleDEACTIVATE (WindowPtr);
extern	void HandleUPDATE (WindowPtr);
extern	void HandleRESUME(void);
extern	void HandleSUSPEND(void);
extern	void HandleREGION(void);
extern	void HandleNULL(void);


// ----	windowutilities.c --------------------------------------------------
extern	void DisplayWindow ( WindowPtr w );




#endif
