
/*
***	Copyright:	1996 Günter Woigk
  				mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

/*	General layout of a movie file:

	one		movie header
	multiple	sequences
	one final	prev_block_len

*/


#include	"options.h"
#if include_movie


// every movie file starts with this header:

typedef struct movhead movhead;
struct movhead
{	long		file_id;						// 'ZMov'		file type identification
	long		appl_id;						// 'ZXSP'		creator
	long		appl_ver;						// '1.90'		version
	long		z80_ver;						// '1c80'		z80 engine version; 'a'ss or 'c'
	long		frames;						// total length in 1/50 sec frames
	char		hardware;						// zx spectrum model
	char		rom;							// rom information
	char		patches;						// applied patches
	char		joystick;						// emulated joystick
	char		sound;						// emulated sound hardware
	char		scr_size;						// screen size (informational only)
	char		options;						// exact r register emulation, highres mode, etc.
	char		filler[8];						// further enhancements...
};


// every sequence starts with this header:

typedef struct seqhead seqhead;
struct seqhead
{	long		prev_block_len;				// length of previous sequence; used for backward stepping
	long		block_len;						// length of this sequence; used for forward stepping
	long		frame;						// starting frame number of this sequence
	short	frames;						// number of frames contained in this sequence
	Char		border;						// initial value of border color register
	Char		bank;						// initial state of bank switching register; $FF on zx16/zx48
	Char		xcore;						// flag: is core dump xored with prev. core dump?
	Char		dummy;
	z80		zreg;						// z80 registers, irptcmd, exit, wuff, cycles, total
};



typedef struct aystate aystate;
struct aystate
{	short	regnr;				// selected ay data register
	Char		reg[16];				// values of the ay data registers
};


// ------------------------------------------------------------
// constants for the movie header:

#define		file_idf		'ZMov'		// file type indiacator (always 'ZMov')
#define		appl_idf		'ZXSP'		// creator application ('ZXSP' is Mac Spectacle)
#define		appl_vers		'1.90'		// application's version number
#define		c_vers		'1c80'		// version of z80 engine; 'c' indicating a c coded engine
#define		a_vers		'1a70'		// version of z80 engine, 'a' indicating an assembler coded engine

#define		zx16			1			// hardware identification
#define		zx48			2
#define		zx128		3
#define		zxplus2		4

#define		rom_48		1			// used standard roms
#define		rom_1280	2
#define		rom_1281	4
#define		rom_if1		8			// if1 attached
#define		rom_ext		128			// external rom attached; included in first core dump!

#define		load_patch	1			// applied patches
#define		save_patch	2
#define		beep_patch	4

#define		KEMPSTON		1			// joystick information

#define		ay_sound		1			// sound information

#define		exact_r		1			// option bits
#define		highres		2			


// ------------------------------------------------------------
// constants for the sequence header:

#define		std_core		0x00		// core compression flag
#define		xored_core	0xff


// ------------------------------------------
// functions:


extern				OpenMovieWindow	( );
extern				CloseMovieWindow	( );
extern				RefreshMovieWindow( );
extern				ClickInMovieWindow	( );
extern				OpenMovie		( FSSpec file );
extern	OSErr		SaveMOVIE		( short fileID );
extern				DoMovieFrame		( );
extern				StartStopMovie		( Boolean running );


#endif



// ------------------------------------------
// data:

extern	WindowPtr	movie_window;
extern	Rect			movieRect;
extern	Boolean		movie_recording;




