
/* ---- String-Verwaltung für das FORCE System -----------------------------

***	Copyright:	1994-1999 Günter Woigk
  				mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.



	Strings werden normal nur über ihren Descriptor gehandhabt.
	Dieser besteht aus einem Zeiger auf den text (.text) und einer Längenangabe (.len)
	Grundsätzlich kann damit jede Zeichenkette im Speicher beschrieben werden,
	typischerweise werden die Zeichenketten jedoch im String-Pool reserviert.
	Der Stringpool wird durch den Aufruf von InitStr() initialisiert.

	Einträge im Pool sind jeweils fest mit einem Descriptor assoziiert. Wird der 
	Descriptor gelöscht oder verschoben, wird der Text im Pool freigegeben. Nur solange 
	der Descriptor gültig ist, bleibt auch der Text reserviert. Substrings aus dem Text
	können adressiert werden solange der Hauptdescriptor gültig ist.
	
	Stringoperationen liefern als Ergebnis wieder einen Stringdescriptor zurück. Dieser
	kann ein Hauptdescriptor sein, ein Subdescriptor oder ein Konstantendescriptor.
	Das ist solange kein Problem, wie kein Hauptdescriptor verändert wird.
	
	Hauptdescriptoren können verändert werden
	1. Durch Funktionen, die in Stringausdrücken aufgerufen werden.
			: killer ( -- str )	a$=""; return "haha!" ;
			: bugger ( -- )		a$ = a$(5 to) + killer()			

	2. Durch parallel laufende Tasks

	3. Zuordnungen mit Stringslicer auf der linken Seite bedürfen besonderer 
	   Beachtung, da hier dann nicht nur Müll gelesen, sondern ggf. falscher
	   Speicher überschrieben wird:
			
			: killer ( -- str )	a$=""; return "haha!" ;
			: bugger ( -- )		a$(0 to 5) = killer() ;

	Hier muß der Programmierer jeweils darauf achten, sich nicht selbst ein paar Fallen zu graben.
	
	4. Durch Überschreiben von Zwischenergebnissen, die einen Hauptdescriptor lieferten,
	   auf dem Heap. Dieser Fall muß vom Compiler erkannt werden. 

	Stringdescriptoren auf dem Heap können folgenden Status haben:
	   Masterdescriptor	 	z.B. nach AddStr ( str str2 -- str )
	   Subdescriptor		z.B. nach SubStr ( str a e -- str )
	   Konstantendescriptor	z.B. nach PushS  ( -- str )

	Sobald das Ergebnis in eine Variable gespeichert wird (nicht, wenn es 'verbraucht'
	wird, z.B. in einem Print-Kommando) muß sichergestellt werden, daß dies jetzt ein
	autarker Text mit eigenem Hauptdescriptor ist. Dies geschieht durch AssignString()


    20.Oct.01   revised for CW7Pro and ANSI C compliance									hdusel@bnv-gz.de
*/


extern	string	nullString;		// prototype for empty string: nullString.len = 0 and nullString.addr is not off-memory


// ----	pool handling ---------------------------------

extern	void	InitStrings	(long poolsize);			// init string module
extern	void	ClearQStrings(void);						// free all temporary strings
extern	string	AllocStr	( long len );				// TEMPORARY; allocate string in pool
extern	string	PoolString	( string q );				// TEMPORARY; move text into pool
extern	void	MoveMaster	( string *s, string q );	// MASTER; move a master string to a new location

#define			AssignString(Z,Q)	MoveMaster(Z,PoolString(Q))

// ----	tests -----------------------------------------

extern	Boolean	EmptyStr	( string s );				// string empty ((may contain white space!)
extern	Boolean	SameStr		( string s, string t );		// compare two strings for equality

// ----	conversions -----------------------------------

extern	string	CharStr		( Char c );					// CONSTANT; convert char to string
extern	string	PStr		( const Str255 p );			// CONSTANT; convert pascal-style text to string
extern	string	CStr		( char *c );				// CONSTANT; convert c-style text to string
extern	string	SpaceStr	( long n );					// TEMPORARY; create a string of spaces
extern	Char* 	PascalStr 	( string s );				// TEMPORARY; convert string back to pascal format

#define			CharString(C)		PoolString(CharStr(C))
#define			PString(P)			PoolString(PStr(P))
#define			CString(C)			PoolString(CStr(C))

// ----	string adding / clipping ----------------------

extern	void	AddString	( string *a, string b );	// MASTER; add string to string
extern	string	CatStr		( string a, string b );		// TEMPORARY; add string to string
extern	string	QSubStr		(string s, long a, long e);	// SLAVE; substring [a ... [e; indexes starting with 0
extern	string	QLeftStr	(string s, long n);
extern	string	QRightStr	(string s, long n);
extern	string	QClipStr	( string s );				// SLAVE; clip white spaces at both ends

#define			SubStr(S,A,E)	PoolString(QSubStr(S,A,E))	// TEMPORARY
#define			LeftStr(S,N)	PoolString(QLeftStr(S,N))	// TEMPORARY
#define			RightStr(S,N)	PoolString(QRightStr(S,N))	// TEMPORARY
#define			ClipStr(S,A,E)	PoolString(QClipStr(S,A,E))	// TEMPORARY

// ----	number conversion -----------------------------

extern	long	ValStr		( string s, long *err );	// evaluate number: "±decimal", "±$hex", "±%bin", "'text'"
extern	float	FValStr		( string s, long *err );	// evaluate number: "±dec.dec^dec"
extern	double	DValStr		( string s, long *err);		// evaluate number: "±dec.dec^dec"

extern	string	NumStr		( long n );					// TEMPORARY; convert integer to decimal string; auto-sized
extern	string	HexStr		( long n, long digits );	// TEMPORARY; convert integer to hex string
extern	string	BinStr		( long n, long digits );	// TEMPORARY; convert integer to binary string
extern	string	DNumStr		( double d, char *format );	// TEMPORARY; convert double float to decimal string

#define			FloatStr(N)		DNumStr(N,"%.8g")		// TEMPORARY; convert float to string
#define			DoubleStr(N)	DNumStr(N,"%.12g")		// TEMPORARY; convert double to string

// ----	miscellaneous ----------------------------------

extern	string	ReplaceStr	( string s, Char old, Char neu );
extern	string	LowerStr	( string s );
extern	string	UpperStr	( string s );
extern	short	SkipSpace	( string s, short i );
extern	short	SkipWord	( string s, short i );
extern	short	FindChar	( string s, Char c, short i );

// ----	Dialog boxes ------------------------------------
//const string Str00 = { 0,0 }; 
//extern	short 	DoAlert 	( short n, string s, string t, string u, string v );




