
/*	Movie recorder window handler

***	Copyright:	1994-1996 Günter Woigk
  				mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/


#include	"options.h"
#if include_movie



#include	<quickdraw.h>
#include	<windows.h>
#include	<string.h>

#include	"InfinityWindoid.h"
#include	"kio.h"
#include	"z80.options"
#include	"z80.h"
#include	"z80_opcodes.h"
#include	"z80 main.h"
#include	"strings.h"
#include	"preferences.h"
#include	"menue.h"
#include	"movie.h"
#include	"terminal.h"
#include	"application.h"
#include	"messages.h"
#include	"display.h"
#include	"keyboard.h"
#include	"buttons.h"
#include	"load+save.h"
#include	"sound.h"
#include	"freeze.h"
#include	"application.h"


// *********************************************************************************

#define		movie_wdef 	kInfinityProc
//#define		tape_wdef 	noGrowDocProc


#define		WIDTH		250				// tape recorder window width
#define		HEIGHT		102				// tape recorder window height (single recorder window)

#define		ANIM_DELAY	4				// video frames per animation frame
#define		SPOOL_DELAY	17				// movie_delay for NextBlock() and PrevBlock()

RGBColor		movie_bkgr_color	= { 23*257,36*257,82*257 };
RGBColor		movie_hilite_color	= { 0*257,187*257,0*257 };


btn			movie_button[7];				// info for 7 movie recorder buttons
enum { NEW,RECORD,FIRST,PREV,START,NEXT,LAST };


// -----	movie_button positions
#define		y_btn		73

#define		x_x			7
#define		x_o			x_x+36
#define		x_rewind		x_o+36
#define		x_prev		x_rewind+33
#define		x_start		x_prev+33
#define		x_next		x_start+33
#define		x_end		x_next+33

// -----	cicn IDs for normal icons on buttons
#define		i_eject		135				// cicn: insert / eject
#define		i_stop_go		155				// cicn: start tape / pause
#define		i_start		145
#define		i_pause		150
#define		i_rewind		160				// cicn: go to first block of tape
#define		i_end		165				// cicn: go to end of tape
#define		i_prev		170				// cicn: go to previous block of tape
#define		i_next		175				// cicn: go to next block of tape
#define		i_record		180				// cicn: recording on
#define		i_erase		190				// cicn: erase block
#define		i_copy		185				// cicn: copy block
#define		i_new_tape	195				// cicn: create new tape file
enum { grey,green,orange,red,yellow };			// offsets to get hilit versions


// -----	cicn IDs for released and pushed buttons
#define		i_button		130				// cicn: movie_button, released
#define		i_pbutton	 	131				// cicn: movie_button, pressed


// -----	cicn IDs & locations for 'camera' logo & animation
#define		i_camera		210				// cicn: camera; following: camera window animation 1,2,3,4
Rect			r_camera	= { 7,7, 7+32,7+32 };


// -----	cicn IDs & locations for LED tape counter
#define		i_led			400				// cicn: LED-"0", normal version; following: 1,2,3,4,5,6,7,8,9,off,-,1,1off,:,:off
#define		y_led		45				// vertical position
Rect			r_movie_led[5]		=  {	y_led,8,  	y_led+21,8+12,  
								y_led,24,	y_led+21,24+12,  
								y_led,37,	y_led+21,37+8,
								y_led,45,	y_led+21,45+12,
								y_led,61,	y_led+21,61+12	};

// -----	pict & location of slide bar indicator
#define		i_slide		133	 							// pict: slide bar indicator
Rect			r_movie_slide_o		= { 54,x_rewind, 54+12,WIDTH-7 };		// outer box (whole pict)
Rect			r_movie_slide_i		= { 59,x_rewind+5,59+2,WIDTH-12 };	// inner box (slider opening)
#define		w_slider		5								// slider width
PicHandle		movie_slide_pic		= nil;

// -----	file name, block info, popup box
#define		textsize		9								// text size of file name, block info and popup menu
#define		x_filename	80								// position of file name
#define		y_filename	(7+textsize)
Rect			r_movie_filename	= { 7,x_rewind, 9+textsize,WIDTH-7 };
#define		x_blockinfo	(x_rewind+3)						// position of block info
#define		y_blockinfo	(25+textsize)						// position of block info
Rect			r_movie_blockinfo	= { 23, x_rewind, 29+textsize, WIDTH-8 };


WindowPtr	movie_window	= nil;
Rect			movieRect		= { 0,0,HEIGHT,WIDTH};				// tape window rect in glob. coord (as used for prefs)

Rect			movie_rect	= { 0,0,HEIGHT,WIDTH };
Rect			r_movie_hi	= { 2,2,HEIGHT-3,WIDTH-3 };
Rect			r_movie_lo	= { 3,3,HEIGHT-2,WIDTH-2 };

short		anim_fram	= 0;


/***************************************************************
***************************************************************/


// -----	data data data ------------------------------------------------------

movhead		mov_head;			// head of movie file
seqhead		seq_head;				// head of currently selected/running sequence
aystate		ay_state;

Handle		h_buffers	= nil;		// temp mem for core[], compr[] and inputs[]
Char*		core;				// 1/3/8 ram pages and optionally 1 rom page (uncompressed) at start of sequence
Char*		compr;				// compression buffer
kbd*			inputs;				// keyboard/joystick inputs buffer; 12 bytes per frame
#define		MAX_FRAMES	500		// max. 10 sec. per sequence
#define		MAIN_FRAMES	30		// every XXth core is not xored


short		subframe;				// frame inside seqence
Boolean		movie_recording= false;

OSErr	 	movie_err	= ok;
short		movieID		= 0;
long			movie_len		= 0;
long			movie_pos	= 0;
Str63		movie_name;


/***************************************************************
***************************************************************/


// -----	animate camera icon ---------------------------------------------------------------
static animate_camera ()
{	long n;

	n = TickCount()+ANIM_DELAY;

	SetPort(movie_window);
	anim_fram = (anim_fram+1) % 8;
	PlotCIcon ( &r_camera, Get_CIcon(i_camera+anim_fram) );
	
	RefreshSpecciWindow();
	while ( (long)TickCount()-n<0 ) PollEvents ( updateMask, n-TickCount(), nil );
}


// -----	draw dimmed or highlighted anim_fram around controls -------------------------------------
static movie_draw_frame ( )
{	RGBColor rgb;

	SetPort(movie_window);
	
	ForeColor ( blackColor );
	FrameRect ( &r_movie_lo );

	rgb.red = movie_bkgr_color.red *2;
	rgb.green = movie_bkgr_color.green *2;
	rgb.blue = movie_bkgr_color.blue *2;
	RGBForeColor ( &rgb );
	FrameRect ( &r_movie_hi );
}


// -----	draw movie counter and slide bar indicator ---------------------------------------
static movie_draw_counter ( )
{	int i;
	Rect box;
	short dx;

	SetPort(movie_window);
	
	i = movieID ? (seq_head.frame+subframe)/50 : 0;
	PlotCIcon ( r_movie_led+0, Get_CIcon(i_led + i/600) );
	PlotCIcon ( r_movie_led+1, Get_CIcon(i_led + (i%600)/60) );
	PlotCIcon ( r_movie_led+2, Get_CIcon(415));
	PlotCIcon ( r_movie_led+3, Get_CIcon(i_led + (i%60)/10) );
	PlotCIcon ( r_movie_led+4, Get_CIcon(i_led + i%10) );

	if (movieID)
	{	
		dx = i * ((r_movie_slide_i.right-r_movie_slide_i.left)-w_slider)/max(mov_head.frames,1);

		ForeColor(blackColor);
		box = r_movie_slide_i; box.right = box.left+dx;
		PaintRect(&box);
		box = r_movie_slide_i; box.left += dx+w_slider;
		PaintRect(&box);

		ForeColor ( yellowColor );
		box = r_movie_slide_i; box.left += dx; box.right = box.left+5;
		PaintRect ( &box );
	}
	else
	{	ForeColor(blackColor);
		PaintRect(&r_movie_slide_i);
	}

}


// -----	draw block info, movie counter and slide bar indicator --------------------------
static movie_draw_blockinfo ( )
{	Char ns[12];
	

	SetPort(movie_window);
	
	ForeColor ( whiteColor );
	EraseRect ( &r_movie_blockinfo );
	FrameRect ( &r_movie_blockinfo );	

	movie_draw_counter( );
	
	if (!movieID) return;

	ForeColor ( whiteColor );
	MoveTo ( x_blockinfo,y_blockinfo );

	if ( movie_pos>=movie_len )
	{	if (movie_pos==0)	DrawString ( localized ("\pEmpty movie","\pLeeres Video") );		else
						DrawString ( localized ("\pEnd of movie","\pVideo-Ende") );
		return;
	}

	DrawString ( localized("\p((block info))", "\p((block info))") );
	return;

xx:	DrawString ( localized("\p••• File error: ", "\p••• Dateifehler: ") );
	NumToString(movie_err,ns);
	DrawString(ns);
}


// -----	draw file movie_name, block info, movie counter and slide bar indicator ------------------
static movie_draw_filename ( )
{	
	SetPort(movie_window);
	
	EraseRect ( &r_movie_filename );
	ForeColor ( whiteColor );
	MoveTo ( x_filename,y_filename );
	if (movieID) DrawString ( movie_name ); else DrawString ( localized ("\p(No movie loaded)","\p(Kein Video eingelegt)") );

	movie_draw_blockinfo();
}


// -----	draw background, anim_fram, file movie_name, block info, movie counter and slide bar indicator -------
//		-> background
//		-> slide bar
//		-> icon
// -----	Redraw window -------------------------------------------------------
RefreshMovieWindow ()
{	int i;

	if ( !movie_window ) return;
	SetPort(movie_window);
	
	EraseRect ( &movie_rect );
	PlotCIcon ( &r_camera, Get_CIcon(i_camera+anim_fram) );
	
	if ( !movie_slide_pic )	movie_slide_pic = GetPicture ( i_slide ); 
	if ( movie_slide_pic && *movie_slide_pic ) DrawPicture ( movie_slide_pic, &r_movie_slide_o );

	movie_draw_filename();
	movie_draw_frame();

	for (i=0;i<7;i++) Draw_Button ( movie_button+i, 0 );	// buttons
}


// -----	Eject selected movie ------------------------------------------------------------------
static eject_movie ( )
{	int i;

	if (h_buffers) DisposeHandle(h_buffers);
	StartStop(false); 
	movie_recording=false;
	movie_pos=movie_len=0;

	if (!movieID) return;		// no movie loaded
	FSClose(movieID);
	movieID = 0;

	if (!movie_window) return;
	movie_draw_filename();	
	for (i=0;i<7;i++) Set_Button(movie_button+i,0);
}


// -----	xor source buffer to target buffer
static BufferXBuffer ( void *source, void *target, long n )
{	union { long*lp; char*cp; void*vp; } q,z,e;

	q.vp = (long*)source;
	z.vp = (long*)target;
	e.cp = q.cp+n;

	while (n&3) { *z.cp++ ^= *q.cp++; n--; }
	while (q.lp<e.lp) *z.lp++ ^= *q.lp++;

}


/* -----	save current sequence -----------------------------
static save_sequence()
{	Char *p;

	if (mov_head.hardware<zx16||mov_head.hardware>zx128) Do_Abort ( "\pPanic error:","\pmovie header garbaged!",0 );

	xrom	= (mov_head.rom&rom_ext) && (movie_pos==sizeof(movhead));	// does core dump contain external rom image?
	ay		= mov_head.sound&ay_sound;								// ay register set included?
	
	p = compr;
	if (xrom) p = Shrink ( RdPtr(0), p, 0x4000 );


	if (seq_head.xcore)
	{
		switch (mov_head.hardware)
		{
		case zx16:	q=(long*)RdPtr(0x4000);
					z=core;
					e=q+0x1000;
					while (q<z) *z++ ^= *q++;
					p = Shrink ( core, p, 0x4000 );
		case zx48:
		}
	}
	else
	{
		switch (mov_head.hardware)
		{
		case zx16:	MemCopy ( RdPtr(0x4000), core, 0x4000 );
					p = Shrink ( core, p, 0x4000 );
					break;
		case zx48:	MemCopy ( RdPtr(0x4000), core+0x0000, 0x4000 );
					MemCopy ( RdPtr(0x8000), core+0x4000, 0x4000 );
					MemCopy ( RdPtr(0xc000), core+0x8000, 0x4000 );
					p = Shrink ( core, p, 0x0c000 );
					break;
		case zx128:	for (i=0;i<8;i++) MemCopy (ram[i],core+i*0x4000,0x4000);
					p = Shrink ( core, p, 0x20000 );
					break;
		}
	}
	
	
	
	

	p = Shrink ( inputs, compr, 12*frames );
	



	movie_err = SetFPos ( movieID, fsFromStart, movie_pos );				if (movie_err) goto xx;
	movie_err = Buffer2File ( seq_head, movieID, sizeof(seqhead) );			if (movie_err) goto xx;
	if ( ay )	movie_err = Buffer2File ( movieID, &ay_state, sizeof(aystate) );	if (movie_err) goto xx;	// ay registers
	




// -----	load current sequence
	movie_err = File2Buffer ( movieID, &seq_head, sizeof(seqhead) );			if (movie_err) goto xx;	// sequence header
	if ( ay )	movie_err = File2Buffer ( movieID, &ay_state, sizeof(aystate) );	if (movie_err) goto xx;	// ay registers
	m = seq_head.block_len-sizeof(seqhead)-(ay?sizeof(ay_state):0);	// size of compressed data
	movie_err = File2Buffer ( movieID, compr, m );						if (movie_err) goto xx;	// xrom[], inputs[] & core[]
	p = compr;

// -----	expand external rom image
	if ( xrom )	p = Expand ( RdPtr(0), compr, 0x4000 );					if (!p) goto xx;	

// -----	expand inputs
	p = Expand ( inputs, p, (8+4)*seq_head.frames );						if (!p) goto xx;

	
// ----- expand core
	switch (mov_head.hardware)
	{	
	case zx16:	n=1+xrom; break;
	case zx48:	n=3+xrom; break;
	case zx128:	n=8+xrom; break;
	}
		
	if (seq_head.xcore)	p = X_Expand ( p, core, n*0x4000 );
	else				p = Expand ( p, core, n*0x4000 );					if ( p != compr + m ) goto xx;

// copy core into specci ram
	p = core;
	switch (mov_head.hardware)
	{
	case zx16:	MemCopy ( p, WrPtr(0x4000), 0x4000 ); p+=0x4000; 
				break;
	case zx48:	MemCopy ( p, WrPtr(0x4000), 0x4000 ); p+=0x4000; 
				MemCopy ( p, WrPtr(0x8000), 0x4000 ); p+=0x4000; 
				MemCopy ( p, WrPtr(0xc000), 0x4000 ); p+=0x4000; 
				break;
	case zx128:	for (i=0;i<8;i++) { MemCopy ( p, ram[i], 0x4000 ); p+=0x4000; }
				break;
	}
	return;	

xx:	if (movie_err)	PrintNumberAlert(CStr(localized("Error while reading from movie:","Fehler beim Lesen von der Videokassette:")),movie_err);
	else			PrintAlert(CStr(localized("Sequence decompression error","Fehler beim Auspacken der Videosequenz"))); 
	eject_movie();
}
*/


/* -----	load current sequence ------------------------------
load_sequence();
load_sequence()
{	Char *p; int i; Boolean xrom, ay; long n,m;

	if (mov_head.hardware<zx16||mov_head.hardware>zx128) Do_Abort ( "\pPanic error:","\pmovie header garbaged!",0 );

	subframe	= 0;													// set actual subframe number
	xrom	= (mov_head.rom&rom_ext) && (movie_pos==sizeof(movhead));	// does core dump contain external rom image?
	ay		= mov_head.sound&ay_sound;								// ay register set included?

	if (movie_pos>=movie_len-4)		// end of file
	{
		movie_err = SetFPos ( movieID, fsFromStart, movie_pos );			if (movie_err) goto xx;
		movie_err = File2Buffer ( movieID, &seq_head.prev_block_len, 4 );		if (movie_err) goto xx;
		seq_head.block_len	= 4;
		seq_head.frame 	= mov_head.frames;
		seq_head.frames	= 0;
		
		seq_head.zreg		= zreg;
		
		seq_head.border	= bordercolor;
		seq_head.bank		= zx128pagectrl;
		seq_head.xcore		= (mov_head.frames/MAX_FRAMES)%MAIN_FRAMES ? xored_core : 0;
		
		if ( ay )
		{
			ay_state.regnr	= AYregnr;
			MemCopy ( AYreg, ay_state.reg, 16 );
		}

		if (xrom)						// we should read the external rom right now
		{							// and there is no core dump
			mov_head &= ~rom_ext;		// (( -> this must be an empty movie! ))
		}

		if (seq_head.xcore)
		{	long *q,*z,*p;
			p = (long*)core;
			switch (mov_head.hardware)
			{
			case zx16:	for ( z=p+0x1000,q=(long*)RdPtr(0x4000); p<z; ) *p++ ^= *q++;
						break;
			case zx48:	for ( z=p+0x1000,q=(long*)RdPtr(0x4000); p<z; ) *p++ ^= *q++;
						for ( z=p+0x1000,q=(long*)RdPtr(0x8000); p<z; ) *p++ ^= *q++;
						for ( z=p+0x1000,q=(long*)RdPtr(0xc000); p<z; ) *p++ ^= *q++;
						break;
			case zx128:	for (i=0;i<8;i++) { for ( z=p+0x1000,q=(long*)ram[i]; p<z; ) *p++ ^= *q++; }
						break;
			}
			// no need to check xrom here: xrom is only included in first core dump, which is never xored
			return;
		}

		p = core;
		switch (mov_head.hardware)
		{
		case zx16:	MemCopy ( RdPtr(0x4000), p, 0x4000 ); p+=0x4000; 
					break;
		case zx48:	MemCopy ( RdPtr(0x4000), p, 0x4000 ); p+=0x4000; 
					MemCopy ( RdPtr(0x8000), p, 0x4000 ); p+=0x4000; 
					MemCopy ( RdPtr(0xc000), p, 0x4000 ); p+=0x4000; 
					break;
		case zx128:	for (i=0;i<8;i++) { MemCopy ( ram[i], p, 0x4000 ); p+=0x4000; }
					break;
		}
		return;
	};


// -----	load current sequence
	movie_err = SetFPos ( movieID, fsFromStart, movie_pos );				if (movie_err) goto xx;
	movie_err = File2Buffer ( movieID, &seq_head, sizeof(seqhead) );			if (movie_err) goto xx;	// sequence header
	if ( ay )	movie_err = File2Buffer ( movieID, &ay_state, sizeof(aystate) );	if (movie_err) goto xx;	// ay registers
	m = seq_head.block_len-sizeof(seqhead)-(ay?sizeof(ay_state):0);	// size of compressed data
	movie_err = File2Buffer ( movieID, compr, m );						if (movie_err) goto xx;	// xrom[], inputs[] & core[]
	p = compr;

// -----	expand external rom image
	if ( xrom )	p = Expand ( RdPtr(0), compr, 0x4000 );					if (!p) goto xx;	

// -----	expand inputs
	p = Expand ( inputs, p, (8+4)*seq_head.frames );						if (!p) goto xx;

	
// ----- expand core
	switch (mov_head.hardware)
	{	
	case zx16:	n=1+xrom; break;
	case zx48:	n=3+xrom; break;
	case zx128:	n=8+xrom; break;
	}
		
	if (seq_head.xcore)	p = X_Expand ( p, core, n*0x4000 );
	else				p = Expand ( p, core, n*0x4000 );					if ( p != compr + m ) goto xx;

// copy core into specci ram
	p = core;
	switch (mov_head.hardware)
	{
	case zx16:	MemCopy ( p, WrPtr(0x4000), 0x4000 ); p+=0x4000; 
				break;
	case zx48:	MemCopy ( p, WrPtr(0x4000), 0x4000 ); p+=0x4000; 
				MemCopy ( p, WrPtr(0x8000), 0x4000 ); p+=0x4000; 
				MemCopy ( p, WrPtr(0xc000), 0x4000 ); p+=0x4000; 
				break;
	case zx128:	for (i=0;i<8;i++) { MemCopy ( p, ram[i], 0x4000 ); p+=0x4000; }
				break;
	}
	return;	

xx:	if (movie_err)	PrintNumberAlert(CStr(localized("Error while reading from movie:","Fehler beim Lesen von der Videokassette:")),movie_err);
	else			PrintAlert(CStr(localized("Sequence decompression error","Fehler beim Auspacken der Videosequenz"))); 
	eject_movie();
}
*/


static store_frame ()
{
	inputs[subframe++] = keyboard;
	mov_head.frames++;
	
	if (subframe>=MAX_FRAMES)	subframe=0; // save_sequence();
}



static get_frame ()
{
	if (subframe>=seq_head.frames)	subframe=0; // load_sequence();
	
	keyboard = inputs[subframe++];
}


StartStopMovie ( Boolean run )
{	
	if (!movieID) return;
	if (movie_window)
	{	SetPort(movie_window);
		Set_Button(movie_button+START, 2-run);
	}
	else
	{
		movie_button[START].state=2-run;
	}
}


static SetRecording(Boolean f)
{
	if (!movieID) return;
	movie_recording = f;
	
	if (movie_window)
	{
		SetPort(movie_window);
		Set_Button ( movie_button+RECORD, 1+f );
	}
}


// -----	truncate movie at current position ------------------
static ClipMovie()
{
}


// -----	Load or save frame informations depending on flag movie_recording
//		automatically clip movie and switch to recording if keyboard[] indicates pressed keys
DoMovieFrame()
{
	if (!movieID) return;

	if (!(subframe%25)) movie_draw_counter();

	if (movie_recording)
	{
		store_frame();
	}
	else
	{
		if ( (keyboard.quad.a&keyboard.quad.b&keyboard.quad.c) != -1 )		// key pressed -> switch to record immediately
		{
			ClipMovie();
			SetRecording(true); 
			store_frame();
		}	
		else
		{	
			get_frame();
		}
	}
}



/*****************************************************************/
/*****************************************************************/
/*****************************************************************/



// -----	Open new movie file -----------------------------------------------------
OpenMovie ( FSSpec file )
{	int i;

	OpenMovieWindow();
	eject_movie();

	movie_err = FSpOpenDF ( &file, fsCurPerm, &movieID );			if (movie_err) { movieID=0; goto xx; }

	movie_err = GetEOF (movieID, &movie_len);					if (movie_err) goto xx;
	movie_err = SetFPos(movieID, fsFromStart,0);				if (movie_err) goto xx;
	movie_err = File2Buffer(movieID,&mov_head,sizeof(movhead));	if (movie_err) goto xx;
	
	memcpy ( movie_name, file.name, 64 );
	StartStop(true);
	movie_recording = false;
	movie_pos = sizeof(movhead);
	subframe = 0;

	if (mov_head.file_id != file_idf) 	{ PrintAlert(CStr(localized("This is not a Specci movie","Das ist kein ZX Spectrum-Video"))); goto xx; }
	if (mov_head.appl_id != appl_idf) 	{ PrintAlert(CStr(localized("Caution: Movie of unknown source!","Vorsicht: Video unbekannter Herkunft"))); }
	else
	if (mov_head.appl_ver < appl_vers)	{ PrintAlert(CStr(localized("Caution: Old version!","Vorsicht: Alte Version"))); }
	else
	if (mov_head.appl_ver > appl_vers)	{ PrintAlert(CStr(localized("Caution: Newer version!","Vorsicht: Neuere Version"))); }
#if !GENERATINGPOWERPC
	if (mov_head.z80_ver == c_vers)	HandleSpeedMenu(iBest); else
	if (mov_head.z80_ver == a_vers)	HandleSpeedMenu(iBetter); else
#else
	if (mov_head.z80_ver != c_vers)	
#endif
	{	PrintAlert(CStr(localized("Wrong version of z80 engine","Falsche Version der Z80-Implementierung")));	}
#if EXACT_R
	if (!(mov_head.options&exact_r)) 
	{	PrintAlert(CStr(localized("Caution: NO exact r register emulation required!","Vorsicht: KEINE exakte R-Register-Emulation benötigt!")));	}
#else
	if (mov_head.options&exact_r)
	{	PrintAlert(CStr(localized("Caution: Exact r register emulation required!","Vorsicht: Exakte R-Register-Emulation wird benötigt!")));	}
#endif
	
	if (z80_hardware!=mov_head.hardware)
		switch (mov_head.hardware)
		{
		case	zx16:	
		case zx48:	HandleOptionsMenu(i48k); break;
		case zx128:	HandleOptionsMenu(i128); break;
		default:		PrintAlert(CStr(localized("Unsupported ZX Spectrum hardware","Unbekannte ZX Spectrum Version"))); 
					goto xx;
		}
	
	if (mov_head.rom & rom_48) {}
	if (mov_head.rom & rom_1280) {}
	if (mov_head.rom & rom_1281) {}
	if (mov_head.rom & rom_if1) {}
	if (mov_head.rom & rom_ext) {}		// •••• !!missing!! ••••
	
	if (mov_head.patches != load_patch+save_patch+beep_patch) {}
	if (mov_head.joystick != KEMPSTON) {}
	
	SetSoundChannels ( mov_head.sound&ay_sound ? on : off );
	
	if (mov_head.options&highres) HandleSpeedMenu (iHighresMode); else HandleSpeedMenu(iExactSpeed);

//	HandleOptionsMenu ( iNormalSize + mov_head.scr_size-1 );		// only informational
//	HandleSpeedMenu ( iRefreshRate + mov_head.scr_rfsh-1 );		// only informational
	
	
	if (!movie_window) return;
	movie_draw_filename();
	for (i=0;i<7;i++) Set_Button(movie_button+i,1);
	Set_Button(movie_button+START,2-z80_running);
	return;

xx:	if (movie_err) PrintNumberAlert(CStr(localized("Error while opening movie:","Fehler beim Einlegen der Videokassette:")),movie_err);
	eject_movie();
}


movie_set_btn ( short idx, short o, short l, short i0, short i1, short i2, short state);
movie_set_btn ( short idx, short o, short l, short i0, short i1, short i2, short state)
{	btn *b; Rect r;

	b=movie_button+idx;
	r=(**Get_CIcon(i_button)).iconPMap.bounds;
	
	Link ( movie_button, b );

	SetRect ( &b->bbox, l,o,l+r.right-r.left,o+r.bottom-r.top );
	b->btn[0] = b->btn[1] = i_button;
	b->pbtn[0] = b->pbtn[1] = i_pbutton;
	b->icn[0] = i0;
	b->icn[1] = i1;
	b->icn[2] = i2;
	b->icn[3] = nil;
	b->nxt_sta[0] = 0;
	b->nxt_sta[1] = 2;
	b->nxt_sta[2] = 1;
	b->nxt_sta[3] = 0;
	b->state = state;
	b->id = idx;
}
	

// -----	Close the movie recorder window -----------------------------------------------------
//		updates movieRect for preferences
CloseMovieWindow ()
{	
	CheckItem ( windowsmenu, iMovie, false );
	if (!movie_window) return;
	SetPort(movie_window);
	movieRect = movie_window->portRect;
	LocalToGlobal ( (Point*)&movieRect ); 
	LocalToGlobal (((Point*)&movieRect)+1 ); 
	DisposeWindow ( movie_window ); 
	movie_window=nil;
}


// -----	Open the movie recorder window ------------------------------------------------------
OpenMovieWindow ()
{	short i; 
	
	if (movie_window) goto xx;							// already open -> only SetPort() etc.

	anim_fram = 0;

	if ( movie_button->next==movie_button+1 ) goto yy;		// re-open: movie[] and movie_button[] already initialized


// init linked list of buttons
	movie_button->next = movie_button->prev = movie_button;
// 7 buttons for movie recorder 1
	i = movieID ? 1 : 0;
	movie_set_btn ( 0,	y_btn,x_x,		i_new_tape, i_erase,i_erase+red,		i); movie_button[0].nxt_sta[0]=3; movie_button[0].nxt_sta[3]=0; movie_button[0].icn[3]=i_new_tape+green;
	movie_set_btn ( 1,	y_btn,x_o,		i_record, i_record, i_record+red,		i);
	movie_set_btn ( 2,	y_btn,x_rewind,	i_rewind, i_rewind, i_rewind+green, 	i);
	movie_set_btn ( 3,	y_btn,x_prev,		i_prev, i_prev, i_prev+green,			i);
	movie_set_btn ( 4,	y_btn,x_start,		i_stop_go, i_start+green, i_pause+red,	i);
	movie_set_btn ( 5,	y_btn,x_next,		i_next, i_next, i_next+green,			i);
	movie_set_btn ( 6,	y_btn,x_end,		i_end, i_end, i_end+green,			i);


// create and open window
yy:	movieRect.right = movieRect.left + (movie_rect.right-movie_rect.left);
	movieRect.bottom = movieRect.top + (movie_rect.bottom-movie_rect.top);
	
	movie_window = NewCWindow (nil, &movieRect, localized("\pMovie recorder","\pVideorekorder"), true, movie_wdef, (WindowPtr)-1, true, 0);
	if (!movie_window)  abort_oomem("\pOpenMovieWindow()"); 

xx:	SetWRefCon(movie_window,iMovie);
	SetPort ( movie_window );
	GetFNum ( "\pGeneva", &i );
	TextFont ( i );
	TextSize ( textsize );
	TextMode ( srcOr );
	RGBBackColor (&movie_bkgr_color );
	CheckItem ( windowsmenu, iMovie, true );
}


// -----	copy screen contents -> highres screen buffer ----------------------------------------
movie_update_highres ( );
movie_update_highres ( )
{	
	while (!EXIT) Do_Cycles();
	EXIT=false;
	while (!EXIT) Do_Cycles();
}


// -----	wait movie_delay -------------------------------------------------------------------------
//		animation of cassette icon
Boolean movie_delay ( long movie_delay );
Boolean movie_delay ( long movie_delay )
{
	movie_delay += TickCount();			// final tick number
	movie_update_highres();
	
	while ( (long)TickCount()-movie_delay<0 )
	{	animate_camera();
		UpdateKeyboard();
		if ( ~keyboard.byte[7] & 1 ) return error;
	}
	return ok;
}


// -----	Erase (delete) rest of movie ----------------------------------------
//		
erase_movieblock ( );
erase_movieblock ( )
{	static long len = 0;

	if ( !movieID || !z80_running ) return;

	subframe=0;		// •••provisional•••

	if (subframe==0)
	{	movie_len = movie_pos+4;
		movie_err = SetEOF(movieID,movie_len);
		movie_err = SetFPos(movieID,fsFromStart,movie_pos);
		movie_err = Buffer2File(&len,movieID,4);
	}
	else 
	{	// ••• missing •••
	}

	subframe = 0;	
	movie_draw_blockinfo();
}


// ----- Step to next data block on movie --------------------------------------
static next_movieblock ( )
{	
	if ( !movieID ) return;							// no movie loaded

	WaitCursor();
	movie_delay(SPOOL_DELAY);
	ArrowCursor();
	
	if (movie_pos>=movie_len) return;					// already at end
	movie_pos += seq_head.block_len;
	subframe = 0;	
	movie_draw_blockinfo();
}


// ----- Step to previous data block on movie ------------------------------------
static prev_movieblock ( )
{
	if (!movieID ) return;

	WaitCursor();
	movie_delay(SPOOL_DELAY);
	ArrowCursor();

	if (movie_pos<=sizeof(movhead)) return;
	movie_pos -= seq_head.prev_block_len;
	subframe		= 0;	
	movie_draw_blockinfo();
}


// -----	Rewind movie --------------------------------------------------------------
static first_movieblock ( )
{
	if (!movieID ) return;

	WaitCursor();
	movie_delay ( movie_pos>0 ? SPOOL_DELAY*2 : SPOOL_DELAY );
	ArrowCursor();
	
	movie_pos	= sizeof(movhead);
	subframe		= 0;	
	movie_draw_blockinfo();
}


// -----	Wind movie to end ------------------------------------------------------
static last_movieblock ( )
{
	if (!movieID ) return;

	WaitCursor();
	movie_delay ( movie_pos<movie_len ? SPOOL_DELAY*2 : SPOOL_DELAY );
	ArrowCursor();
	
	movie_pos 	= movie_len-4;
	subframe		= 0;	
	movie_draw_blockinfo();
}


// -----	create and insert an empty movie ----------------------------------------------
static empty_movie ( )
{	StandardFileReply reply;
	FInfo fndrInfo;
	Boolean fndrInfoValid = false;
	int i;
	static long len = 0;
		
	OpenMovieWindow();
	if (movieID) eject_movie();

// querry file name and create empty file
	StandardPutFile  (  localized ( "\pName the empty movie:","\pName für die leere Kassette:" ), "\pEmpty.movie", &reply );
	if (!reply.sfGood) return;
	
	if (reply.sfReplacing)
	{
		movie_err = FSpGetFInfo (&reply.sfFile, &fndrInfo); 		if (movie_err) goto xx;
		fndrInfo.fdType='ZMov';
		fndrInfo.fdCreator='ZXSP';
		fndrInfoValid = true;
		movie_err = FSpDelete (&reply.sfFile);					if (movie_err) goto xx;
	}

	movie_err = FSpCreate ( &reply.sfFile, 'ZXSP', 'ZMov', 0 );		if (movie_err) goto xx;

	if (fndrInfoValid) FSpSetFInfo (&reply.sfFile, &fndrInfo);

// build movie header from current settings
	memset ( &mov_head, 0, sizeof(movhead) );
	mov_head.file_id	= file_idf;
	mov_head.appl_id	= appl_idf;
	mov_head.appl_ver	= appl_vers;
	mov_head.z80_ver	= c_vers;
#if !GENERATINGPOWERPC
	if (use_better_z80) mov_head.z80_ver = a_vers;
#endif
	mov_head.frames	= 0;
	mov_head.hardware	= z80_hardware;
	mov_head.rom		= ( z80_hardware==zx128 ? rom_1280+rom_1281 : rom_48 ) + ( romID ? rom_ext : 0 );
	mov_head.patches	= load_patch + save_patch + beep_patch;
	mov_head.joystick	= KEMPSTON;
	mov_head.sound	= GetSoundChannels() ? ay_sound : off;
	mov_head.scr_size	= scrnsize;
	mov_head.options	= EXACT_R*exact_r + run_highres*highres;

// write movie header and 'final prev-block-len' to file
	movie_err = FSpOpenDF (&reply.sfFile, fsRdWrPerm, &movieID);	if (movie_err) { movieID=0; goto xx; }
	movie_err = Buffer2File (&mov_head,movieID,sizeof(movhead));	if (movie_err) goto xx;
	movie_err = Buffer2File (&len,movieID,4);					if (movie_err) goto xx;

// allocate last-state and compression buffer
	h_buffers	= TempNewHandle ( pagesize*((z80_hardware==zx128?8:3)+(romID!=0))*22/10 + MAX_FRAMES*(8+1), &movie_err );
	if (movie_err) goto xx;
	HLock(h_buffers);

	inputs	= (kbd*)*h_buffers;										// inputs recording buffer
	core		= (Char*)(inputs+MAX_FRAMES);							// old core buffer
	compr	= core + pagesize*((z80_hardware==zx128?8:3)+(romID!=0));	// compression buffer
	
	movie_len 		= sizeof(movhead)+4;
	movie_pos		= sizeof(movhead);
	memcpy ( movie_name, reply.sfFile.name, 64 );
	StartStop(false);
	movie_recording	= false;
	subframe			= 0;

	movie_draw_filename();		// validates block info too
	for (i=0;i<7;i++) Set_Button(movie_button+i,1);
	return;

xx:	if (movie_err) PrintNumberAlert(CStr(localized("Error while creating movie:","Fehler beim Anlegen der Videokassette:")),movie_err);
	eject_movie();
}


// -----	Handle "Insert Movie" --------------------------------------------------------------
//		Ask for a file and open it
insert_movie ( );
insert_movie ( )
{	StandardFileReply reply;	
	SFTypeList	typeList 	= {'ZMov',0,0,0};	// list of file types 

	StandardGetFile ( nil, 1, typeList, &reply ); 
	if (reply.sfGood) OpenMovie(reply.sfFile);
}


// -----	Mouse click in movie recorder window -------------------------------------------
ClickInMovieWindow ()
{	Point p; btn *b;

	if (!movie_window) return;
	SetPort ( movie_window );

	p = event.where;

	if ( b = Find_Button(movie_button,p) )
	{	if ( Handle_Button(b) )
		{
			switch (b->id)
			{
			case 0:	if (b->state==2)	erase_movieblock();			// erase block
					else				empty_movie();	break;	// insert empty movie
			case 1:	SetRecording(b->state==2);			return;	// record on/off
			case 2:	SetRecording(off);StartStop(false);first_movieblock();					break;	// rewind	
			case 3:	SetRecording(off);StartStop(false);prev_movieblock();					break;	// prev block
			case 4:	HandleMENU (mFile,b->state==1?iResume:iSuspend);		return;	// start/stop
			case 5:	SetRecording(off);StartStop(false);next_movieblock();					break;	// next block
			case 6:	SetRecording(off);StartStop(false);last_movieblock();					break;	// end of movie
			}
		
		// reset movie_button icon for those, who breaked (and not returned)
			if (b->state>=2)
			{	SetPort(movie_window);
				Set_Button ( b, 3-b->state );
			}
		}
		return;
	}
	
	if ( PtInRect ( p, &r_camera ) )	// insert/eject movie
	{	if (movieID) eject_movie(); else insert_movie();
		return;
	}

	if ( PtInRect ( p, &r_movie_blockinfo ) )	// popup menu of blocks
	{	return;
	}
}


// -----	Save snapshot as movie -------------------------------------------------------
//		••• not yet implemented •••
OSErr SaveMOVIE ( short fileID )	
{
	fileID;
}


#endif	 // include_movie



