

/*	Button Handler

***	Copyright:	1994-1999 Günter Woigk
  				mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

    20.Oct.01   revised for CW7Pro and ANSI C compliance								hdusel@bnv-gz.de
*/

#include	"kio.h"
#include	"options.h"
#include	"buttons.h"
#include	"messages.h"
#include	"strings.h"


void Link (btn* a, btn* b)	// 27.Dec.1999 kio
{
	b->next = a->next;
	b->prev = a;

	a->next->prev = b;
	a->next = b;

}


// ----	redraw button
//		-> draws the normal or pressed version depending on flag
//		-> if required button handle is nil, redraw_button() tries to guess a button
void Draw_Button ( btn *btn, Boolean pressed )
{	int i,j;
	short n; 
	CIconHandle h;
	Rect r,r2;

	i = btn->state &= MAX_BTN_STATES-1;

	n = pressed ? btn->pbtn[i] : btn->btn[i];

	if ( !n )							// no button defined for this state -> guess
	{	
		if (!btn->btn[0])  btn->btn[0]  = btn->pbtn[0];
		if (!btn->pbtn[0]) btn->pbtn[0] = btn->btn[0];
		for (j=1;j<MAX_BTN_STATES;j++)
		{	if (!btn->btn[j])  btn->btn[j]  = btn->btn[j-1];
			if (!btn->pbtn[j]) btn->pbtn[j] = btn->pbtn[j-1];
		}
		n = pressed ? btn->pbtn[i] : btn->btn[i];
		if (!n) return;
	}

	h = Get_CIcon(n);
	if (!h||!*h) return;		
	r = btn->bbox;
	PlotCIcon (&r, h);			// draw button
	
	n = btn->icn[i];
	if (!n) return;
	h = Get_CIcon(n);
	if (!h||!*h) return;		
	
	r2=(**h).iconPMap.bounds;
	r.left  += ( (r.right-r.left)-(r2.right-r2.left) ) / 2 + pressed;
	r.top   += ( (r.bottom-r.top)-(r2.bottom-r2.top) ) / 2 + pressed;
	r.right  = r.left + (r2.right-r2.left);
	r.bottom = r.top  + (r2.bottom-r2.top);
	PlotCIcon (&r, h);		// draw symbol on button
}



// ----	set button to specific state
//		-> similar to setting new state manually and calling Draw_Button()
//		-> sound effect / animation may be added later
//		-> does not redraw button, if state does not change
void Set_Button (btn *btn, short new_state)
{
	new_state  &= MAX_BTN_STATES-1;			// force legal
	btn->state &= MAX_BTN_STATES-1;
	
	if (btn->state == new_state) return;	// nothing changed

	btn->state = new_state;					// normal state change
	Draw_Button(btn,0);
}



// ----	handle mouse click on button
//		-> pressed-icon, mouse tracking, state switching
//		-> returns true if button was actually switched, else false
Boolean Handle_Button (btn *btn)
{	Point p;
	char old_state;
	char new_state;
	
	old_state = btn->state & (MAX_BTN_STATES-1);
	new_state = btn->nxt_sta[old_state] & (MAX_BTN_STATES-1);

	while(true)
	{
		btn->state = new_state;
		Draw_Button(btn,1);								// pressed version
		do
		{	PollEvents (updateMask|mUpMask, 1, nil);
			if (!StillDown())							// released inside button => pressed!
			{	Draw_Button(btn,0);
				return true;
			}
			GetMouse(&p); 
		} while (PtInRect(p,&btn->bbox));
	
		btn->state = old_state;
		Draw_Button(btn,0);								// released version
		do
		{	PollEvents (updateMask|mUpMask, 1, nil);
			if (!StillDown()) return false;				// released outside button => canceled!
			GetMouse(&p);
		} while (!PtInRect(p,&btn->bbox));
	}
}



// ----	Find button in which user clicked
//		-> mouse point in local coordinates 
//		-> linked list of buttons 
//		-> returns nil if not inside any button
//		-> list may be cyclic or nil-terminated
btn* Find_Button ( btn *btn, Point p )
{	struct btn *b;
		
	b=btn; 
	
	while (!PtInRect(p,&b->bbox))
	{	b=b->next;
		if (!b || b==btn) return nil;
	}

	return b;
}



// ======================================================================
// =================== Color Icon Handling ==============================
// ======================================================================


/* these constants must be defined in options.h:
#define		MAX_CIC		100		// number of cicns
#define		MIN_CIC_ID	128		// minimum cicn number to look for
#define		MAX_CIC_ID	512		// maximum cicn number to look for (excl.)
*/

CIconHandle	cic[MAX_CIC];							// CIcon handles

#if MAX_CIC<=256
Char		cic_id_array[MAX_CIC_ID-MIN_CIC_ID];	// cicn number -> cic[] index
Char*		cic_id		= 0;						// actually used pointer (normalized by MIN_CIC_ID)
#else
short		cic_id_array[MAX_CIC_ID-MIN_CIC_ID];
short*		cic_id		= 0;
#endif



// ----	Get Color Icon ------------------------------------------------
//		returns handle of CIcon
//		returns nil if n oorange or CIcon not ex.
CIconHandle Get_CIcon ( short n )
{
	if (n>=MIN_CIC_ID && n<MAX_CIC_ID)	return cic[cic_id[n]];
	else								return nil;
}	



// ----	Read CIcons into memory ---------------------------------------
//		reads all color icons into memory
//		aborts if cic[] overflows (can only happen while working on a program)
void Read_CIcons (void)
{	short i,n; CIconHandle h;

	if (cic_id) return;								// multiple call

	cic_id = cic_id_array-MIN_CIC_ID;				// init pointer -> cic_id[0]
	cic[0] = nil;									// first entry := dummy for n.ex. cicn id's
	n = 1;											// number of used cic[] slots

	for (i=MIN_CIC_ID;i<MAX_CIC_ID;i++)				// read color icons
	{
		if ( (h = GetCIcon(i)) != 0 )
		{
			cic_id[i] = n;
			cic[n++] = h;
			if (n >= MAX_CIC) do_abort("\pRead_CIcon():","\parray exceeded",0);
		}
		else
		{	cic_id[i] = 0;							// cic[0] = nil
		}
	}
}	






























