

/*	FREEZE / MELT - file compressor / decompressor

***	Copyright:	1991-1999 Günter Woigk
  				mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#include	"kio.h"
#include	"options.h"
#include	"freeze.h"



// -----	Settings for Freeze() ---------------------------------------------------------
#define		SHIFT		7				// 7=fastest, for high redundance … 3=slowest, for low redundance
#define		WIN_SIZE		(1<<(15-SHIFT))	// 7->256  6->512  5->1024  4->2048  3->4096  
#define		MAX_LEN		((1<<SHIFT)+1)		// 7->129  6->65   5->33    4->17    3->9
#define		WIN_MASK	(WIN_SIZE-1)


/* -----	compress source buffer -------------------------------------
		returns pointer to compressed data end (excl.)
		••• destination buffer must be at least src_len+(src_len/MAX_LEN)+2 bytes in size! •••
		••• destination buffer overflow is not checked! •••
*/
Char* Freeze ( Char* src, Char* dest, long n )
{	register Char *q, *z, *p;
	Char *src_anf, *src_end, *cnt;
	Char *a, c;

	Char *prev[WIN_SIZE];
	Char *last[256];

// init
	MemClear ( last,256*sizeof(*last));		// clear last_occurance_of_byte array
	src_anf = src;							// source start
	src_end = src+n;						// source end
	cnt = nil;								// pointer to uncompressed bytes counter

	while ( src<src_end )
	{

// search 'template' for byte stream at current position
		c = *src;							// first byte of next byte stream
		n = 0;							// best match stream length
		
		for ( p=last[c]; p>=src-WIN_SIZE && *p==c; p = prev[(long)p&WIN_MASK] )
		{
			q=p;z=src;
			while (*q==*z) { q++; z++; }
			if (q-p<=n) continue;
			a=p; n=q-p;
			if (n<MAX_LEN) continue;
			n=MAX_LEN; break;
		}	

		if (src+n>src_end) n=src_end-src;
		//	=>	a->pos; n=len

// compressible sequence && sequence long enough to save space?
		if ( n>=2+(cnt!=nil) )		
		{
			*(short*)dest = a-src + n-2;		// store compression information
			dest += 2;						// increment destination pointer
			while (n)
			{	c=*src;
				prev[(long)src&WIN_MASK] = last[c];	// set prev[addr]
				last[c] = src++;					// set last[byte]
				n--;
			}
			cnt = nil;						// uncompressed byte counter no longer valid
			continue;
		}
		
// uncompressible byte
		*dest++ = c;						// store byte
		prev[(long)src&WIN_MASK] = last[c];	// set prev[addr]
		last[c] = src++;					// set last[byte]
		
		if (cnt)							// are we within an uncompressible sequence?
		{	*cnt++;						// increment uncompressed byte counter
			if (*cnt<=127) continue;			// loop
			cnt = nil; continue;				// counter reached maximum -> start a new sequence
		}

		cnt = dest++;						// new uncompressed bytes counter
		*cnt = 1;							// uncompressed bytes := 1
	}

	return dest;
}



/* -----	Fast compression of simple byte repetitions -------------------------------------
		returns pointer to compressed data end (excl.)
		••• destination buffer should be src_len+20% bytes in size! •••
		••• destination buffer overflow is not checked! •••

sheme:

1.		dc.b		$FF			; repetition of $00
		dc.b		cnt			; number of bytes = cnt
	   ((	dc.w		cnt2	 ))		; if cnt==0 then cnt := cnt2+1

2.		dc.b		$FE			; byte repetition
		dc.b		cnt			; number of bytes
	   ((	dc.w		cnt2	 ))		; if cnt==0 then cnt := cnt2+1
		dc.b		byte			; repeated byte

3.		dc.b		cnt			; uncompressed bytes; cnt = 0 ... $FD
	   ((	dc.w		cnt2	 ))		; if cnt==0 then cnt := cnt2+1
		dc.b		data[]		; uncompressed bytes
*/
Char* Shrink ( Char* src, Char* dest, long n )
{	Char *src_end, *p, c;

	while (n>0x10000) 		// eliminate need to check for count overflow
	{	dest=Shrink(dest,src,0x10000); src+=0x10000; n-=0x10000; 
	};

	p = src;
	src_end = src+n;

	while ( p<src_end )
	{
		src=p;

		c=*p++;
		while (*p==c&&p<src_end) p++;

		n=p-src;
		
		if ( n>=3 )
		{
			if (c==0)
			{	if (n<0x00fe) { *dest++ = 0xFF; *dest++ = n; continue; }
				*(long*)dest = 0xFF000000 + n-1; dest+=4; continue;			// little endian!!
			}
			else 
			{	if (n<0x00fe) { *dest++ = 0xFE; *dest++ = n; *dest++=c; continue; }
				*(long*)dest = 0xFE000000 + n-1; dest+=4; *dest++=c; continue;	// little endian!!
			}
		}

a:		while (*p!=c && p<src_end ) c=*p++;
		if ( *(p+1)!=c && p<src_end ) { p++; goto a; }
		p--; if ( p >= src_end-3 ) p = src_end;
		
		n = p-src;

		if (n<0x00fe) { *dest++ = n; }
		else { *dest++ = 0; *(short*)dest = n-1; dest+=2; }						// little endian!!
		
		MemCopy ( src, dest, n );
		dest+=n;
	}

	return dest;
}	


/* -----	expand shrinked buffer --------------------------------------------
		returns pointer to compressed source data end
		returns nil if source ended not at sequence boundary
*/
Char* Expand ( Char* src, Char* dest, long n )
{	Char *dest_end;

	dest_end = dest+n;
	
	while ( dest<dest_end )
	{
		switch ( *src++ )
		{
		case 0x0ff:	// repetition of $00
			n = *src++;
			if (!n) { n = *(Short*)src+1; src+=2; }
			if (dest+n>dest_end) { n=dest_end-dest; src=nil; }
			MemClear ( dest, n );
			dest+=n; 
			continue;
		case 0x0fe:		// repetition of byte
			n = *src++;
			if (!n) { n = *(Short*)src+1; src+=2; }
			if (dest+n>dest_end) { n=dest_end-dest; src=nil; }
			MemSet ( *src++, dest, n );
			dest+=n; 
			continue;
		default:		// uncompressed data (flag byte omitted)
			n = src[-1];
			if (!n) { n = *(Short*)src+1; src+=2; }
			if (dest+n>dest_end) { n=dest_end-dest; src=(Char*)(-n); }
			MemCopy ( src, dest, n );
			dest+=n; src+=n; 
			continue;
		}
		
	}
	
	return src;
}


/* -----	expand shrinked buffer --------------------------------------------
		XOR target with source - intended for repeated compressions of slightly modified buffers
		returns pointer to compressed source data end
		returns nil if source ended not at sequence boundary
*/
Char* X_Expand ( Char* src, Char* dest, long n )
{	Char *dest_end;
	Char c;

	dest_end = dest+n;
	
	while ( dest<dest_end )
	{
		switch ( *src++ )
		{
		case 0x0ff:		// repetition of $00 -> no change of source
			n = *src++;
			if (!n) { n = *(Short*)src+1; src+=2; }
			dest+=n; 
			if (dest>dest_end) { dest=dest_end; src=nil; }
			continue;
		case 0x0fe:		// repetition of byte
			n = *src++;
			if (!n) { n = *(Short*)src+1; src+=2; }
			if (dest+n>dest_end) { n=dest_end-dest; src=nil; }
			c=*src++; while (n--) *dest++ ^= *src;
			continue;
		default:			// uncompressed data (flag byte omitted)
			n = src[-1];
			if (!n) { n = *(Short*)src+1; src+=2; }
			if (dest+n>dest_end) { n=dest_end-dest; src=(Char*)(-n); }
			while (n--) *dest++ ^= *src++;
			continue;
		}
		
	}
	
	return src;
}























