
/*	application dependent z80 I/O instruction handlers

***	Copyright:	1994-1996 Günter Woigk
  				mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	23.Jan.95:	Created this file collecting some stuff from other files		KIO !
	24.Mar.95:	Log sound output statement with time stamp					KIO !
	04.Jun.95:	Revised sound output logging to combine ear and mic output	KIO !
	05.Jun.95:	Border color logging										KIO !
	11.Jun.95:	Kempston interface											KIO !
    20.Oct.01   revised for CW7Pro and ANSI C compliance					hdusel@bnv-gz.de
*/

#include	"kio.h"
#include	"strings.h"
#include	"z80.h"
#include	"keyboard.h"
#include	"application.h"
#include	"messages.h"
#include	"display.h"
#include	"sound.h"
#include	"z80 main.h"
#include	"help+about.h"


#define	now	(TOTAL-CYCLES)


/* -----	Sound output to Mic and Tape -------------------------------------------------
*/
#define	m_ear		0x10
#define	m_mic		0x08
#define	snd_mask		(m_ear+m_mic)	// both outputs mixed to Mac speaker

#define	in_fe			0xff				// value read from port 0xfe if line silent: 0xff (early specci rev. 2)  or  0xdf (later models)


// ======================================================================


// -----	Handler for Input instructions --------------------------------------------
Char Do_Input ( Short addr )
{	Char c;

	switch ( (Char)addr )
	{	
	case 0x00fe:		// Keyboard Matrix lesen
		c = in_fe;  	// bit 5 is line input
		addr = ~addr;
		if ( addr&0x0100 ) c &= keyboard.byte[0] & kbdpict[0];
		if ( addr&0x0200 ) c &= keyboard.byte[1] & kbdpict[1];
		if ( addr&0x0400 ) c &= keyboard.byte[2] & kbdpict[2];
		if ( addr&0x0800 ) c &= keyboard.byte[3] & kbdpict[3];
		if ( addr&0x1000 ) c &= keyboard.byte[4] & kbdpict[4];
		if ( addr&0x2000 ) c &= keyboard.byte[5] & kbdpict[5];
		if ( addr&0x4000 ) c &= keyboard.byte[6] & kbdpict[6];
		if ( addr&0x8000 ) c &= keyboard.byte[7] & kbdpict[7];
		return c;

	case 0x001f:	return ( ~keyboard.joy.kempston );				// Kempston interface

	case 0x00ff:		// ULA won't stop -> read mixture of 0xff (idle bus) and screen byte / attr byte
		if (CYCLES<=224*64) return 0xff;						// above screen
		if (CYCLES>=224*(64+192)) return 0xff;					// below screen
		return video_page[(CYCLES-224*64)/(224/32)];			// return screen byte
		
	case 0x00fd:		// ---- zx128/+2: paging, sound chip
		switch (addr)
		{
		case 0xfffd:	// ---- zx128/+2: sound chip: read data register
			return Read_AY_Register();
		};
	};

	if ( msg>=log_level )
	{	PrintStr	(PStr("\pIN ("));
		PrintHex4	(addr);
		PrintChr	(')');
		PrintNL	();
	};
	
	return 0xFF;
}


// -----	Handler for output instructions -------------------------------------------
void Do_Output ( Short addr, Char n )
{	static short oldborder = 99;
	static Char oldsound = 0x00;	
	
	switch ( (Char)addr )
	{
	case 0x00fe:	// ---- border, mic & ear sound output
// border change
		bordercolor = n&0x07;
		if (bordercolor!=oldborder)
		{	oldborder = bordercolor;
			if (brdrlogptr<brdrlogend)	*brdrlogptr++ = (now&0xfffffff8)+oldborder;
		};
// sound output
		if ((n&snd_mask)!=oldsound)
		{	oldsound = n&snd_mask;
			if (sndlogptr<sndlogend)	{ *sndlogptr++ = now; *sndlogptr++ = 0x1000 + oldsound; }
		};
		return;

	case 0x00fd:		// ---- zx128/+2: paging, sound chip
		switch (addr)
		{
		case 0x7ffd:	ZX128PageCtrl(n);		return;
		case 0xfffd:	Select_AY_Register(n);	return;
		case 0xbffd:	//if ( msg>=log_level ) { PrintHex8(now); PrintStr("\pAY:",0); PrintHex2(AYregnr); PrintChar('='); PrintHex2(n); PrintNL(); };	
					Write_AY_Register(n);	return;
		};
	};

	if ( msg>=log_level )
	{	PrintStr	(PStr("\pOUT ("));
		PrintHex4	(addr);
		PrintChr	(')');
		PrintHex2	(n);
		PrintNL	();
	};
}


