
/*	RST 0 Handler for Z80 Emulator

***	Copyright:	1994-1996 Günter Woigk
  				mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	01.07.94:	Started work on this file								kio !
    20.Oct.01   revised for CW7Pro and ANSI C compliance				hdusel@bnv-gz.de
*/

#include	"kio.h"
#include	"z80.h"
#include	"keyboard.h"
#include	"display.h"
#include	"application.h"
#include	"messages.h"
#include	"menue.h"
#include	"cassette.h"
#include	"terminal.h"
#include	"load+save.h"
#include	"z80 main.h"
#include	"strings.h"


//	Diese Datei enthält Austauschroutinen in 68000er Code für Spectrum ROM Routinen
//	----------------------------------------------------------------------------
//
//	Aufruf aus Z80 code einfach mit einem RST0


void rst0_load_verify (void);
void rst0_save (void);


// -----	Handle RST 0 patch -------------------------------------------------------
void Handle_RST0()
{	switch ( PC-1 )
	{
	case 0x0556:							//  Load / Verify Tape
		rst0_load_verify();
		CYCLES=0;
		frames_done=frames-1;
		return;
	case 0x04c2:							// Save Tape
		rst0_save();		
		CYCLES=0;
		frames_done=frames-1;
		return;
	case	0x03d5:							// BEEP
/*		this is a very low level patch:
		the timing loop inside the beep function is patched.
		output to mic/ear is tracked by normal Do_Output() function !!!

03D5   0C				INC	C			<-- patched with RST 0
03D6   0D		HALBZE:	DEC	C
03D7   20FD			JR	NZ,HALBZE
03D9   0E3F			LD	C,3FH
03DB   05 			DEC	B
03DC   C2D603			JP	NZ,HALBZE
03DF
		time consumed = 1024 * B  +  16 * C  -  988  T cycles
		minus 11 T cycles which are already counted by RST 0 command in Z80()
*/
		CYCLES -= 1024*(long)RB+16*(long)RC-999;
		BC = 0;
		PC = 0x03DF;
		return;
	default:	
		if (msg>=log_level)
		{	PrintStr ( CStr(localized("RST 0 Instruction executed at:", "RST 0 Befehl bei Adresse:" )));
			PrintHex4 ( PC-1 );
			PrintNL ( );
		}
		Poke(--SP,PCH);
		Poke(--SP,PCL);
		PC = 0;
//		LoadResROM(48,rom[0]);			// why ?!?
		return;
	};
}


// -----	Load / Verify Trap ----------------------------------------------
//		CY -> Load
//		NC -> Verify
// 		A  = block type
// 		DE = length
// 		IX = address
// 		Return NC -> Break
// 		Return CY -> ok
void rst0_load_verify (void)		
{		
	PrintDebug(CatStr(CStr(RF&1?"Load address: ":"Verify address: "),NumStr(IX)));
	PrintDebug(CatStr(CStr("Data length:  "),NumStr(DE)));

	switch ( RF&1 ? LoadBlock ( nil, RA, IX, DE ) : VerifyBlock ( nil, RA, IX, DE ) )
	{	
	case noErr:	// ok
		PrintDebug(CStr(RF&1?"loaded ok":"verified ok"));
		PC = 0x053f;				// return address
		RF |= 1;					// set carry -> OK
		return;
	case 'ubrk':	// User breaked
		PrintDebug(CStr("breaked"));
		RF &= ~1; 				// nc -> error
		PC = 0x0552;				// -> Break
		return;
	case 'wtyp':	// falscher Blocktyp
		PrintDebug(CStr(localized("Wrong block type","falscher Blocktyp")));
		PC = 0x053f;				// return address
		RF &= ~1; 				// nc -> error
		return;
	case 'rder':	// Lesefehler
		PrintDebug(CStr(localized("Read error","Lesefehler")));
		PC = 0x053f;				// return address
		RF &= ~1; 				// nc -> error
		return;
	case 'wcrc':	// Prüfsumme falsch
		PrintDebug(CStr(localized("Wrong checksum","Prüfsumme falsch")));
		PC = 0x053f;				// return address
		RF &= ~1; 				// nc -> error
		return;
	case 'wait':	// Band läuft nicht
		PC--;					// rst 0 will be executed immediately again
		Wait ( 6 );					// wait 1/10 sec
		return;
	default:
		PrintDebug(CStr(localized("Unknown error (rst0_load_verify)","Unbekannter Fehler (rst0_load_verify)")));
		PC = 0x053f;				// return address
		RF &= ~1; 				// nc -> error
		return;
	};
}


// -----	SAVE Ersatzroutine -------------------------------------------------
// 		A  = block type
// 		DE = length
// 		IX = address
// 		Return NC -> Break
// 		Return CY -> ok

void rst0_save (void)
{
	PrintDebug(CatStr(CStr("Save address: "),NumStr(IX)));
	PrintDebug(CatStr(CStr("Data length:  "),NumStr(DE)));

	switch ( SaveBlock ( nil, RA, IX, DE ) )
	{	
	case noErr:	// ok
		PrintDebug(CStr("saved ok"));
		PC = 0x053f;					// return address
		RF |= 1;						// set carry -> OK
		return;
	case 'ubrk':	// User breaked
		PrintDebug(CStr("breaked"));
		PC = 0x0552;					// -> Break
		RF &= ~1; 					// nc -> error
		return;
	case 'wrer':	// Schreibfehler
		PrintDebug(CStr(localized("Write error","Schreibfehler")));
		PC = 0x053f;					// return address
		RF &= ~1; 					// nc -> error
		return;
	case 'wait':	// Band läuft nicht
		PC--;
		Wait (6);	// wait 1/10 sec
		return;
	default:
		PrintDebug(CStr(localized("Unknown error (rst0_save)","Unbekannter Fehler (rst0_save)")));
		PC = 0x053f;					// return address
		RF &= ~1; 					// nc -> error
		return;
	};
}








