/*	Z80 Emulator: CB instruction handlers

***	Copyright:	1995-1999 Günter Woigk
  				mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	based on fMSX; Copyright (C) Marat Fayzullin 1994,1995

	Handler for CB prefixed instructions
	All opcodes covered, no unsupported opcodes!


note:

this is the 'compact version' which fit's in a 32k code segment for 68020 version
and which is faster than a case construct with 256 distinct handlers on a PowerPC, 
probably due to more cache hits. Therefore i have removed the version with the 256
handlers on 19.Nov.1995. KIO !

    20.Oct.01   revised for CW7Pro and ANSI C compliance									hdusel@bnv-gz.de
	20.Oct.01	#define USE_INLINE to replace macros by inline functions (req. for CW7Pro)	hdusel@bnv-gz.de
*/


COUNT_CB_INSTR;			// profiler
increment_r;			// prefixed instructions increment r by 2
time(8);				// all register operations take 8 T cycles


switch(PEEK(pc)&0x07)		// register; b,c,d,e,h,l,(hl),a
{	
case 0:	c=RB; break;
case 1:	c=RC; break;
case 2:	c=RD; break;
case 3:	c=RE; break;
case 4:	c=RH; break;
case 5:	c=RL; break;
case 6:	more(4);c=PEEK(HL); break;
case 7:	c=ra; break;
};

switch (PEEK(pc)>>3)		// instruction: shift/bit/res/set
{
case RLC_B>>3:	M_RLC(&rf, &c);		break;		
case RRC_B>>3: 	M_RRC(&rf, &c);		break;
case RL_B>>3:	M_RL(&rf, &c);		break;
case RR_B>>3:	M_RR(&rf, &c);		break;
case SLA_B>>3: 	M_SLA(&rf, &c);		break;
case SRA_B>>3: 	M_SRA(&rf, &c);		break;
case SLL_B>>3: 	M_SLL(&rf, &c);
#if INFO_ILLEGALS
	switch(n(&pc)&0x07)		// write back result & incr pc
	{	
	case 0:	RB=c; loop_ill2;
	case 1:	RC=c; loop_ill2;
	case 2:	RD=c; loop_ill2;
	case 3:	RE=c; loop_ill2;
	case 4:	RH=c; loop_ill2;
	case 5:	RL=c; loop_ill2;
	case 6:	more(3); POKE(HL,c); loop_ill2;
	case 7:	ra=c; loop_ill2;
	};
#else
								break;
#endif
case SRL_B>>3:	M_SRL(&rf, &c);		break;
case BIT0_B>>3:	M_BIT(&rf, 0x01,c);	pc++;	loop;
case BIT1_B>>3:	M_BIT(&rf, 0x02,c);	pc++;	loop;
case BIT2_B>>3:	M_BIT(&rf, 0x04,c);	pc++;	loop;
case BIT3_B>>3:	M_BIT(&rf, 0x08,c);	pc++;	loop;
case BIT4_B>>3:	M_BIT(&rf, 0x10,c);	pc++;	loop;
case BIT5_B>>3:	M_BIT(&rf, 0x20,c);	pc++;	loop;
case BIT6_B>>3:	M_BIT(&rf, 0x40,c);	pc++;	loop;
case BIT7_B>>3:	M_BIT(&rf, 0x80,c);	pc++;	loop;
case RES0_B>>3:	c&=~0x01;		break;
case RES1_B>>3:	c&=~0x02;		break;
case RES2_B>>3:	c&=~0x04;		break;
case RES3_B>>3:	c&=~0x08;		break;
case RES4_B>>3:	c&=~0x10;		break;
case RES5_B>>3:	c&=~0x20;		break;
case RES6_B>>3:	c&=~0x40;		break;
case RES7_B>>3:	c&=~0x80;		break;
case SET0_B>>3:	c|=0x01;		break;
case SET1_B>>3:	c|=0x02;		break;
case SET2_B>>3:	c|=0x04;		break;
case SET3_B>>3:	c|=0x08;		break;
case SET4_B>>3:	c|=0x10;		break;
case SET5_B>>3:	c|=0x20;		break;
case SET6_B>>3:	c|=0x40;		break;
case SET7_B>>3:	c|=0x80;		break;
};

switch(n(&pc)&0x07)		// write back result and increment pc
{	
case 0:	RB=c; loop;
case 1:	RC=c; loop;
case 2:	RD=c; loop;
case 3:	RE=c; loop;
case 4:	RH=c; loop;
case 5:	RL=c; loop;
case 6:	more(3); POKE(HL,c); loop;
case 7:	ra=c; loop;
};

loop;						// never used




