/*	Z80 Emulator

***	Copyright:	1995-1999 Günter Woigk
  				mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	based on fMSX; Copyright (C) Marat Fayzullin 1994,1995

	16.Jan.95 	Fixed wrong calculation of parity flag (68k version)				KIO !
	28.Mar.95 	Started work on this file											KIO !
	01.Apr.95	Finished, yee it works!												KIO !
	06.Apr.95	Removed "final bug": im2 interrupt handling							KIO !
	12.Jun.95	revised EI_WITH_DELAY handling										KIO !
	12.Jun.95	all Info-Calls and profiling added									KIO !
	12.Jun.95	exact r register emulation completed								KIO !
	13.Jun.95	All opcodes covered, no unsupported opcodes!						KIO !
	01.Nov.95 	Paged memory supported												KIO !
	26.Nov.95	Engine now completely hardware independent!							KIO !
	28.Mar.96	Core2Core() works even if src and dest block overlap at both ends	KIO !
	28.Mar.96	removed option to switch off latency of 'EI' instruction			KIO !
	02.Mai.96	load_cc after IN and OUT added to make extra wait states possible	KIO !

    20.Oct.01   revised for CW7Pro and ANSI C compliance							hdusel@bnv-gz.de

	20.Oct.01   Replaced some defines by code which will be
                inlined on demand. This was necessary because
                the macros caused malformed code if will be
                compiled with CW7.0 pro ! - However yo have the 
                choice to switch back to the old macros if you
                omit to define the new preprocessor symbol
                'USE_INLINE' which has been introduced to support
                the replacement by inlined code.
                
                The new strategy to use inlined code instead
                of just using the preprocessor requires some
                adaptions in the old macros as well.                 				hdusel@bnv-gz.de
*/

#include	<string.h>

#include	"kio.h"				// all-times needed


#include	"z80.options"		// options
#include	"z80.h"				// main header file for z80 engine
#include	"z80_opcodes.h"		// opcode enumerations for case constructs
#include	"Z80_ppc.macros"	// memory accesss, opcode defining macros etc.


/* ----	The following defines should be made in "z80.options":
#define	maxrampages		max. number of ram pages 				e.g. 3
#define	maxrompages		max. number of rom pages				e.g. 1
#define	pagesize		page_size								e.g. 0x4000
#define	pagemask		page_size - 1							e.g. 0x3FFF
#define	pagebits		significant address bits inside page	e.g. 14
#define	corepages		pages inside Z80 address space			e.g. 4
*/


// -----	The Z80 core --------------------------------------------------------------
Char**	CORE_M		= nil;		// master handle for all allocated RAM and ROM
Char*	CORE 		= nil;		// do not use any longer!!!
Char*	rpage[corepages];		// mapped in pages for reading
Char*	wpage[corepages];		// mapped in pages for writing
Char*	ram[maxrampages];		// real RAM pages
Char*	rom[maxrompages];		// real ROM pages
Char*	norom		= nil;		// dummy page for writing to ROM


// -----	The Z80 registers --------------------------------------------------
z80		zreg;					// contains z80 registers on entry/return of Z80()


// -----	Flags after logic operations -------------------------------------
Char		zlog_flags[256];	// convert:   A -> Z80-flags with V=parity, S, Z, and C=0
#if !GENERATINGPOWERPC
Char		mlog_flags[256];	// convert:   A-register  ->  68000 flags with V=parity and C=0
Char		z80flags[256];		// convert:   68000 flagbyte  ->  Z80 flagbyte
Char		m68flags[256];		// convert:   Z80 flagbyte  ->  68000 flagbyte
#endif


// ----	Misc. ----------------------------------------------------------------
long		tfu;				// after read/write to file: bytes actually transferred


// ==================================================================================

/* ----	Initialialize the z80 engine --------------------------------------------
		allocate rom, ram, and norom pages
		initialize rom[], ram[] and norom
		initialize rpage[] and wpage[]
		CORE points to reset core which is allocated in 4 consecutive pages!
		init random generator for r register handling
		calculate flag tables
 		initialize zreg
		reset z80
		On a 68k machine:
		• CORE is aligned to a multiple of 0x00010000
		• some byte behind the last byte a preset with rst0 opcodes
*/
void Init_Z80 ( long roms, long rams )
{	long i;
	OSErr err=noErr;
	Char *p;

	if (roms<1||rams<corepages-1||roms>maxrompages||rams>maxrampages) do_abort("\pInit_Z80()","\pParameter out of range",0);

// allocate dummy write page for ROMs
	if (!norom)	norom = (Char*)NewPtr(pagesize);
	if (!norom) abort_oomem("\pInit_Z80()");

// dispose of old core
	if (CORE_M) DisposHandle((Handle)CORE_M);
	CORE = nil; CORE_M = nil;

#if !GENERATINGPOWERPC 
//	put rom[0] and ram[0...2] in a memory block aligned to a multiple of 0x10000
//	additional ram and rom is located physically before pointer CORE !!!
	CORE_M = (Char**)TempNewHandle ((roms+rams)*pagesize+0x10010,&err);
	if (!CORE_M||err) abort_oomem ( "\pInit_Z80()" );
	HLock((Handle)CORE_M);
	CORE = *CORE_M+(roms+rams)*pagesize;
	CORE -= (Short)CORE;
	memset ( CORE+0x10000, 199, 0x10 ); 	// rst 0
#else
// simpler allocation in PowerPC version
	CORE_M = (Char**)TempNewHandle ((roms+rams)*pagesize,&err); 
	if (err||!CORE_M) abort_oomem("\pInit_Z80()");
	HLock((Handle)CORE_M);
	CORE = *CORE_M + (roms+rams-4)*pagesize;	// to mimic 68k version
#endif

// initialize rom[] and ram[]
// one ROM at start of CORE, rest is RAM
	rom[0] = p=CORE;
	for (i=0;i<corepages-1;i++) ram[i] = p+=pagesize;	

	p=CORE;
	for (i=1;i<roms;) 			rom[i++] = p-=pagesize;	while (i<maxrompages) rom[i++] = NULL;
	for (i=corepages-1;i<rams;) ram[i++] = p-=pagesize;	while (i<maxrampages) ram[i++] = NULL;

// initialize rpage[], wpage[]
	for (i=corepages;i--;)	rpage[i] = wpage[i] = CORE;
	wpage[0] = (Char*)norom;

// init z80 registers:
	BC=DE=HL=IX=IY=PC=SP=BC2=DE2=HL2=0;		// clear registers
	RA=RA2=RF=RF2=RR=RI=0;
	IFF1=IFF2 = disabled;					// disable interrupts
	IM = 1;									// interrupt mode := 1	
	WUFF = 0;								// clear irpt, nmi
	EXIT = 0;								// clear watchdog
	IRPTCMD = 0xff;							// byte read from bus in im0 or im2 mode

// build flag tables:
// Z80:	%SZ-H-VNC	--- H & N ignored
// m68:	%---XSZVC	--- X ignored
// the unused flags correspond as follows:   z80{%--1H2-N-}  ==  m68{%1H2N----}	   11.12.94 KIO !
	for (i=0; i<256; i++)
	{	
	zlog_flags[i]	= (i&0x80) + ((i==0)<<6) + (  ( (~i+(i>>1)+(i>>2)+(i>>3)+(i>>4)+(i>>5)+(i>>6)+(i>>7))&1 ) << 2  );
#if !GENERATINGPOWERPC
	mlog_flags[i]	= ((i>=128)<<3) + ((i==0)<<2) + (  ( (~i+(i>>1)+(i>>2)+(i>>3)+(i>>4)+(i>>5)+(i>>6)+(i>>7))&1 ) << 1  );
	z80flags[i]	= ((i<<4)&0xc0) + ((i<<1)&4) + (i&1)   +   ((i>>2)&0x38) + ((i>>3)&2);
	m68flags[i]	= ((i>>4)&0x0c) + ((i>>1)&2) + (i&1)   +   ((i<<2)&0xe0) + ((i<<3)&0x10);
#endif
	};
}



// ====	Utilities ========================================================


// ----	Peek, Poke and calculate EA for peeking and poking -----------------------------
//		These are inappropriate for macros, because the address is used twice
//		note: rpage[] and wpage[] pointers do not point to the first byte of a page,
//		they are defined that way, that the unmasked target address must be added
//		to point to the right byte in the Macintosh's RAM !

Char*  RdPtr ( Short addr )	{  return  rpage[addr>>pagebits] + addr; 	}	// ptr for reading
Char*  WrPtr ( Short addr )	{  return wpage[addr>>pagebits] + addr; 	}	// ptr for writing

Char   Peek  ( Short addr )				{  return  rpage[addr>>pagebits][addr];  		}
void   Poke  ( Short addr, Char byte )	{		   wpage[addr>>pagebits][addr] = byte;	}


// ----	Write from/to core, buffer and file -------------------------------------------
//		file operations:	return 0 if ok else OS error
//							'tfu' contains number of bytes actually transferred


/* ----	Write buffer to file ----------------------------------------------------------
*/
OSErr Buffer2File ( void* bu, short fileID, long len )
{	
	tfu = len;
	return FSWrite ( fileID, &tfu, bu );
}


/* ----	Read buffer from file --------------------------------------------------------
*/
OSErr File2Buffer ( short fileID, void* bu, long len )
{
	tfu = len;  
	return FSRead ( fileID, &tfu, bu );
}


/* ----	Write core to file ------------------------------------------------------------
		saves what the z80 cpu would read
*/
OSErr Core2File ( Short addr, short fileID, long len )
{	OSErr err; long cnt;

	tfu = 0;
	for ( err=noErr; !err && len; )
	{	cnt = min ( len, pagesize-(addr&pagemask) );
		err = FSWrite ( fileID, &cnt, RdPtr(addr) );
		addr += cnt;
		len -= cnt;
		tfu += cnt;
	}
	return err;
}


/* ----	Read core from file -----------------------------------------------------------
		stores what the z80 cpu would write
		ROM is automatically protected
*/
OSErr File2Core ( short fileID, Short addr, long len )
{	OSErr err; long cnt;

	tfu = 0;
	for ( err=noErr; !err && len; )
	{	cnt = min ( len, pagesize-(addr&pagemask ) );
		err = FSRead ( fileID, &cnt, WrPtr(addr) );
		addr += cnt;
		len -= cnt;
		tfu += cnt;
	}
	return err;
}


/* ----	Copy core to buffer -----------------------------------------------------------
		copies what the z80 cpu would read
*/
void Core2Buffer ( Short addr, void* bu, long len )
{	long cnt;

	while ( len )
	{	cnt = min ( len, pagesize-(addr&pagemask ) );
		memcpy ( bu, RdPtr(addr), cnt );
		bu = (Char*)bu+cnt;
		addr += cnt;
		len -= cnt;
	}
}

/* ----	Copy buffer to core ------------------------------------------------------------
		copies what the z80 cpu would write
*/
void Buffer2Core ( void* bu, Short addr, long len )
{	long cnt;

	while ( len )
	{	cnt = min ( len, pagesize-(addr&pagemask ) );
		memcpy ( WrPtr(addr), bu, cnt );
		bu = (Char*)bu+cnt;
		addr += cnt;
		len -= cnt;
	}
}


/* ----	Core to core copying ----------------------------------------------------------------
		There are two functions which mimic the Z80 LDIR and LDDR instruction
		and a function which copies memory correctly, even if source and destination overlap at both ends.
		If len=0x0000 then the entire core is moved.
*/

void Do_LDIR ( Short src, Short dest, Short len )
{	Short cnt; Char *q,*z;

	do
	{	q = RdPtr(src); z = WrPtr(dest);
		cnt = pagesize-max(src&pagemask,dest&pagemask);
		if (cnt>len)  cnt = len?len:cnt;
		src+=cnt; dest+=cnt; len-=cnt;
		while (cnt--) *z++=*q++;
	} while (len);
}

void Do_LDDR ( Short src, Short dest, Short len )
{	Short cnt; Char *q,*z;

	do
	{	q = RdPtr(src); z = WrPtr(dest);
		cnt = 1+Min(src&pagemask,dest&pagemask);	// kio 12.feb.2000
		if (cnt>len && len!=0)  cnt = len;
		src-=cnt; dest-=cnt; len-=cnt;
		while (cnt--) *z--=*q--;
	} 
	while (len);
}


void Core2Core ( Short src, Short dest, Short len )
{	long l,d,s; Ptr p;
	
	l = len ? len : 0x10000;		// block size

	d = (Short)(src-dest);			// distance for LDIR
	if ( d+l <= 0x10000 )			// LDIR possible ?
	{	Do_LDIR ( src, dest, len );
		return;
	}

	if ( d>0x08000 )				// LDDR distance < LDIR distance ? (allows bigger block to copy)
	{	d = (Short)(dest-src);		// distance for LDDR
		if ( d+l <= 0x10000 )		// LDDR possible ?
		{	Do_LDDR ( src+len-1, dest+len-1, len );
			return;
		}
	}
	
//	source and destination blocks overlap at both ends -> use buffer
	s = d + l - 0x10000;
	p = NewPtr ( s ); if (!p) abort_oomem("\pCore2Core()");
	Core2Buffer ( src, p, s );
	Core2Core ( src+s, dest+s, len-s );
	Buffer2Core ( p, dest, s );
	DisposPtr ( p );
}


/* ----	Copy buffer to buffer (just for completeness) ------------------------------------
*/
void Buffer2Buffer ( void* src, void* dest, long len )
{
	memmove ( dest, src, len );
}


/* ==========================================================================================
   ====	The Z80 Engine ======================================================================
   ==========================================================================================

	Features:
	•	all legal instructions
	•	all known illegal instructions (no opcode gaps)
	•	all flags (8 bit)
	•	exact T cycle counting
		-> exact speed
		-> timing loop depending sound
		-> timing loop depending tape i/o
		-> high resolution screen update
	•	ROM write protection
	•	paged memory support
	•	exact R register emulation
	•	1 instruction latency after EI
	•	notification when processing 'interesting' legal instructions
	•	notification when processing illegal and weird illegal instructions
	•	instruction and address profiling
	•	single stepping

	global struct zreg keeps z80 registers on entry and return of Z80_PPC()
	some registers are copied to local variables while Z80_PPC() is running
	rpage[] and wpage[] contain pointers pointing to the paged-in memory pages for 
	reading and writing. This provides paged memory support and ROM write protection.
	
	Z80_PPC() processing priority:
	1.	process NMI, if NMI is pending
	2.	process interrupt, if IRPT is pending and not disabled
	3.	execute z80 instructions, until T cycle counter expires
		Interrupts are only tested again after an EI instruction
	4.	if T cycle counter reaches 0, Do_Cycles() is called which may
		•	reload the T cycle counter
		•	trigger NMI or IRPT
		•	set EXIT requiring Z80_PPC() to exit

	Single instructions may be executed by calling Z80_PPC() with CYCLES=4.
	Then only NMI or interrupt request is processed (if pending and enabled) or the next
	instruction is executed (EI and the next instruction is processed in one go)

	If Z80_PPC() returns, it's result value indicates one of the following conditions:
	•	EXIT was set by Do_Cycles()
	•	halt instruction executed at pc-1
	•	rst0 instruction executed at pc-1
	•	interrupt in mode 0 read not supported instruction from bus (actually commented out)
	•	illegal instruction at pc-2/-3/-4 was executed
	•	weird illegal instruction at pc-2/-3/-4 was executed
*/


// ----	jumping -----------------------------------------------------------
#define	loop		goto nxtcmnd			// loop to next instruction
#define	exit(N)		{ c=N; goto exit_z80; }	// exit Z80_PPC()


// ----	load and store local variables from/to global registers ------------
#define	store_cc	CYCLES = cc
#define	load_cc		cc = CYCLES

#define	load_registers								\
	load_r;				/* refresh counter R	*/	\
	load_cc;			/* T cycle counter		*/	\
	pc = PC;			/* program counter PC	*/	\
	sp = SP;			/* stack pointer SP		*/	\
	ra = RA;			/* register A			*/	\
	rf = RF;			/* register F			*/

#define	save_registers								\
	store_r;			/* refresh counter R	*/	\
	store_cc;			/* T cycle counter		*/	\
	PC = pc;			/* program counter PC	*/	\
	SP = sp;			/* stack pointer SP		*/	\
	RA = ra;			/* register A			*/	\
	RF = rf;			/* register F			*/


/* ====	The Z80 ENGINE ====================================================
*/
short Z80_PPC()
{	

register long	cc;							// T cycle counter

Short	pc;							// z80 program counter
Short	sp;							// z80 stack pointer
Char	ra;							// z80 a register
Char	rf;							// z80 flags

#define			time(N)	cc-=N-4				// T cycles for instruction
#define			more(N)	cc-=N				// adjust T cycles after use of time()

register Short	*izp;						// pointer to IX or IY after 0xDD or 0xFD
#define			iz		(*izp)				// IX or IY
#define			rzh		(*(Char*)izp)		// XH or YH
#define			rzl		(*(((Char*)izp)+1))	// XL or YL

Char	c;							// general purpose byte register
	
register Short	w;							// general purpose word register
#define			wl		(Char)w				// access low byte of w
#define			wh		(w>>8)				// access high byte of w


#ifndef USE_INLINE
    register Short	wm;							// scratch for macro internal use
    #define			wml		(Char)wm			// access low byte of wm
    #define			wmh		(wm>>8)				// access high byte of wm
#endif

long			ccx = 0;					// limit for cc -> EI_WITH_DELAY handling

#if EXACT_R
register Char	r;							// r register bit 0...6
#endif


// load local variables from global registers:
	load_registers;
	

// check for interrupts:
// interrupts are only tested once at start of Z80_PPC()
// and only tested again after EI processing!
check_wuff:
	if (WUFF)
	{	if (WUFF&0x80)					// NMI:  0x0066; 11 T cycles
		{	do_info_nmi;
			WUFF&=0x7F;
			IFF2=IFF1;
			IFF1=disabled;
			PUSHD(&sp, pc);
			pc = 0x0066;
			more(11);
			loop;
		}
		if (IFF1==enabled)				// interrupts enabled?
		{	do_info_irpt;
			WUFF--;						// clear interrupt request
			IFF1=IFF2=disabled;			// disable iff1 & iff2
			PUSHD(&sp, pc);					// push actual pc
			switch (IM)
			{
			case 0:			// mode 0: read instr; 1 T cycle + normal instr timing
			//	if (IRPTCMD==RST38)
			//	{	pc = 0x0038;
			//		more(12);		// RST==11T + 1T
			//		loop;
			//	};
			//	exit(irpt_error);	// not supported instruction read in irpt ackn cycle
			case 1:			// mode 1:  rst 56; 13 T cycles
				pc = 0x0038;
				more(13);
				loop;
			case 2:			// mode2:  jmp via table; 19 T cycles
				w = IRPTVEK;
				pc = PEEK(w+1)*256 + PEEK(w);
				more(19);
				loop;
			default:
				exit(irpt_error);
			};
		};
	};

// ----	THE MAIN INSTRUCTION DISPATCHER --------------------------------

nxtcmnd:
	if ((cc-=4)>=ccx)				// -4 T cycles for most frequent instructions
	{	
#include "Codes.c"
	}
	else
	{	cc+=4;
		if (ccx) { ccx=0; goto check_wuff; }	// for correct ei handling
		store_cc; 
		Do_Cycles(); 
		load_cc;
		if (!EXIT) goto check_wuff;
		exit(watchdog_irpt);
	};


// ----	Exit points -------------------------------------------------------------

#if INFO_ILLEGALS
exit_ill2:	if (!Z80_Info_Illegal(cc,pc-2)) loop;	exit(ill_instr);
exit_ill3:	if (!Z80_Info_Illegal(cc,pc-3)) loop;	exit(ill_instr);
exit_ill4:	if (!Z80_Info_Illegal(cc,pc-4)) loop;	exit(ill_instr);
#endif

#if INFO_WEIRD
exit_weird2:if (!Z80_Info_Weird (cc,pc-2))	loop;	exit(weird_instr);
exit_weird3:if (!Z80_Info_Weird (cc,pc-3))	loop;	exit(weird_instr);
exit_weird4:if (!Z80_Info_Weird (cc,pc-4))	loop;	exit(weird_instr);
#endif

exit_z80:				// write back registers and return
	save_registers;
	return c;			// result code
}



