
/*	Z80 Emulator

***	Copyright:	1994-2000 Günter Woigk
  				mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	01.Mai.94 	Started work on this file					KIO !
	11.Dec.94 	Now preserving all bits with push/pop AF	KIO !
	18.Jan.95 	Improved interrupt dispatcher				KIO !
	22.Jan.95 	Trace mode									KIO !
	28.Jan.95 	CMD_PROFILE									KIO !
	05.Mar.95	full r-register emulation					KIO !
	12.Mar.95	T cycle counter								KIO !
	17.Mar.95	revised exception handling of Z80()			KIO !
	29.Mrz.95 	Separated this file from Init_Z80.c			KIO !
	10.Apr.95 	ROM write protection						KIO !
	26.Oct.96	removed hiwords (core base addr.) from zreg	KIO !
	23.Jan.2000	found bug in bc register access macro		kio !
*/


#include	"kio.h"

#include	"z80.options"
#include	"z80.h"
#include	"z80.macros"


/* -----	The Z80 engine ----------------------------------------------------------

	• T cycle counting  =>  timing loop depending sound, tape i/o possible
	• ROM write protection

	On entry/return some registers are loaded from and written back to struct zreg
	• a and a2 are loaded into one long register: a
	• f and f2 are converted and loaded into one long register: f
	• pc and sp are loaded into address registers, adding core base address
	• bc, de, hl, ix, iy are loaded into a local struct, adding core base address

	WUFF (irpt and nmi flags) is tested
	•  on entry of Z80()
	•  after ei instruction
	•  after Do_Cycles() after CYCLES reached 0 and EXIT is not set 
	
	EXIT (watchdog) and WUFF (nmi/irpt) are tested
	•  Do_Cycles() is called after CYCLES reaches 0
	    after Do_Cycles() returns, EXIT and WUFF are tested

	Single instructions may be executed by calling Z80() with CYCLES=4
	Then one of the following is processed: (in priority order)
	•  handle nmi request
	•  handle irpt request (if interrupts are enabled)
	•  execute one instruction, except:
	•  if instruction is EI, handle EI and next instruction in one go

	If Z80() returns, it's result value indicates one of the following conditions:
	•  not supported instruction at zreg.IP
	•  rst0 instruction at zreg.IP-1
	•  halt instruction at zreg.IP-1
	•  EXIT set
*/


// registers bc,de,hl,ix,iy are copied to long variables;
// highword = core base address
typedef union ppx ppx;
union ppx
{	Char		*ptr;
	struct	{	Short corebase; Short reg; }	rr;
	struct	{	Short corebase; Char hi,lo; }	r;
};

struct
{	ppx	bc,de,hl,ix,iy;
}	lreg;


// undefine register definitions, which would use the zreg struct
#undef	BC
#undef	DE
#undef	HL
#undef	IX
#undef	IY

#undef	RB
#undef	RC
#undef	RD
#undef	RE
#undef	RH
#undef	RL
#undef	XH
#undef	XL
#undef	YH
#undef	YL


// define register definitions to use the lreg struct
#define	AHL	lreg.hl.ptr
#define	ADE	lreg.de.ptr
#define	ABC	lreg.bc.ptr			// 23.jan.2000 kio

#define	BC	lreg.bc.rr.reg
#define	DE	lreg.de.rr.reg
#define	HL	lreg.hl.rr.reg
#define	IX	lreg.ix.rr.reg
#define	IY	lreg.iy.rr.reg

#define	RB	lreg.bc.r.hi
#define	RC	lreg.bc.r.lo
#define	RD	lreg.de.r.hi
#define	RE	lreg.de.r.lo
#define	RH	lreg.hl.r.hi
#define	RL	lreg.hl.r.lo
#define	XH	lreg.ix.r.hi
#define	XL	lreg.ix.r.lo
#define	YH	lreg.iy.r.hi
#define	YL	lreg.iy.r.lo


// -----	T cycle counting and ROM write protection ------------------------------

#define	rom_protection	1
#define	exact_bit		1
#define	exact_xy_bit	1

asm short Z80 ( )
{
#include	"Z80_proc.c"
}









